@@ -14,7 +14,7 @@ which is a tad faster.
 
 #define CSS_ERROR_NONE 0
 #define CSS_ERROR_NOSEL 1
-#define CSS_ERROR_AT 2
+#define CSS_ERROR_MANYNOT 2
 #define CSS_ERROR_BRACES 3
 #define CSS_ERROR_SEL0 4
 #define CSS_ERROR_NORULE 5
@@ -29,13 +29,18 @@ which is a tad faster.
 #define CSS_ERROR_PE 14
 #define CSS_ERROR_UNSUP 15
 #define CSS_ERROR_MULTIPLE 16
-#define CSS_ERROR_DELIM 17
-#define CSS_ERROR_LAST 18
+#define CSS_ERROR_ATEMPTY 17
+#define CSS_ERROR_NOTMEDIA 18
+#define CSS_ERROR_BADMEDIA 19
+// DELIM is special not really an error, true errors come before
+#define CSS_ERROR_DELIM 20
+#define CSS_ERROR_ATPROC 21
+#define CSS_ERROR_LAST 22
 
 static const char *const errorMessage[] = {
 	"ok",
 	"no selectors",
-	"@",
+	"many selectors under not",
 	"nested braces",
 	"empty selector",
 	"no rules",
@@ -48,11 +53,14 @@ static const char *const errorMessage[] = {
 	"tag link",
 	"inject high",
 	"pseudo element",
-	": unsupported",
+	":unsupported",
 	"multiple",
+	"@ empty",
+	"@ not media",
+	"@ bad media",
 	"==========",
-	"==========",
-	0
+	"@ processed",
+	"eof",
 };
 
 static int errorBuckets[CSS_ERROR_LAST];
@@ -85,7 +93,8 @@ static int closeString(char *s, char delim)
 {
 	int i;
 	char c, qc = 0;		// quote character
-	if (delim != '[')
+	int pc = 0;		// paren count
+	if (delim == '"' || delim == '\'')
 		qc = delim;
 	for (i = 0; (c = s[i]); ++i) {
 		if (c == qc) {	// close quote
@@ -103,9 +112,15 @@ static int closeString(char *s, char delim)
 // something is wrong or misaligned. End the string here.
 		if (c == '\n')
 			return i + 1;
-		if (c == ']' && delim == '[' && !qc)
+		if (qc)
+			continue;
+		if (c == ']' && delim == '[')
 			return i + 1;
-		if (!qc && (c == '"' || c == '\''))
+		if (c == '(' && delim == '(')
+			++pc;
+		if (c == ')' && delim == '(' && --pc < 0)
+			return i + 1;
+		if (c == '"' || c == '\'')
 			qc = c;
 	}
 	return -1;
@@ -233,6 +248,7 @@ struct desc {
 	struct sel *selectors;
 	struct rule *rules;
 	int highspec;		// specificity when this descriptor matches
+	bool underat;
 };
 
 // selector
@@ -260,6 +276,8 @@ struct mod {
 	struct mod *next;
 	char *part;
 	bool isclass, isid, negate;
+// notchain is used only if negate is true, not(ul.lister), recursive
+	struct asel *notchain;
 };
 
 struct rule {
@@ -291,8 +309,10 @@ struct cssmaster {
 static void cssPiecesFree(struct desc *d);
 static void cssPiecesPrint(const struct desc *d);
 static void cssAtomic(struct asel *a);
+static void cssParseLeft(struct desc *d);
 static void cssModify(struct asel *a, const char *m1, const char *m2);
-static bool onematch, skiproot, gcsmatch, bulkmatch, bulktotal;
+static void chainFree(struct asel *asel);
+static bool onematch, topmatch, skiproot, gcsmatch, bulkmatch, bulktotal;
 static char matchtype;		// 0 plain 1 before 2 after
 static bool matchhover;		// match on :hover selectors.
 static jsobjtype rootobj;
@@ -332,6 +352,73 @@ static void intoShortCache(const char *url, char *data)
 	c->data = data;
 }
 
+void writeShortCache(void)
+{
+	struct shortcache *c;
+	FILE *f;
+	int n = 0;
+	struct cssmaster *cm = cf->cssmaster;
+	if (!cm)
+		return;
+	f = fopen("implocal", "w");
+	if (!f)
+		return;
+	for (c = cm->cache; c; c = c->next) {
+		char filename[20];
+		++n;
+		sprintf(filename, "i%d.css", n);
+		if (!memoryOutToFile(filename, c->data, strlen(c->data),
+				     MSG_TempNoCreate2, MSG_NoWrite2)) {
+/* oops, can't write the file */
+			unlink(filename);
+			showError();
+			continue;
+		}
+		fprintf(f, "%s:%s\n", filename, c->url);
+	}
+	fclose(f);
+}
+
+static void readShortCache(struct cssmaster *cm)
+{
+	FILE *f = fopen("implocal", "r");
+	struct shortcache *c;
+	char *s;
+	int length, n = 0;
+	char line[400];
+	if (!f)
+		return;
+	while (fgets(line, sizeof(line), f)) {
+		s = strchr(line, '\n');
+		if (!s) {
+			fprintf(stderr, "implocal line too long\n");
+			fclose(f);
+			return;
+		}
+		*s = 0;
+		if (s > line && s[-1] == '\r')
+			*--s = 0;
+		if (line[0] == '#' || line[0] == 0)
+			continue;
+		s = strchr(line, ':');
+		if (!s) {
+			fprintf(stderr, "jslocal line has no :\n");
+			continue;
+		}
+		*s++ = 0;
+		c = allocMem(sizeof(struct shortcache));
+		c->url = cloneString(s);
+		fileIntoMemory(line, &c->data, &length);
+// fileIntoMemory puts a null byte on the end
+		c->next = cm->cache;
+		cm->cache = c;
+		++n;
+	}
+	fclose(f);
+	if (n)
+		debugPrint(3, "%d import file replacements", n);
+}
+
 // Step back through a css string looking for the base url.
 // The result is allocated.
 static char *cssBase(const char *start, const char *end)
@@ -421,7 +508,7 @@ copy:
 }
 
 // This is a crude measure of specificity, contained in a 5 digit number.
-static int specificity(const struct sel *sel)
+static int specificity(const struct sel *sel, bool underat)
 {
 	const struct asel *a = sel->chain;
 	const struct mod *mod;
@@ -454,20 +541,195 @@ static int specificity(const struct sel *sel)
 		for (mod = sel->chain->modifiers; mod; mod = mod->next)
 			++n;
 	}
+// @ media tag has more specificity; I have no idea how much more.
+	if (underat)
+		n += 10;
+
 	return n;
 }
 
+static bool mediaPiece(struct desc *d, char *t)
+{
+	char *a;
+	int n, s_width, s_height;
+	bool atnot = false, atsome = false;
+
+	if (!*t)		// nothing there
+		return false;
+
+	if (!strncmp(t, "only", 4) && !isalnum(t[4])) {
+// only is meaningless
+		t += 4;
+		skipWhite2(&t);
+	}
+	if (!strncmp(t, "not", 3) && !isalnum(t[3])) {
+		atnot = true;
+		t += 3;
+		skipWhite2(&t);
+	}
+	if (!strncmp(t, "screen", 6) && !isalnum(t[6])) {
+		atsome = true;
+		t += 6;
+		skipWhite2(&t);
+	} else if (!strncmp(t, "all", 3) && !isalnum(t[3])) {
+		atsome = true;
+		t += 3;
+		skipWhite2(&t);
+	}
+
+and:
+	if (!*t) {
+		if (atsome)
+			return !atnot;
+		d->error = CSS_ERROR_BADMEDIA;
+		return false;
+	}
+
+	if (!strncmp(t, "and", 3) && !isalnum(t[3])) {
+		t += 3;
+		skipWhite2(&t);
+	}
+	if (*t != '(') {
+		d->error = CSS_ERROR_BADMEDIA;
+		return false;
+	}
+	++t;
+	skipWhite2(&t);
+// A well written website can skip the hover text, which edbrowse can't
+// manage very well, so we should respond to this one.
+	if (!strncmp(t, "hover)", 6)
+	    || !strncmp(t, "any-hover)", 10)
+	    || !strncmp(t, "pointer)", 8)
+	    || !strncmp(t, "any-pointer)", 12)
+	    || !strncmp(t, "inverted-colors)", 16))
+		return atnot;
+	if (!strncmp(t, "scripting)", 10)) {
+		if (!isJSAlive)
+			return atnot;
+		t += 10;
+		skipWhite2(&t);
+		atsome = true;
+		goto and;
+	}
+// I only handle min or max on certain parameters
+	if ((strncmp(t, "max", 3) && strncmp(t, "min", 3)) ||
+	    t[3] != '-' || (strncmp(t + 4, "height", 6)
+			    && strncmp(t + 4, "width", 5)
+			    && strncmp(t + 4, "color", 5)
+			    && strncmp(t + 4, "monochrome", 10))) {
+// not recognized
+		d->error = CSS_ERROR_BADMEDIA;
+		return false;
+	}
+	a = t + 9;
+	if (t[4] == 'h')
+		++a;
+	if (t[4] == 'm')
+		a += 5;
+	if (*a != ':') {
+		d->error = CSS_ERROR_BADMEDIA;
+		return false;
+	}
+	++a;
+	skipWhite2(&a);
+	if (!isdigit(*a)) {
+		d->error = CSS_ERROR_BADMEDIA;
+		return false;
+	}
+	n = strtol(a, &a, 10);
+
+// screen is always 1024 by 768
+	s_width = 1024, s_height = 768;
+
+// px or em
+	if (a[0] == 'p' && a[1] == 'x')
+		a += 2;
+	if (a[0] == 'e' && a[1] == 'm') {
+		a += 2;
+		s_width /= 16, s_height /= 16;
+	}
+	if (*a != ')') {
+		d->error = CSS_ERROR_BADMEDIA;
+		return false;
+	}
+
+	if (t[1] == 'i' && t[4] == 'w' && n > s_width)
+		return atnot;
+	if (t[1] == 'a' && t[4] == 'w' && n < s_width)
+		return atnot;
+	if (t[1] == 'i' && t[4] == 'h' && n > s_height)
+		return atnot;
+	if (t[1] == 'a' && t[4] == 'h' && n < s_height)
+		return atnot;
+	if (t[1] == 'i' && t[4] == 'c' && n > 8)
+		return atnot;
+	if (t[1] == 'a' && t[4] == 'c' && n < 8)
+		return atnot;
+	if (t[1] == 'i' && t[4] == 'm' && n > 4)
+		return atnot;
+	if (t[1] == 'a' && t[4] == 'm' && n < 4)
+		return atnot;
+
+	t = a + 1;
+	skipWhite2(&t);
+	atsome = true;
+	goto and;
+}
+
+static bool media(struct desc *d, char *t)
+{
+	char *p = t;
+	char c;
+	int n;
+
+	while ((c = *t)) {
+		if (strchr("'\"[(", c)) {
+			n = closeString(t + 1, c);
+			if (n < 0) {	// should not happen
+				d->error = CSS_ERROR_BADMEDIA;
+				return false;
+			}
+			t += n + 1;
+			continue;
+		}
+		if (c == ',') {
+			*t = 0;
+			trim(p);
+			if (mediaPiece(d, p))
+				return true;
+			p = t + 1;
+		}
+		++t;
+	}
+
+// last piece
+	trim(p);
+	if (mediaPiece(d, p))
+		return true;
+
+	if (!d->error)
+		d->error = CSS_ERROR_ATPROC;
+	return false;
+}
+
+// The above is used by window.matchMedia
+bool matchMedia(char *t)
+{
+	struct desc d0;
+	return media(&d0, t);
+}
+
 // The input string is assumed allocated, it could be reallocated.
 static struct desc *cssPieces(char *s)
 {
 	int bc = 0;		// brace count
 	struct desc *d1 = 0, *d2, *d = 0;
-	struct sel *sel, *sel2;
-	struct asel *asel, *asel2;
+	struct sel *sel;
+	struct asel *asel;
 	int n;
-	char c, last_c;
+	char c;
 	char *lhs;
-	char *a, *t;
+	char *a, *t, *at_end_marker = 0;
 	char *iu1, *iu2, *iu3;	// for import url
 
 	loadcount = 0;
@@ -520,7 +782,7 @@ top:
 				g.thisfile = cf->fileName;
 				g.uriEncoded = true;
 				g.url = newurl;
-				if (httpConnect(&g)) {
+				if (!intFlag && httpConnect(&g)) {
 					if (g.code == 200) {
 						a = force_utf8(g.buffer,
 							       g.length);
@@ -593,6 +855,8 @@ top2:
 	lhs = s;
 
 	while ((c = *s)) {
+		if (s >= at_end_marker)
+			at_end_marker = 0;
 		if (c == '"' || c == '\'') {
 			n = closeString(s + 1, c);
 			if (n < 0) {
@@ -612,9 +876,53 @@ top2:
 				goto copy;
 // bc is 0, end of the descriptor
 			*s++ = 0;
-			lhs = s;
+			lhs = s;	// next descriptor
 			trim(d->lhs);
 			trim(d->rhs);
+
+// some special @ code here
+			t = d->lhs;
+			if (*t != '@')
+				goto past_at;
+// our special code to insert a delimiter into debugging output
+			if (!strncmp(t, "@ebdelim", 8)) {
+				d->error = CSS_ERROR_DELIM;
+				goto past_at;
+			}
+			if (d->bc > 2) {
+				d->error = CSS_ERROR_BRACES;
+				goto past_at;
+			}
+			if (d->bc == 1) {
+				d->error = CSS_ERROR_ATEMPTY;
+				goto past_at;
+			}
+			d->bc = 1;
+			++t;
+			skipWhite2(&t);
+			if (strncmp(t, "media", 5) || isalnum(t[5])) {
+				d->error = CSS_ERROR_NOTMEDIA;
+				goto past_at;
+			}
+			t += 5;
+			skipWhite2(&t);
+			if (!media(d, t))
+				goto past_at;
+			d->error = CSS_ERROR_ATPROC;
+
+// Here comes some funky string manipulation.
+// The descriptors are in rhs.
+// I nulled out the trailing }, make it a space.
+			while (s[-1] == 0)
+				*--s = ' ';
+			at_end_marker = s;
+			lhs = s = d->rhs;
+			d->rhs = "follow";
+
+past_at:
+			if (d->bc > 1)
+				d->error = CSS_ERROR_BRACES;
+
 			if (!d1)
 				d1 = d2 = d;
 			else
@@ -636,12 +944,15 @@ top2:
 			*s++ = 0;
 			d->rhs = s;
 			d->bc = bc = 1;
+			if (at_end_marker)
+				d->underat = true;
 			continue;
 		}
 copy:		++s;
 	}
 
 	nzFree(d);
+
 	if (!d1) {
 // no descriptors, nothing to do,
 // except free the incoming string.
@@ -653,152 +964,9 @@ copy:		++s;
 
 // Now let's try to understand the selectors.
 	for (d = d1; d; d = d->next) {
-		char combin;
-		char *a1, *a2;	// bracket the atomic selector
-
 		if (d->error)
 			continue;
-		s = d->lhs;
-		if (!s[0]) {
-			d->error = CSS_ERROR_NOSEL;
-			continue;
-		}
-// our special code to insert a delimiter into debugging output
-		if (!strncmp(s, "@ebdelim", 8)) {
-			d->error = CSS_ERROR_DELIM;
-			continue;
-		}
-// leading @ doesn't apply to edbrowse.
-		if (s[0] == '@') {
-			d->error = CSS_ERROR_AT;
-			continue;
-		}
-		if (d->bc > 1) {
-			d->error = CSS_ERROR_BRACES;
-			continue;
-		}
-
-		a1 = s;		// start of the atomic selector
-		sel = 0;	// the selector being built
-
-		last_c = 0;
-		while ((c = *s)) {
-			if (c == '"' || c == '\'' || c == '[') {
-				n = closeString(s + 1, c);
-				if (n < 0)	// should never happen
-					break;
-				s += n + 1;
-				last_c = 0;
-				continue;
-			}
-// Ambiguous, ~ is combinator or part of [foo~=bar].
-// Simplistic check here for ~=
-// I don't really need it any more, now that I treat [ stuf ] as a string.
-// But I still need the next one.
-// Ambiguous, + is combinator or part of "nth_child(n+3)
-// Simplistic check here, next selector should not begin with a digit
-			if ((c == '~' && s[1] == '=') ||
-			    (c == '+' && isdigit(s[1]))) {
-				last_c = c;
-				++s;
-				continue;
-			}
-
-			if (last_c == '\\') {
-				last_c = 0;
-				++s;
-				continue;
-			}
-
-			combin = 0;	// look for combinator
-			a2 = s;
-			while (strchr(", \t\n\r>~+", c)) {
-				if (isspace(c)) {
-					if (!combin)
-						combin = ' ';
-				} else {
-					if (combin && combin != ' ')
-						break;
-					combin = c;
-					last_c = c;
-				}
-				c = *++s;
-			}
-			if (!combin) {
-				last_c = c;
-				++s;
-				continue;
-			}
-// it's a combinator or separator
-			last_c = c;
-			if (a2 == a1) {	// empty piece
-// I'll allow it if it's just an extra comma
-				if (combin == ',' && !sel) {
-					a1 = s;
-					continue;
-				}
-				sel->error = CSS_ERROR_SEL0;
-				break;
-			}
-
-			t = allocMem(a2 - a1 + 1);
-			memcpy(t, a1, a2 - a1);
-			t[a2 - a1] = 0;
-			asel = allocZeroMem(sizeof(struct asel));
-			asel->part = t;
-			asel->combin = combin;
-			if (!sel) {
-				sel = allocZeroMem(sizeof(struct sel));
-				if (!d->selectors)
-					d->selectors = sel2 = sel;
-				else
-					sel2->next = sel, sel2 = sel;
-			}
-// I'm reversing the order of the atomic selectors here
-			asel2 = sel->chain;
-			sel->chain = asel;
-			asel->next = asel2;
-
-			cssAtomic(asel);
-// pass characteristics of the atomic selector back up to the selector
-			if (asel->error && !sel->error)
-				sel->error = asel->error;
-			if (combin == ',')
-				sel = 0;
-			a1 = s;
-		}
-
-// This is the last piece, pretend like combin is a comma
-		combin = ',';
-		a2 = s;
-		if (a2 == a1) {	// empty piece
-			if (!sel) {
-				if (!d->selectors)
-					d->error = CSS_ERROR_NOSEL;
-				continue;
-			}
-			sel->error = CSS_ERROR_SEL0;
-			continue;
-		}
-		t = allocMem(a2 - a1 + 1);
-		memcpy(t, a1, a2 - a1);
-		t[a2 - a1] = 0;
-		asel = allocZeroMem(sizeof(struct asel));
-		asel->part = t;
-		asel->combin = combin;
-		if (!sel) {
-			sel = allocZeroMem(sizeof(struct sel));
-			if (!d->selectors)
-				d->selectors = sel;
-			else
-				sel2->next = sel;
-		}
-		asel2 = sel->chain;
-		sel->chain = asel;
-		asel->next = asel2;
-		cssAtomic(asel);
-		if (asel->error && !sel->error)
-			sel->error = asel->error;
+		cssParseLeft(d);
 	}
 
 // pull before and after up from atomic selector to selector
@@ -848,7 +1016,7 @@ copy:		++s;
 		for (sel = d->selectors; sel; sel = sel->next) {
 			if (!sel->error) {
 // as good a time as any to compute specificity
-				sel->spec = specificity(sel);
+				sel->spec = specificity(sel, d->underat);
 				across = false;
 				continue;
 			}
@@ -938,7 +1106,7 @@ lastrule:
 			a = allocMem(t - r1 + 1);
 			memcpy(a, r1, t - r1);
 			a[t - r1] = 0;
-			cssAttributeCrunch(a);
+			camelCase(a);
 			rule->atname = a;
 			++t;
 			while (isspace(*t))
@@ -1001,7 +1169,150 @@ lastrule:
 	return d1;
 }
 
-// determine the tag, and build the chain of modifiers
+static void cssParseLeft(struct desc *d)
+{
+	char *s = d->lhs;
+	char *a1 = s;		// start of the atomic selector
+	char *a2, *t;
+	char c, last_c = 0, combin;
+	struct sel *sel = 0;	// the selector being built
+	struct sel *sel2;
+	struct asel *asel, *asel2;
+	int n;
+
+	if (!s[0]) {
+		d->error = CSS_ERROR_NOSEL;
+		return;
+	}
+
+	while ((c = *s)) {
+		if (c == '"' || c == '\'' || c == '[') {
+			n = closeString(s + 1, c);
+			if (n < 0)	// should never happen
+				break;
+			s += n + 1;
+			last_c = 0;
+			continue;
+		}
+// Ambiguous, ~ is combinator or part of [foo~=bar].
+// Simplistic check here for ~=
+// I don't really need it any more, now that I treat [ stuf ] as a string.
+// But I still need the next one.
+// Ambiguous, + is combinator or part of "nth_child(n+3)
+// Simplistic check here, next selector should not begin with a digit
+		if ((c == '~' && s[1] == '=') || (c == '+' && isdigit(s[1]))) {
+			last_c = c;
+			++s;
+			continue;
+		}
+
+		if (last_c == '\\') {
+			last_c = 0;
+			++s;
+			continue;
+		}
+// :not( code, rather like closing a string.
+		if (!strncmp(s, ":not(", 5)) {
+			n = closeString(s + 5, '(');
+			if (n < 0)	// should never happen
+				break;
+			s += n + 5;
+			last_c = 0;
+			continue;
+		}
+
+		combin = 0;	// look for combinator
+		a2 = s;
+		while (strchr(", \t\n\r>~+", c)) {
+			if (isspace(c)) {
+				if (!combin)
+					combin = ' ';
+			} else {
+				if (combin && combin != ' ')
+					break;
+				combin = c;
+				last_c = c;
+			}
+			c = *++s;
+		}
+		if (!combin) {
+			last_c = c;
+			++s;
+			continue;
+		}
+// it's a combinator or separator
+		last_c = c;
+		if (a2 == a1) {	// empty piece
+// I'll allow it if it's just an extra comma
+			if (combin == ',' && !sel) {
+				a1 = s;
+				continue;
+			}
+			sel->error = CSS_ERROR_SEL0;
+			break;
+		}
+
+		t = allocMem(a2 - a1 + 1);
+		memcpy(t, a1, a2 - a1);
+		t[a2 - a1] = 0;
+		asel = allocZeroMem(sizeof(struct asel));
+		asel->part = t;
+		asel->combin = combin;
+		if (!sel) {
+			sel = allocZeroMem(sizeof(struct sel));
+			if (!d->selectors)
+				d->selectors = sel2 = sel;
+			else
+				sel2->next = sel, sel2 = sel;
+		}
+// I'm reversing the order of the atomic selectors here
+		asel2 = sel->chain;
+		sel->chain = asel;
+		asel->next = asel2;
+
+		cssAtomic(asel);
+// pass characteristics of the atomic selector back up to the selector
+		if (asel->error && !sel->error)
+			sel->error = asel->error;
+		if (combin == ',')
+			sel = 0;
+		a1 = s;
+	}
+
+// This is the last piece, pretend like combin is a comma
+	combin = ',';
+	a2 = s;
+	if (a2 == a1) {		// empty piece
+		if (!sel) {
+			if (!d->selectors)
+				d->error = CSS_ERROR_NOSEL;
+			return;
+		}
+		sel->error = CSS_ERROR_SEL0;
+		return;
+	}
+	t = allocMem(a2 - a1 + 1);
+	memcpy(t, a1, a2 - a1);
+	t[a2 - a1] = 0;
+	asel = allocZeroMem(sizeof(struct asel));
+	asel->part = t;
+	asel->combin = combin;
+	if (!sel) {
+		sel = allocZeroMem(sizeof(struct sel));
+		if (!d->selectors)
+			d->selectors = sel;
+		else
+			sel2->next = sel;
+	}
+	asel2 = sel->chain;
+	sel->chain = asel;
+	asel->next = asel2;
+	cssAtomic(asel);
+	if (asel->error && !sel->error)
+		sel->error = asel->error;
+}
+
+// determine the tag and build the chain of modifiers
 static void cssAtomic(struct asel *a)
 {
 	char c, last_c;
@@ -1034,11 +1345,8 @@ static void cssAtomic(struct asel *a)
 		return;
 	m1 = s;
 	++s;
-// special code for :not(:whatever)
-	if (!strncmp(s, "not(", 4) && s[4])
-		s += 5;
-
 	last_c = 0;
+
 	while ((c = *s)) {
 		if (c == '"' || c == '\'') {
 			last_c = 0;
@@ -1048,18 +1356,25 @@ static void cssAtomic(struct asel *a)
 			s += n + 1;
 			continue;
 		}
+
+		if (s == m1 + 1 && !strncmp(m1, ":not(", 5)) {
+			last_c = 0;
+			n = closeString(s + 4, '(');
+			if (n < 0)	// should never happen
+				break;
+			s += n + 4;
+			continue;
+		}
 // I assume \ is an escape, though this could fail  foo\\:
 		if (!strchr(".[#:", c) || last_c == '\\') {
 			++s;
 			last_c = c;
 			continue;
 		}
+
 		cssModify(a, m1, s);
 		m1 = s;
 		++s;
-// special code for :not(:whatever)
-		if (!strncmp(s, "not(", 4) && s[4])
-			s += 5;
 	}
 
 // last modifier
@@ -1097,14 +1412,41 @@ static void cssModify(struct asel *a, const char *m1, const char *m2)
 	} else
 		a->modifiers = mod;
 
-// :not(foo) is jquery css syntax, whatever that is,
-// and can have any kind of complex selector inside,
-// but I'll do at least the simplest case, :not(modifier)
-	if (!strncmp(t, ":not(", 5) && t[n - 1] == ')') {
+// Handle not() first because it's weird.
+	if (!strncmp(t, ":not(", 5)) {
+		struct desc d0;	// dummy descriptor for not()
+		if (t[n - 1] != ')') {
+			a->error = CSS_ERROR_UNSUP;
+			return;
+		}
 		t[--n] = 0;
 		strmove(t, t + 5);
-		n -= 5;
+		trim(t);
+		memset(&d0, 0, sizeof(d0));
+		d0.lhs = t;
+		cssParseLeft(&d0);
+		if (!d0.selectors)	//   :not()
+			return;
 		mod->negate = true;
+		if (d0.selectors->next) {
+// more than one chain beneath
+			struct sel *sel, *sel2;
+			sel = d0.selectors;
+			while (sel) {
+				chainFree(sel->chain);
+				sel2 = sel->next;
+				free(sel);
+				sel = sel2;
+			}
+			a->error = CSS_ERROR_MANYNOT;
+			return;
+		}
+// one chain, that's good
+		mod->notchain = d0.selectors->chain;
+		free(d0.selectors);
+		a->error = d0.error;	// pass the error upward
+// should we check for before after hover?
+		return;
 	}
 // See if the modifier makes sense
 	h = t[0];
@@ -1206,18 +1548,10 @@ static void cssModify(struct asel *a, const char *m1, const char *m2)
 			if (isupper(c))
 				*w = tolower(c);
 		}
-		if (!*w)
-			break;
-// [foo=] isn't well defined. I'm going to call it [foo]
-		if (w[1])
-			break;
-		*w-- = 0;
-		if (strchr("|~^$*", *w))
-			*w = 0;
 		break;
 
 	default:
-// we could get here via :not(crap)
+// not sure how we would get here
 		a->error = CSS_ERROR_UNSUP;
 		return;
 	}			// switch
@@ -1244,17 +1578,23 @@ void cssDocLoad(jsobjtype thisobj, char *start, bool pageload)
 {
 	struct ebFrame *save_cf = cf;
 	struct cssmaster *cm;
+	bool recompile = false;
 	frameFromWindow(thisobj);
 	cm = cf->cssmaster;
-	if (!cm)
+	if (!cm) {
 		cf->cssmaster = cm = allocZeroMem(sizeof(struct cssmaster));
+		readShortCache(cm);
+	}
 // This could be run again and again, if the style nodes change.
 	if (cm->descriptors) {
 		debugPrint(3,
 			   "free and recompile css descriptors due to dom changes");
 		cssPiecesFree(cm->descriptors);
+		recompile = true;
 	}
 	cm->descriptors = cssPieces(start);
+	if (recompile)
+		debugPrint(3, "css complete");
 	if (!cm->descriptors)
 		goto done;
 	if (debugCSS) {
@@ -1283,12 +1623,32 @@ done:
 	set_js_globals();
 }
 
+static void chainFree(struct asel *asel)
+{
+	struct asel *asel2;
+	struct mod *mod, *mod2;
+	while (asel) {
+		mod = asel->modifiers;
+		while (mod) {
+			mod2 = mod->next;
+			nzFree(mod->part);
+			if (mod->notchain)
+				chainFree(mod->notchain);
+			free(mod);
+			mod = mod2;
+		}
+		nzFree(asel->part);
+		nzFree(asel->tag);
+		asel2 = asel->next;
+		free(asel);
+		asel = asel2;
+	}
+}
+
 static void cssPiecesFree(struct desc *d)
 {
 	struct desc *d2;
 	struct sel *sel, *sel2;
-	struct asel *asel, *asel2;
-	struct mod *mod, *mod2;
 	struct rule *r, *r2;
 	if (!d)
 		return;		// nothing to do
@@ -1296,21 +1656,7 @@ static void cssPiecesFree(struct desc *d)
 	while (d) {
 		sel = d->selectors;
 		while (sel) {
-			asel = sel->chain;
-			while (asel) {
-				mod = asel->modifiers;
-				while (mod) {
-					mod2 = mod->next;
-					nzFree(mod->part);
-					free(mod);
-					mod = mod2;
-				}
-				nzFree(asel->part);
-				nzFree(asel->tag);
-				asel2 = asel->next;
-				free(asel);
-				asel = asel2;
-			}
+			chainFree(sel->chain);
 			sel2 = sel->next;
 			free(sel);
 			sel = sel2;
@@ -1348,21 +1694,51 @@ void cssFree(struct ebFrame *f)
 }
 
 // for debugging
+static FILE *cssfile;
+static void chainPrint(struct asel *asel)
+{
+	const struct mod *mod;
+	while (asel) {
+		char *tag = asel->tag;
+		if (!tag)
+			tag = (asel->modifiers ? emptyString : "*");
+		if (asel->combin != ',')
+			fprintf(cssfile, "%c",
+				(asel->combin == ' ' ? '^' : asel->combin));
+		fprintf(cssfile, "%s", tag);
+		for (mod = asel->modifiers; mod; mod = mod->next) {
+			if (mod->negate) {
+				if (mod->notchain) {
+					fprintf(cssfile, ":not(");
+					chainPrint(mod->notchain);
+					fprintf(cssfile, ")");
+				} else
+					fprintf(cssfile, ":not()");
+				continue;
+			}
+			if (!strncmp(mod->part, "[class~=", 8))
+				fprintf(cssfile, ".%s", mod->part + 8);
+			else if (!strncmp(mod->part, "[id=", 4))
+				fprintf(cssfile, "#%s", mod->part + 4);
+			else
+				fprintf(cssfile, "%s", mod->part);
+		}
+		asel = asel->next;
+	}
+}
+
 static void cssPiecesPrint(const struct desc *d)
 {
-	FILE *f;
 	const struct sel *sel;
-	const struct asel *asel;
-	const struct mod *mod;
 	const struct rule *r;
 
 	if (!debugCSS)
 		return;
-	f = fopen(cssDebugFile, "a");
-	if (!f)
+	cssfile = fopen(cssDebugFile, "a");
+	if (!cssfile)
 		return;
 	if (!d) {
-		fclose(f);
+		fclose(cssfile);
 		return;
 	}
 
@@ -1371,58 +1747,35 @@ static void cssPiecesPrint(const struct desc *d)
 			if (d->error == CSS_ERROR_DELIM) {
 				char which = d->lhs[8];
 				if (which == '0')
-					fprintf(f, "%s start %s\n",
+					fprintf(cssfile, "%s start %s\n",
 						errorMessage[d->error],
 						d->lhs + 9);
 				if (which == '1')
-					fprintf(f, "%s open %s\n",
+					fprintf(cssfile, "%s open %s\n",
 						errorMessage[d->error],
 						d->lhs + 9);
 				if (which == '2')
-					fprintf(f, "%s close\n",
+					fprintf(cssfile, "%s close\n",
 						errorMessage[d->error]);
 			} else
-				fprintf(f, "<%s>%s\n", errorMessage[d->error],
-					d->lhs);
+				fprintf(cssfile, "<%s>%s\n",
+					errorMessage[d->error], d->lhs);
 			continue;
 		}
 		for (sel = d->selectors; sel; sel = sel->next) {
 			if (sel != d->selectors)
-				fprintf(f, ",");
+				fprintf(cssfile, ",");
 			if (sel->error)
-				fprintf(f, "<%s|", errorMessage[sel->error]);
-			for (asel = sel->chain; asel; asel = asel->next) {
-				char *tag = asel->tag;
-				if (!tag)
-					tag =
-					    (asel->modifiers ? emptyString :
-					     "*");
-				if (asel->combin != ',')
-					fprintf(f, "%c",
-						(asel->combin ==
-						 ' ' ? '^' : asel->combin));
-				fprintf(f, "%s", tag);
-				for (mod = asel->modifiers; mod;
-				     mod = mod->next) {
-					if (mod->negate)
-						fprintf(f, "~");
-					if (!strncmp(mod->part, "[class~=", 8))
-						fprintf(f, ".%s",
-							mod->part + 8);
-					else if (!strncmp(mod->part, "[id=", 4))
-						fprintf(f, "#%s",
-							mod->part + 4);
-					else
-						fprintf(f, "%s", mod->part);
-				}
-			}
+				fprintf(cssfile, "<%s|",
+					errorMessage[sel->error]);
+			chainPrint(sel->chain);
 		}
-		fprintf(f, "{");
+		fprintf(cssfile, "{");
 		for (r = d->rules; r; r = r->next)
-			fprintf(f, "%s:%s;", r->atname, r->atval);
-		fprintf(f, "}\n");
+			fprintf(cssfile, "%s:%s;", r->atname, r->atval);
+		fprintf(cssfile, "}\n");
 	}
-	fclose(f);
+	fclose(cssfile);
 }
 
 /*********************************************************************
@@ -1741,6 +2094,9 @@ That chain is considered, or not considered, based on before after hover
 criteria in qsa2() and qsaMatchGroup(), so we need not test for those here.
 *********************************************************************/
 
+static bool qsaMatchChain(struct htmlTag *t, jsobjtype obj,
+			  const struct asel *a);
+
 static bool qsaMatch(struct htmlTag *t, jsobjtype obj, const struct asel *a)
 {
 	bool rc;
@@ -1767,6 +2123,20 @@ static bool qsaMatch(struct htmlTag *t, jsobjtype obj, const struct asel *a)
 		char c = p[0];
 		int i, ntype, ns;
 
+		if (!c)		// empty modifier
+			continue;
+
+		if (negate) {
+			if (mod->notchain) {
+				if (qsaMatchChain(t, obj, mod->notchain))
+					return false;
+// the notchain fails, which is what we want, so on we go.
+				continue;
+			}
+// empty not()
+			continue;
+		}
+
 		if (mod->isclass && t
 		    && (bulkmatch || (gcsmatch && a->combin == ','))) {
 			char *v = t->class;
@@ -1781,12 +2151,8 @@ static bool qsaMatch(struct htmlTag *t, jsobjtype obj, const struct asel *a)
 					continue;
 				if (q[l] && !isspace(q[l]))
 					continue;
-				if (negate)
-					return false;
 				goto next_mod;
 			}
-			if (negate)
-				goto next_mod;
 			return false;
 		}
 
@@ -1795,7 +2161,7 @@ static bool qsaMatch(struct htmlTag *t, jsobjtype obj, const struct asel *a)
 			char *v = t->id;
 			if (!v)
 				v = emptyString;
-			if (stringEqual(v, p + 4) ^ negate)
+			if (stringEqual(v, p + 4))
 				goto next_mod;
 			return false;
 		}
@@ -1825,25 +2191,32 @@ static bool qsaMatch(struct htmlTag *t, jsobjtype obj, const struct asel *a)
 			if (bulkmatch && t)
 				v = (char *)attribVal(t, p + 1);
 			else if (obj) {
-				v = get_property_string_nat(obj, p + 1);
+				if (!strncmp(p + 1, "data-", 5)) {
+					jsobjtype ds =
+					    get_property_object_nat(obj,
+								    "dataset");
+					if (ds) {
+						char *k = cloneString(p + 6);
+						camelCase(k);
+						v = get_property_string_nat(ds,
+									    k);
+						nzFree(k);
+					}
+				} else
+					v = get_property_string_nat(obj, p + 1);
 				valloc = true;
 			}
 			if (cut)
 				*cut = cutc;
-			if (!v || !*v) {
-				if (negate)
-					goto next_mod;
+			if (!v)
 				return false;
-			}
 			if (!cutc) {
 				if (valloc)
 					nzFree(v);
-				if (negate)
-					return false;
 				goto next_mod;
 			}
 			if (cutc == '=') {	// easy case
-				rc = (stringEqual(v, value) ^ negate);
+				rc = (stringEqual(v, value));
 				if (valloc)
 					nzFree(v);
 				if (rc)
@@ -1853,7 +2226,6 @@ static bool qsaMatch(struct htmlTag *t, jsobjtype obj, const struct asel *a)
 			if (cutc == '|') {
 				rc = (!strncmp(v, value, l)
 				      && (v[l] == 0 || v[l] == '-'));
-				rc ^= negate;
 				if (valloc)
 					nzFree(v);
 				if (rc)
@@ -1861,7 +2233,7 @@ static bool qsaMatch(struct htmlTag *t, jsobjtype obj, const struct asel *a)
 				return false;
 			}
 			if (cutc == '^') {
-				rc = (!strncmp(v, value, l) ^ negate);
+				rc = !strncmp(v, value, l);
 				if (valloc)
 					nzFree(v);
 				if (rc)
@@ -1874,7 +2246,6 @@ static bool qsaMatch(struct htmlTag *t, jsobjtype obj, const struct asel *a)
 				rc = false;
 				if (l1 >= l2)
 					rc = !strncmp(v + l1 - l2, value, l);
-				rc ^= negate;
 				if (valloc)
 					nzFree(v);
 				if (rc)
@@ -1882,7 +2253,7 @@ static bool qsaMatch(struct htmlTag *t, jsobjtype obj, const struct asel *a)
 				return false;
 			}
 			if (cutc == '*') {
-				rc = ((! !strstr(v, value)) ^ negate);
+				rc = (! !strstr(v, value));
 				if (valloc)
 					nzFree(v);
 				if (rc)
@@ -1899,14 +2270,10 @@ static bool qsaMatch(struct htmlTag *t, jsobjtype obj, const struct asel *a)
 					continue;
 				if (valloc)
 					nzFree(v0);
-				if (negate)
-					return false;
 				goto next_mod;
 			}
 			if (valloc)
 				nzFree(v0);
-			if (negate)
-				goto next_mod;
 			return false;
 		}
 // At this point c should be a colon.
@@ -1918,7 +2285,7 @@ static bool qsaMatch(struct htmlTag *t, jsobjtype obj, const struct asel *a)
 			continue;
 
 		if (!strncmp(p, ":lang(", 6)) {
-			if (languageSpecial(t, obj, p + 6) ^ negate)
+			if (languageSpecial(t, obj, p + 6))
 				goto next_mod;
 			return false;
 		}
@@ -1988,11 +2355,8 @@ nth_good:
 			ns = spreadElem(ns);
 			if (oftype)
 				ns = spreadType(ns);
-			if (!ns) {
-				if (negate)
-					goto next_mod;
+			if (!ns)
 				return false;
-			}
 // find myself
 			for (i = 0; i < ns; ++i)
 				if (sibs[i].myself)
@@ -2032,11 +2396,8 @@ nth_bad:
 			ns = spreadElem(ns);
 			if (strstr(p, "of-type"))
 				ns = spreadType(ns);
-			if (!ns) {
-				if (negate)
-					goto next_mod;
+			if (!ns)
 				return false;
-			}
 			if (p[1] == 'f')
 				rc = sibs[0].myself;
 			if (p[1] == 'l')
@@ -2044,7 +2405,7 @@ nth_bad:
 			if (p[1] == 'o')
 				rc = (ns == 1 && sibs[0].myself);
 			free(sibs);
-			if (rc ^ negate)
+			if (rc)
 				goto next_mod;
 			return false;
 		}
@@ -2063,11 +2424,11 @@ all the div sections just below the current node.
 *********************************************************************/
 			if (t) {
 				if (!rootobj) {
-					if ((t->action == TAGACT_HTML) ^ negate)
+					if ((t->action == TAGACT_HTML))
 						goto next_mod;
 					return false;
 				}
-				if ((t->jv == rootobj) ^ negate)
+				if ((t->jv == rootobj))
 					goto next_mod;
 				return false;
 			}
@@ -2076,11 +2437,11 @@ all the div sections just below the current node.
 				    get_property_string_nat(obj, "nodeName");
 				rc = (a && stringEqualCI(a, "document"));
 				cnzFree(a);
-				if (rc ^ negate)
+				if (rc)
 					goto next_mod;
 				return false;
 			}
-			if ((obj == rootobj) ^ negate)
+			if ((obj == rootobj))
 				goto next_mod;
 			return false;
 		}
@@ -2123,7 +2484,7 @@ all the div sections just below the current node.
 					break;
 			}
 			nzFree(sibs);
-			if (rc ^ negate)
+			if (rc)
 				goto next_mod;
 			return false;
 		}
@@ -2139,7 +2500,7 @@ all the div sections just below the current node.
 				if (p[1] == 'e')
 					rc ^= 1;
 			}
-			if (rc ^ negate)
+			if (rc)
 				goto next_mod;
 			return false;
 		}
@@ -2156,7 +2517,7 @@ all the div sections just below the current node.
 				if (p[6] == 'w')
 					rc ^= 1;
 			}
-			if (rc ^ negate)
+			if (rc)
 				goto next_mod;
 			return false;
 		}
@@ -2170,7 +2531,7 @@ all the div sections just below the current node.
 					rc = get_property_bool_nat(obj,
 								   "checked");
 			}
-			if (rc ^ negate)
+			if (rc)
 				goto next_mod;
 			return false;
 		}
@@ -2511,6 +2872,8 @@ static void build1_doclist(struct htmlTag *t)
 	}
 	doclist[doclist_n++] = t;
 	t->highspec = 0;
+	if (topmatch)		// top only
+		return;
 // can't descend into another frame
 	if (t->action == TAGACT_FRAME)
 		return;
@@ -2557,6 +2920,8 @@ static struct htmlTag **qsaInternal(const char *selstring, struct htmlTag *top)
 	}
 	build_doclist(top);
 	skiproot = ! !top;
+	if (topmatch)
+		skiproot = false;
 	a = qsa2(d0);
 	nzFree(doclist);
 	cssPiecesFree(d0);
@@ -2613,16 +2978,22 @@ jsobjtype querySelector(const char *selstring, jsobjtype topobj)
 	return node;
 }
 
-// foo-bar has to become fooBar
-void cssAttributeCrunch(char *s)
+bool querySelector0(const char *selstring, jsobjtype topobj)
 {
-	char *t, *w;
-	for (t = w = s; *t; ++t)
-		if (*t == '-' && isalpha(t[1]))
-			t[1] = toupper(t[1]);
-		else
-			*w++ = *t;
-	*w = 0;
+	struct htmlTag *top = 0, **a;
+	jsobjtype node = 0;
+	rootobj = topobj;
+	if (topobj)
+		top = tagFromJavaVar(topobj);
+	onematch = topmatch = true;
+	a = qsaInternal(selstring, top);
+	onematch = topmatch = false;
+	if (!a)
+		return false;
+	if (a[0])
+		node = a[0]->jv;
+	nzFree(a);
+	return (node ? true : false);
 }
 
 // replace each attr(foo) with the value of attribute foo
@@ -2679,7 +3050,7 @@ Set a flag if that is found.
 
 static void do_rules(jsobjtype obj, struct rule *r0, int highspec)
 {
-	struct rule *r;
+	struct rule *r, *r1;
 	char *s, *s_attr;
 	int sl;
 	jsobjtype textobj, original = obj;
@@ -2706,11 +3077,12 @@ in fact it's easier to list the tags that allow it.
 	if (matchtype) {
 		bool forbidden = true;
 		static const char *const ok2inject[] = {
-			"A", "ADDRESS", "BLOCKQUOTE", "BODY", "BUTTON",
+			"A", "ADDRESS", "Q", "BLOCKQUOTE", "BODY", "BUTTON",
 			"CAPTION", "CITE",
 			"DIV", "FOOTER", "H1", "H2", "H3", "H4", "H5", "H6",
 			"HEADER", "LABEL", "LI", "MENU",
 			"P", "SPAN", "TD", "TH", "XMP",
+			"LISTING", "STRONG", "EM", "S", "STRIKE", "I", "U", "B",
 			0
 		};
 		s = get_property_string_nat(obj, "nodeName");
@@ -2796,11 +3168,18 @@ in fact it's easier to list the tags that allow it.
 		if (has && !what)
 			continue;
 
-// only write if the specificity is higher
+// don't repeat an attribute. Hardly ever happens except for acid test 0.
+		for (r1 = r0; r1 != r; r1 = r1->next)
+			if (stringEqual(r1->atname, r->atname))
+				break;
+		if (r1 != r)
+			continue;
+
+// Don't write if the specificity is less
 		a = allocMem(strlen(r->atname) + 6);
 		sprintf(a, "%s$$scy", r->atname);
 		spec = get_property_number_nat(obj, a);
-		if (spec >= highspec) {
+		if (spec > highspec) {
 			free(a);
 			continue;
 		}
@@ -2883,6 +3262,10 @@ void cssText(jsobjtype node, const char *rulestring)
 // check arguments.
 	if (!node || !rulestring)
 		return;
+// Is this suppose to replace the existing properties, or add to them?
+// I have no idea; for now it just adds to them.
+	if (!*rulestring)
+		return;		// empty
 // Compile the selector. The string has to be allocated.
 	s = allocMem(strlen(rulestring) + 20);
 	sprintf(s, "*{%s}", rulestring);
