@@ -85,8 +85,6 @@ const char *attribVal(const struct htmlTag *t, const char *name)
 	if (j < 0)
 		return 0;
 	v = t->atvals[j];
-	if (!v || !*v)
-		return 0;
 	return v;
 }				/* attribVal */
 
@@ -450,7 +448,7 @@ void htmlInputHelper(struct htmlTag *t)
 	bool isbutton = stringEqual(t->info->name, "button");
 
 	t->itype = (isbutton ? INP_BUTTON : INP_TEXT);
-	if (s) {
+	if (s && *s) {
 		n = stringInListCI(inp_types, s);
 		if (n < 0) {
 			n = stringInListCI(inp_others, s);
@@ -514,11 +512,9 @@ char *displayOptions(const struct htmlTag *sel)
 	const struct htmlTag *t;
 	char *opt;
 	int opt_l;
-	int i;
 
 	opt = initString(&opt_l);
-	for (i = 0; i < cw->numTags; ++i) {
-		t = tagList[i];
+	for (t = cw->optlist; t; t = t->same) {
 		if (t->controller != sel)
 			continue;
 		if (!t->checked)
@@ -551,7 +547,7 @@ static void prerenderNode(struct htmlTag *t, bool opentag)
 	switch (action) {
 	case TAGACT_NOSCRIPT:
 // If javascript is enabled kill everything under noscript
-		if (allowJS && !opentag)
+		if (isJSAlive && !opentag)
 			underKill(t);
 		break;
 
@@ -660,7 +656,8 @@ static void prerenderNode(struct htmlTag *t, bool opentag)
 						t->javapost = true;
 					else if (stringEqualCI(prot, "https"))
 						t->secure = true;
-					else if (!stringEqualCI(prot, "http"))
+					else if (!stringEqualCI(prot, "http") &&
+						 !stringEqualCI(prot, "gopher"))
 						debugPrint(3,
 							   "form cannot submit using protocol %s",
 							   prot);
@@ -671,7 +668,7 @@ static void prerenderNode(struct htmlTag *t, bool opentag)
 			radioCheck = initString(&radio_l);
 		}
 		if (!opentag && currentForm) {
-			if (t->href && !t->submitted) {
+			if (t->href && t->ninp && !t->submitted) {
 				makeButton();
 				t->submitted = true;
 			}
@@ -757,10 +754,11 @@ static void prerenderNode(struct htmlTag *t, bool opentag)
 /* like the other value fields, it can't be null */
 				t->rvalue = t->value = emptyString;
 			}
-			if (whichproc == 'e') {
-				j = sideBuffer(0, t->value, -1, 0);
-				t->lic = j;
-			}
+#if 0
+// When textarea buffers were allocated at browse time, before the ib command
+			if (whichproc == 'e')
+				t->lic = sideBuffer(0, t->value, -1, 0);
+#endif
 			currentTA = 0;
 		}
 		break;
@@ -824,13 +822,21 @@ static void prerenderNode(struct htmlTag *t, bool opentag)
 	case TAGACT_FRAME:
 		if (opentag)
 			break;
-// If somebody wrote <frame><p><a></frame>, those tags should be excised,
+// If somebody wrote <frame><p>foo</frame>, those tags should be excised.
 		underKill(t);
 		cdt = newTag("document");
 		t->firstchild = cdt;
 		cdt->parent = t;
 		break;
 
+	case TAGACT_MUSIC:
+		if (opentag)
+			break;
+// If somebody wrote <audio><p>foo</audio>, those tags should be excised.
+// However <source> tags should be kept and/or expanded. Not yet implemented.
+		underKill(t);
+		break;
+
 	}			/* switch */
 }				/* prerenderNode */
 
@@ -1044,6 +1050,7 @@ Don't do any of this if the tag is itself <style>. */
 		ato = instantiate(io, "attributes", "NamedNodeMap");
 		set_property_object(ato, "owner", io);
 		set_property_object(io, "ownerDocument", cf->docobj);
+		instantiate(io, "dataset", "Object");
 
 // only anchors with href go into links[]
 		if (list && stringEqual(list, "links") &&
@@ -1082,11 +1089,9 @@ Don't do any of this if the tag is itself <style>. */
 		io = w;
 	}
 
-	if (symname)
-		set_property_string(io, "name", symname);
-
-	if (idname)
-		set_property_string(io, "id", idname);
+	set_property_string(io, "name", (symname ? symname : emptyString));
+	set_property_string(io, "id", (idname ? idname : emptyString));
+	set_property_string(io, "last$id", (idname ? idname : emptyString));
 
 	if (href && href_url)
 // This use to be instantiate_url, but with the new side effects
@@ -1191,6 +1196,8 @@ static void link_css(struct htmlTag *t)
 	const char *a;
 	const char *a1 = attribVal(t, "type");
 	const char *a2 = attribVal(t, "rel");
+	const char *altsource;
+
 	if (a1)
 		set_property_string(t->jv, "type", a1);
 	if (a2)
@@ -1200,11 +1207,15 @@ static void link_css(struct htmlTag *t)
 	if ((!a1 || !stringEqualCI(a1, "text/css")) &&
 	    (!a2 || !stringEqualCI(a2, "stylesheet")))
 		return;
+
 // Fetch the css file so we can apply its attributes.
 	a = NULL;
-	if (browseLocal && !isURL(t->href)) {
-		debugPrint(3, "css source %s", t->href);
-		if (!fileIntoMemory(t->href, &b, &blen)) {
+	altsource = fetchReplace(t->href);
+	if (!altsource)
+		altsource = t->href;
+	if (browseLocal && !isURL(altsource)) {
+		debugPrint(3, "css source %s", altsource);
+		if (!fileIntoMemory(altsource, &b, &blen)) {
 			if (debugLevel >= 1)
 				i_printf(MSG_GetLocalCSS);
 		} else {
@@ -1328,18 +1339,20 @@ Needless to say that's not good!
 		a = attribVal(t, "type");
 		if (a)
 			set_property_string(t->jv, "type", a);
-		a = attribVal(t, "src");
-		if (a) {
-			set_property_string(t->jv, "src", a);
-		} else {
-			set_property_string(t->jv, "src", "");
-		}
 		a = attribVal(t, "data");
 		if (a) {
 			set_property_string(t->jv, "data", a);
 		} else {
 			set_property_string(t->jv, "data", "");
 		}
+		a = attribVal(t, "src");
+		if (a) {
+			set_property_string(t->jv, "src", a);
+			if (down_jsbg && a[0])	// from another source, let's get it started
+				prepareScript(t);
+		} else {
+			set_property_string(t->jv, "src", "");
+		}
 		break;
 
 	case TAGACT_FORM:
@@ -1478,6 +1491,10 @@ Needless to say that's not good!
 		link_css(t);
 		break;
 
+	case TAGACT_MUSIC:
+		domLink(t, "Audio", "src", 0, cf->docobj, 0);
+		break;
+
 	default:
 // Don't know what this tag is, or it's not semantically important,
 // so just call it an html element.
@@ -1548,12 +1565,13 @@ static void pushAttributes(const struct htmlTag *t)
 // There are some exceptions, some attributes that we handle individually.
 		static const char *const exclist[] = {
 			"name", "id", "class",
-			"checked", "value", "type", "style",
+			"checked", "value", "type",
 			"href", "src", "action",
 			0
 		};
 		static const char *const dotrue[] = {
-			"multiple", "readonly", "disabled", 0
+			"required",
+			"multiple", "readonly", "disabled", "async", 0
 		};
 		static const char *const handlers[] = {
 			"onload", "onunload", "onclick", "onchange",
@@ -1561,9 +1579,31 @@ static void pushAttributes(const struct htmlTag *t)
 			0
 		};
 		const char *u;
-		if (stringInListCI(exclist, a[i]) >= 0)
+
+// Should we drop attribute name to lower case? I don't, for now.
+		u = v[i];
+		if (!u)
+			u = emptyString;
+
+// attributes on HTML tags that begin with "data-" should be available under a
+// "dataset" object in JS
+		if (strncmp(a[i], "data-", 5) == 0) {
+			jsobjtype dso = get_property_object(t->jv, "dataset");
+			if (dso) {
+// must convert to camelCase
+				char *a2 = cloneString(a[i] + 5);
+				camelCase(a2);
+				set_property_string(dso, a2, u);
+				nzFree(a2);
+				run_function_onestring(t->jv, "markAttribute",
+						       a[i]);
+			}
 			continue;
-// I surely haven't thought of everything, so check generally.
+		}
+
+		if (stringEqual(a[i], "style"))	// no clue
+			continue;
+
 // Maybe they wrote <a firstChild=foo>
 // See if the name is in the prototype, and not a handler,
 // as handlers have setters.
@@ -1573,17 +1613,16 @@ static void pushAttributes(const struct htmlTag *t)
 				   t->info->name, a[i]);
 			continue;
 		}
-// Should we drop attribute name to lower case? I don't, for now.
-		u = v[i];
-		if (!u)
-			u = emptyString;
 // There are some, like multiple or readonly, that should be set to true,
 // not the empty string.
 		if (!*u && stringInList(dotrue, a[i]) >= 0) {
 			set_property_bool(t->jv, a[i], true);
-			continue;
+		} else {
+// standard attribute here
+			if (stringInListCI(exclist, a[i]) < 0)
+				set_property_string(t->jv, a[i], u);
 		}
-		set_property_string(t->jv, a[i], u);
+		run_function_onestring(t->jv, "markAttribute", a[i]);
 	}
 }				/* pushAttributes */
 
@@ -1677,6 +1716,7 @@ const struct tagInfo availableTags[] = {
 	{"unknown0", "an html entity", TAGACT_UNKNOWN, 5, 1},
 	{"object", "an html object", TAGACT_OBJECT, 5, 3},
 	{"a", "an anchor", TAGACT_A, 0, 1},
+	{"htmlanchorelement", "an anchor element", TAGACT_A, 0, 1},
 	{"input", "an input item", TAGACT_INPUT, 0, 4},
 	{"element", "an input element", TAGACT_INPUT, 0, 4},
 	{"title", "the title", TAGACT_TITLE, 0, 0},
@@ -1709,13 +1749,14 @@ const struct tagInfo availableTags[] = {
 	{"map", "a map of images", TAGACT_NOP, 5, 0},
 	{"blockquote", "a quoted paragraph", TAGACT_NOP, 10, 1},
 	{"document", "a document", TAGACT_DOC, 5, 1},
+	{"fragment", "a document fragment", TAGACT_FRAG, 5, 1},
 	{"comment", "a comment", TAGACT_COMMENT, 0, 2},
-	{"h1", "a level 1 header", TAGACT_NOP, 10, 1},
-	{"h2", "a level 2 header", TAGACT_NOP, 10, 1},
-	{"h3", "a level 3 header", TAGACT_NOP, 10, 1},
-	{"h4", "a level 4 header", TAGACT_NOP, 10, 1},
-	{"h5", "a level 5 header", TAGACT_NOP, 10, 1},
-	{"h6", "a level 6 header", TAGACT_NOP, 10, 1},
+	{"h1", "a level 1 header", TAGACT_H, 10, 1},
+	{"h2", "a level 2 header", TAGACT_H, 10, 1},
+	{"h3", "a level 3 header", TAGACT_H, 10, 1},
+	{"h4", "a level 4 header", TAGACT_H, 10, 1},
+	{"h5", "a level 5 header", TAGACT_H, 10, 1},
+	{"h6", "a level 6 header", TAGACT_H, 10, 1},
 	{"dt", "a term", TAGACT_DT, 2, 4},
 	{"dd", "a definition", TAGACT_DD, 1, 4},
 	{"li", "a list item", TAGACT_LI, 1, 5},
@@ -1807,16 +1848,14 @@ static void freeTag(struct htmlTag *t)
 void freeTags(struct ebWindow *w)
 {
 	int i, n;
-	struct htmlTag *t;
-	struct htmlTag **e;
+	struct htmlTag *t, **e;
 
 /* if not browsing ... */
 	if (!(e = w->tags))
 		return;
 
 /* drop empty textarea buffers created by this session */
-	for (i = 0; i < w->numTags; ++i, ++e) {
-		t = *e;
+	for (t = w->inputlist; t; t = t->same) {
 		if (t->action != TAGACT_INPUT)
 			continue;
 		if (t->itype != INP_TA)
@@ -1826,7 +1865,6 @@ void freeTags(struct ebWindow *w)
 		freeEmptySideBuffer(n);
 	}			/* loop over tags */
 
-	e = w->tags;
 	for (i = 0; i < w->numTags; ++i, ++e) {
 		t = *e;
 		disconnectTagObject(t);
@@ -1836,11 +1874,12 @@ void freeTags(struct ebWindow *w)
 	free(w->tags);
 	w->tags = 0;
 	w->numTags = w->allocTags = w->deadTags = 0;
+	w->inputlist = w->scriptlist = w->optlist = 0;
 }				/* freeTags */
 
 struct htmlTag *newTag(const char *name)
 {
-	struct htmlTag *t;
+	struct htmlTag *t, *t1, *t2 = 0;
 	const struct tagInfo *ti;
 
 	for (ti = availableTags; ti->name[0]; ++ti)
@@ -1860,6 +1899,34 @@ struct htmlTag *newTag(const char *name)
 	t->seqno = cw->numTags;
 	t->nodeName = cloneString(name);
 	pushTag(t);
+	if (t->action == TAGACT_SCRIPT) {
+		for (t1 = cw->scriptlist; t1; t1 = t1->same)
+			if (!t1->slash)
+				t2 = t1;
+		if (t2)
+			t2->same = t;
+		else
+			cw->scriptlist = t;
+	}
+	if (t->action == TAGACT_INPUT || t->action == TAGACT_SELECT ||
+	    t->action == TAGACT_TA) {
+		for (t1 = cw->inputlist; t1; t1 = t1->same)
+			if (!t1->slash)
+				t2 = t1;
+		if (t2)
+			t2->same = t;
+		else
+			cw->inputlist = t;
+	}
+	if (t->action == TAGACT_OPTION) {
+		for (t1 = cw->optlist; t1; t1 = t1->same)
+			if (!t1->slash)
+				t2 = t1;
+		if (t2)
+			t2->same = t;
+		else
+			cw->optlist = t;
+	}
 	return t;
 }				/* newTag */
 
@@ -1879,15 +1946,16 @@ static struct htmlTag *treeAttach;
 static int tree_pos;
 static bool treeDisable;
 static void intoTree(struct htmlTag *parent);
+static const int tdb = 5;	// tree debug level
 
 void htmlNodesIntoTree(int start, struct htmlTag *attach)
 {
 	treeAttach = attach;
 	tree_pos = start;
 	treeDisable = false;
-	debugPrint(4, "@@tree of nodes");
+	debugPrint(tdb, "@@tree of nodes");
 	intoTree(0);
-	debugPrint(4, "}\n@@end tree");
+	debugPrint(tdb, "}\n@@end tree");
 }				/* htmlNodesIntoTree */
 
 /* Convert a list of html nodes, properly nested open close, into a tree.
@@ -1901,9 +1969,9 @@ static void intoTree(struct htmlTag *parent)
 	int action;
 
 	if (!parent)
-		debugPrint(4, "root {");
+		debugPrint(tdb, "root {");
 	else
-		debugPrint(4, "%s %d {", parent->info->name, parent->seqno);
+		debugPrint(tdb, "%s %d {", parent->info->name, parent->seqno);
 
 	while (tree_pos < cw->numTags) {
 		t = tagList[tree_pos++];
@@ -1914,12 +1982,12 @@ static void intoTree(struct htmlTag *parent)
 				if (t->dead)
 					++cw->deadTags;
 			}
-			debugPrint(4, "}");
+			debugPrint(tdb, "}");
 			return;
 		}
 
 		if (treeDisable) {
-			debugPrint(4, "node skip %s", t->info->name);
+			debugPrint(tdb, "node skip %s", t->info->name);
 			t->dead = true;
 			++cw->deadTags;
 			intoTree(t);
@@ -1933,7 +2001,7 @@ static void intoTree(struct htmlTag *parent)
  * to the children below. */
 			action = t->action;
 			if (action == TAGACT_HEAD) {
-				debugPrint(4, "node skip %s", t->info->name);
+				debugPrint(tdb, "node skip %s", t->info->name);
 				t->dead = true;
 				++cw->deadTags;
 				treeDisable = true;
@@ -1942,7 +2010,7 @@ static void intoTree(struct htmlTag *parent)
 				continue;
 			}
 			if (action == TAGACT_HTML || action == TAGACT_BODY) {
-				debugPrint(4, "node pass %s", t->info->name);
+				debugPrint(tdb, "node pass %s", t->info->name);
 				t->dead = true;
 				++cw->deadTags;
 				intoTree(t);
@@ -1956,8 +2024,8 @@ static void intoTree(struct htmlTag *parent)
 				const char *w = "root";
 				if (treeAttach)
 					w = treeAttach->info->name;
-				debugPrint(4, "node up %s to %s", t->info->name,
-					   w);
+				debugPrint(tdb, "node up %s to %s",
+					   t->info->name, w);
 				t->parent = treeAttach;
 				if (treeAttach) {
 					struct htmlTag *c =
@@ -2009,6 +2077,8 @@ checkattributes:
 			t->disabled = true;
 		if (stringInListCI(t->attributes, "multiple") >= 0)
 			t->multiple = true;
+		if (stringInListCI(t->attributes, "async") >= 0)
+			t->async = true;
 		if ((j = stringInListCI(t->attributes, "name")) >= 0) {
 /* temporarily, make another copy; some day we'll just point to the value */
 			v = t->atvals[j];
@@ -2195,7 +2265,7 @@ static void processStyles(jsobjtype so, const char *stylestring)
 			trimWhite(sv);
 // the property name has to be nonempty
 			if (*s) {
-				cssAttributeCrunch(s);
+				camelCase(s);
 				set_property_string(so, s, sv);
 // Should we set a specification level here, perhaps high,
 // so the css sheets don't overwrite it?
