@@ -46,6 +46,7 @@
 #else
 #include <unistd.h>
 #endif
+#include <pthread.h>
 
 #ifndef O_BINARY
 #define O_BINARY 0
@@ -87,6 +88,16 @@ extern jsobjtype docobj;	// document object
 extern const char *jsSourceFile; // sourcefile providing the javascript
 extern int jsLineno; // line number
 
+/*********************************************************************
+duktape is not threadsafe, so says the programmer's guide.
+So javascript runs in the foreground as we push buttons etc,
+or, there is at most one thread running javascript in the background,
+if we choose to implement that.
+Here is the js background thread. It is 0 if no thread is running.
+*********************************************************************/
+
+extern pthread_t jsbt;
+
 enum ej_proptype {
 	EJ_PROP_NONE,
 	EJ_PROP_STRING,
@@ -191,10 +202,12 @@ struct i_get {
 	const char *url;
 	char *urlcopy;
 	int urlcopy_l;
+	const char *custom_h; // custom http headers passed to curl
 	char *cfn; // changed filename
 	const char *thisfile;
 	char *referrer;
 	CURL *h;
+	int tsn; // thread sequence number
 // State of download to disk, see http.c for state values.
 	int down_state;
 	int down_fd;	/* downloading file descriptor */
@@ -203,6 +216,7 @@ struct i_get {
 	const char *down_file2;	/* without download directory */
 	int down_length;
 	bool down_ok;
+	uchar down_force;
 	bool uriEncoded;
 	bool foreground;
 	bool pg_ok; // watch for plugins
@@ -221,7 +235,7 @@ struct i_get {
  * or case sensitive, so keep our own sanitized copy. */
 	char content[60];
 	char *charset;	/* extra content info such as charset */
-int hcl;			/* http content length */
+long long hcl;			/* http content length */
 	char *cdfn;		/* http content disposition file name */
 	time_t modtime;	/* http modification time */
 	char *etag;		/* the etag in the header */
@@ -240,14 +254,6 @@ struct MACCOUNT {		/* pop3 account */
 extern struct MACCOUNT accounts[];	/* all the email accounts */
 extern int maxAccount;		/* how many email accounts specified */
 
-struct PXENT {			/* proxy entry */
-	char *prot;		/* null means any */
-	char *domain;		/* null means any */
-	char *proxy;		/* null means direct */
-};
-extern struct PXENT proxyEntries[];
-extern int maxproxy;
-
 struct MIMETYPE {
 	char *type, *desc;
 	char *suffix, *prot, *program;
@@ -275,20 +281,24 @@ extern CURLSH *global_share_handle;
 extern int debugLevel;		/* 0 to 9 */
 extern bool debugClone, debugEvent, debugThrow, debugCSS;
 extern bool demin; // deminimize javascript
-extern char *uvw; // undefined variable watch
+extern bool uvw; // trace points
 extern bool gotimers; // run javascript timers
+extern int rr_interval; // rerender the screen after this many seconds
 extern FILE *debugFile;
 extern char *debugFileName;
 extern char *sslCerts;		/* ssl certificates to validate the secure server */
 extern int verifyCertificates;	/* is a certificate required for the ssl connection? */
-extern int displayLength;	/* when printing a line */
+extern int displayLength;	// when printing a line
+extern int formatLineLength;	// when formatting html
+bool formatOverflow;
 extern int webTimeout, mailTimeout;
 extern uchar browseLocal;
 extern bool sqlPresent;		/* Was edbrowse compiled with SQL built in? */
 extern bool curlActive; // is curl running?
 extern bool ismc;		/* Is the program running as a mail client? */
 extern bool isimap;		/* Is the program running as an imap client? */
-extern bool down_bg;		/* download in background */
+extern bool down_bg;		// download in background
+extern bool down_jsbg;		// download javascript in background
 extern char whichproc; // which edbrowse-xx process
 extern char showProgress; // feedback as a file is downloaded
 extern char eb_language[];		/* edbrowse language, determined by $LANG */
@@ -301,6 +311,7 @@ extern bool passMail;		/* pass mail across the filters */
 extern bool errorExit;		/* exit on any error, for scripting purposes */
 extern bool isInteractive;
 extern volatile bool intFlag;	/* set this when interrupt signal is caught */
+extern time_t intStart;
 extern bool binaryDetect;
 extern bool inputReadLine;
 extern bool curlAuthNegotiate;  /* try curl negotiate (SPNEGO) auth */
@@ -324,13 +335,14 @@ extern bool caseInsensitive, searchStringsAll, searchWrap;
 extern bool allowRedirection;	/* from http code 301, or http refresh */
 extern bool sendReferrer;	/* in the http header */
 extern bool allowJS;		/* javascript on */
-extern bool allowXHR;		/* xhr on */
+extern bool blockJS; // javascript is blocked
 extern bool htmlGenerated;
 extern bool ftpActive;
 extern bool helpMessagesOn;	/* no need to type h */
 extern bool pluginsOn;		/* plugins are active */
 extern bool showHiddenFiles;	/* during directory scan */
 extern bool showHover; // messages that appear when you hover
+extern bool doColors;
 extern int context;		/* which session (buffer) are we in? */
 extern pst linePending;
 extern char *changeFileName;
@@ -343,8 +355,8 @@ extern char *recycleBin;	/* holds deleted files */
 extern char *configFile, *sigFile, *sigFileEnd;
 extern char *cookieFile;	/* persistent cookies */
 extern char *cacheDir;	/* directory for a persistent cache of http pages */
-extern int cacheSize; /* in megabytes */
-extern int cacheCount; /* number of cache files */
+extern int cacheSize; // in megabytes
+extern int cacheCount; // number of cache files
 
 struct listHead {
 	void *next;
@@ -378,7 +390,7 @@ struct ebFrame {
 	struct ebWindow *owner;
 	struct htmlTag *frametag;
 	char *fileName;		/* name of file or url */
-	char *firstURL;		/* before http redirection */
+	char *firstURL;		// before http redirection
 	char *hbase; /* base for href references */
 	bool render1; // rendered via protocol or urlmatch
 	bool render2; // rendered via suffix
@@ -397,6 +409,7 @@ struct ebFrame {
 	jsobjtype docobj;	/* window.document */
 	const struct MIMETYPE *mt;
 	void *cssmaster;
+	short jtmin;
 };
 extern struct ebFrame *cf;	/* current frame */
 
@@ -425,6 +438,7 @@ struct ebWindow {
 	char *referrer; // another web page that brought this one to life
 	char *baseDirName;	/* when scanning a directory */
 	char *htmltitle, *htmldesc, *htmlkey;	/* title, description, keywords */
+	char *saveURL;		// for the fu command
 	char *mailInfo;
 	char lhs[MAXRE], rhs[MAXRE];	/* remembered substitution strings */
 	struct lineMap *map, *r_map;
@@ -441,6 +455,7 @@ struct ebWindow {
  * and used thereafter for hyperlinks, fill-out forms, etc. */
 	struct htmlTag **tags;
 	int numTags, allocTags, deadTags;
+	struct htmlTag *scriptlist, *inputlist, *optlist;
 	bool mustrender:1;
 	bool sank:1; /* jSyncup has been run */
 	bool lhs_yes:1;
@@ -455,6 +470,7 @@ struct ebWindow {
 	bool utf32Mode:1;
 	bool bigMode:1; // big-endian
 	bool iso8859Mode:1;
+	bool dosMode:1; // \r\n
 	bool browseMode:1;	/* browsing html */
 	bool changeMode:1;	/* something has changed in this file */
 	bool quitMode:1;	/* you can quit this buffer any time */
@@ -493,7 +509,7 @@ extern struct ebSession *cs;	/* current session */
 extern int maxSession;
 
 /* The information on an html tag */
-#define MAXTAGNAME 12
+#define MAXTAGNAME 20
 struct tagInfo {
 	const char name[MAXTAGNAME];
 	const char *desc;
@@ -519,6 +535,7 @@ struct htmlTag {
 	struct htmlTag *parent, *firstchild, *sibling;
 /* connect <foo> and </foo> */
 	struct htmlTag *balance;
+	struct htmlTag *same; // same action
 	struct ebFrame *f0; /* frame that owns this tag */
 	struct ebFrame *f1; /* subordinate frame if this is a <frame> tag */
 	jsobjtype jv;		/* corresponding javascript variable */
@@ -535,13 +552,19 @@ struct htmlTag {
 	const char **atvals;
 /* the form that owns this input tag */
 	struct htmlTag *controller;
-	uchar step; /* prerender, decorate, runscript */
+	pthread_t loadthread;
+	long hcode;
+	bool loadsuccess;
+	uchar step; // prerender, decorate, load script, runscript
 	bool slash:1;		/* as in </A> */
 	bool textin:1; /* <a> some text </a> */
 	bool deleted:1; /* deleted from the current buffer */
 	bool dead:1; // removed by garbage collection
 	bool contracted:1; /* frame is contracted */
 	bool multiple:1;
+	bool async:1; // asynchronous script
+	bool intimer:1; // asynchronous script in timer
+	bool inxhr:1; // script is really an xhr
 	bool rdonly:1;
 	bool disabled:1;
 	bool clickable:1;	/* but not an input field */
@@ -563,9 +586,16 @@ struct htmlTag {
 	bool doorway:1; /* doorway to javascript */
 	bool visited:1;
 	bool masked:1;
+	bool iscolor:1;
 	char subsup;		/* span turned into sup or sub */
 	uchar itype;		/* input type = */
 	uchar itype_minor;
+#define DIS_INVISIBLE 1
+#define DIS_HOVER 2
+#define DIS_COLOR 3
+#define DIS_TRANSPARENT 4
+#define DIS_HOVERCOLOR 5
+	uchar disval; // displayable value for the node
 	int ninp;		/* number of nonhidden inputs */
 	char *name, *id, *class, *nodeName, *value, *href;
 	const char *rvalue; /* for reset */
@@ -579,8 +609,8 @@ enum {
 	TAGACT_HTML, TAGACT_A, TAGACT_INPUT, TAGACT_TITLE, TAGACT_TA,
 	TAGACT_BUTTON, TAGACT_SELECT, TAGACT_OPTION, TAGACT_LABEL,
 	TAGACT_NOP, TAGACT_JS, TAGACT_H, TAGACT_SUB, TAGACT_SUP, TAGACT_OVB,
-	TAGACT_OL, TAGACT_UL, TAGACT_DL,
-	TAGACT_TEXT, TAGACT_BODY, TAGACT_HEAD, TAGACT_DOC, TAGACT_COMMENT,
+	TAGACT_OL, TAGACT_UL, TAGACT_DL, TAGACT_TEXT,
+	TAGACT_BODY, TAGACT_HEAD, TAGACT_DOC, TAGACT_FRAG, TAGACT_COMMENT,
 	TAGACT_MUSIC, TAGACT_IMAGE, TAGACT_BR, TAGACT_IBR, TAGACT_P,
 	TAGACT_BASE, TAGACT_META, TAGACT_LINK, TAGACT_PRE,
 	TAGACT_TBODY, TAGACT_THEAD, TAGACT_TFOOT,
@@ -623,6 +653,14 @@ extern nodeFunction traverse_callback;
 #ifdef DOSLIKE
 /* windows mkdir takes only one argument */
 #define mkdir(a,b) _mkdir(a)
+// give the above we probably don't need rwx mode but here we go
+#define MODE_rw 0600
+#define MODE_rwx 0700
+#define MODE_private 0600
+#else
+#define MODE_rw 0666
+#define MODE_rwx 0777
+#define MODE_private 0600
 #endif
 
 /* function prototypes */
