@@ -189,6 +189,26 @@ void jsRunScript(jsobjtype obj, const char *str, const char *filename,
 	nzFree(s);
 }				/* jsRunScript */
 
+void jsRunData(jsobjtype obj, const char *filename, int lineno)
+{
+// this never runs from the j process.
+	if (whichproc != 'e') {
+		debugPrint(1, "jsRunData run from the js process");
+		return;
+	}
+	if (!allowJS || !cf->winobj || !obj)
+		return;
+	debugPrint(5, "> script:");
+	jsSourceFile = filename;
+	jsLineno = lineno;
+	set_js_globals();
+	whichproc = 'j';
+	run_data_nat(obj);
+	whichproc = 'e';
+	jsSourceFile = NULL;
+	debugPrint(5, "< ok");
+}
+
 /* does the member exist in the object or its prototype? */
 bool has_property(jsobjtype obj, const char *name)
 {
@@ -673,7 +693,27 @@ int run_function_onearg(jsobjtype obj, const char *name, jsobjtype a)
 	whichproc = 'e';
 	debugPrint(5, "< ok");
 	return rc;
-}				/* run_function_onearg */
+}
+
+void run_function_onestring(jsobjtype obj, const char *name, const char *s)
+{
+	if (!allowJS || !cf->winobj)
+		return;
+	if (!obj) {
+		debugPrint(3, "run_function_onestring(0, %s", name);
+		return;
+	}
+	if (whichproc == 'j') {
+		run_function_onestring_nat(obj, name, s);
+		return;
+	}
+	debugPrint(5, "> function %s", name);
+	set_js_globals();
+	whichproc = 'j';	// this line is totally important!
+	run_function_onestring_nat(obj, name, s);
+	whichproc = 'e';
+	debugPrint(5, "< ok");
+}
 
 /*********************************************************************
 Everything beyond this point is, perhaps, part of a DOM support layer
@@ -690,6 +730,14 @@ void set_basehref(const char *h)
 	if (!h)
 		h = emptyString;
 	set_property_string(cf->winobj, "eb$base", h);
+// This is special code for snapshot simulations.
+// If the file jslocal is present, push base over to window.location etc,
+// as though you were running that page.
+	if (!access("jslocal", 4) && h[0]) {
+		run_function_bool(cf->winobj, "eb$base$snapshot");
+		nzFree(cf->fileName);
+		cf->fileName = cloneString(h);
+	}
 }				/* set_basehref */
 
 #ifdef DOSLIKE			// port of uname(p), and struct utsname
@@ -792,9 +840,12 @@ void setupJavaDom(void)
 	set_property_string(hist, "current", cf->fileName);
 
 	set_property_string(d, "referrer", cw->referrer);
-	instantiate_url(d, "URL", cf->fileName);
+	set_property_string(d, "URL", cf->fileName);
 	set_property_string(d, "location", cf->fileName);
 	set_property_string(w, "location", cf->fileName);
+	jsRunScript(w,
+		    "window.location.replace = document.location.replace = function(s) { this.href = s; };Object.defineProperty(window.location,'replace',{enumerable:false});Object.defineProperty(document.location,'replace',{enumerable:false});",
+		    "locreplace", 1);
 	set_property_string(d, "domain", getHostURL(cf->fileName));
 	if (debugClone)
 		set_property_bool(cf->winobj, "cloneDebug", true);
@@ -802,8 +853,6 @@ void setupJavaDom(void)
 		set_property_bool(cf->winobj, "eventDebug", true);
 	if (debugThrow)
 		set_property_bool(cf->winobj, "throwDebug", true);
-	if (uvw)
-		set_property_string(cf->winobj, "$uv$watch", uvw);
 }				/* setupJavaDom */
 
 /* Get the url from a url object, special wrapper.
@@ -858,44 +907,46 @@ so this tells you the options underneath have changed.
 
 static void rebuildSelector(struct htmlTag *sel, jsobjtype oa, int len2)
 {
-	int i1, i2, len1;
+	int i2 = 0;
 	bool check2;
 	char *s;
 	const char *selname;
 	bool changed = false;
-	struct htmlTag *t;
+	struct htmlTag *t, *t0 = 0;
 	jsobjtype oo;		/* option object */
 
-	len1 = cw->numTags;
-	i1 = i2 = 0;
 	selname = sel->name;
 	if (!selname)
 		selname = "?";
-	debugPrint(4, "testing selector %s %d %d", selname, len1, len2);
-
+	debugPrint(4, "testing selector %s %d", selname, len2);
 	sel->lic = (sel->multiple ? 0 : -1);
+	t = cw->optlist;
 
-	while (i1 < len1 && i2 < len2) {
+	while (t && i2 < len2) {
+		t0 = t;
 /* there is more to both lists */
-		t = tagList[i1++];
-		if (t->action != TAGACT_OPTION)
-			continue;
-		if (t->controller != sel)
-			continue;
-		if (t->dead)
+		if (t->controller != sel) {
+			t = t->same;
 			continue;
+		}
 
 /* find the corresponding option object */
 		if ((oo = get_array_element_object(oa, i2)) == NULL) {
 /* Wow this shouldn't happen. */
 /* Guess I'll just pretend the array stops here. */
 			len2 = i2;
-			--i1;
 			break;
 		}
 
 		if (t->jv != oo) {
 			debugPrint(5, "oo switch");
+/*********************************************************************
+Ok, we freed up the old options, and garbage collection
+could well kill the tags that went with these options,
+i.e. the tags we're looking at now.
+I'm bringing the tags back to life.
+*********************************************************************/
+			t->dead = false;
 			disconnectTagObject(t);
 			connectTagObject(t, oo);
 		}
@@ -925,23 +976,27 @@ static void rebuildSelector(struct htmlTag *sel, jsobjtype oa, int len2)
 			t->value = s;
 		} else
 			nzFree(s);
+		t = t->same;
 	}
 
 /* one list or the other or both has run to the end */
 	if (i2 == len2) {
-		for (; i1 < len1; ++i1) {
-			t = tagList[i1];
-			if (t->action != TAGACT_OPTION)
-				continue;
-			if (t->controller != sel)
+		for (; t; t = t->same) {
+			if (t->controller != sel) {
+				t0 = t;
 				continue;
+			}
 /* option is gone in js, disconnect this option tag from its select */
 			disconnectTagObject(t);
 			t->controller = 0;
 			t->action = TAGACT_NOP;
+			if (t0)
+				t0->same = t->same;
+			else
+				cw->optlist = t->same;
 			changed = true;
 		}
-	} else if (i1 == len1) {
+	} else if (!t) {
 		for (; i2 < len2; ++i2) {
 			if ((oo = get_array_element_object(oa, i2)) == NULL)
 				break;
