@@ -70,6 +70,7 @@ void createJavaContext(void) ;
 void freeJavaContext(struct ebFrame *f) ;
 char *jsRunScriptResult(jsobjtype obj, const char *str, const char *filename, int lineno) ;
 void jsRunScript(jsobjtype obj, const char *str, const char *filename, int lineno) ;
+void jsRunData(jsobjtype obj, const char *filename, int lineno);
 enum ej_proptype typeof_property(jsobjtype obj, const char *name) ;
 bool has_property(jsobjtype obj, const char *name) ;
 #define handlerPresent(obj, name) (typeof_property(obj, name) == EJ_PROP_FUNCTION)
@@ -100,6 +101,7 @@ jsobjtype create_event(jsobjtype parent, const char *evname);
 void unlink_event(jsobjtype parent);
 bool run_event_bool(jsobjtype obj, const char *pname, const char *evname, jsobjtype evobj);
 int run_function_onearg(jsobjtype obj, const char *name, jsobjtype o);
+void run_function_onestring(jsobjtype parent, const char *name, const char *s);
 void set_basehref(const char *b);
 
 /* sourcefile=fetchmail.c */
@@ -119,6 +121,7 @@ void breakLineSetup(void);
 char *htmlReformat(char *buf);
 void extractEmailAddresses(char *line);
 void cutDuplicateEmails(char *tolist, char *cclist, const char *reply);
+bool isEmailAddress(const char *s);
 int byteOrderMark(const uchar *buf, int buflen);
 bool looksBinary(const unsigned char *buf, int buflen);
 void looks_8859_utf8(const uchar *buf, int buflen, bool * iso_p, bool * utf8_p);
@@ -129,30 +132,37 @@ char *uni2utf8(unsigned int unichar);
 void utfLow(const char *inbuf, int inbuflen, char **outbuf_p, int *outbuflen_p, int bom);
 char *force_utf8( char *buf, int buflen);
 char *base64Encode(const char *inbuf, int inlen, bool lines);
-void iuReformat(const char *inbuf, int inbuflen, char **outbuf_p, int *outbuflen_p) ;
+uchar base64Bits(char c);
+int base64Decode(char *start, char **end);
+void iuReformat(const char *inbuf, int inbuflen, char **outbuf_p, int *outbuflen_p);
 bool parseDataURI(const char *uri, char **mediatype, char **data, int *data_l);
+uchar fromHex(char d, char e) ;
+char *closeColor(const char *s);
 
 /* sourcefile=html.c */
-bool tagHandler(int seqno, const char *name) ;
-void jSideEffects(void) ;
-void jSyncup(bool fromtimer) ;
+bool tagHandler(int seqno, const char *name);
+void jSideEffects(void);
+void jSyncup(bool fromtimer);
 void jClearSync(void);
 void htmlMetaHelper(struct htmlTag *t);
-extern void runScriptsPending(void);
+void prepareScript(struct htmlTag *t);
+void runScriptsPending(void);
 void preFormatCheck(int tagno, bool * pretag, bool * slash) ;
 char *htmlParse(char *buf, int remote) ;
 bool htmlTest(void) ;
 void infShow(int tagno, const char *search) ;
-bool infReplace(int tagno, const char *newtext, bool notify) ;
-bool infPush(int tagno, char **post_string) ;
+void itext(void);
+bool infReplace(int tagno, const char *newtext, bool notify);
+bool infPush(int tagno, char **post_string);
 struct htmlTag *tagFromJavaVar(jsobjtype v);
-void javaSubmitsForm(jsobjtype v, bool reset) ;
-bool bubble_event(const struct htmlTag *t, const char *name) ;
-void runningError(int msg, ...) ;
+void javaSubmitsForm(jsobjtype v, bool reset);
+bool bubble_event(const struct htmlTag *t, const char *name);
+void runningError(int msg, ...);
 void rerender(bool notify);
 void delTags(int startRange, int endRange);
-extern void runOnload(void);
+void runOnload(void);
 void javaSetsTimeout(int n, const char *jsrc, jsobjtype to, bool isInterval);
+void scriptSetsTimeout( struct htmlTag *t);
 bool timerWait(int *delay_sec, int *delay_ms);
 void delTimers(struct ebFrame *f);
 void runTimer(void);
@@ -186,19 +196,18 @@ void html_from_setter( jsobjtype innerParent, const char *h);
 
 /* sourcefile=http.c */
 size_t eb_curl_callback(char *incoming, size_t size, size_t nitems, struct i_get *g) ;
-uchar base64Bits(char c);
-int base64Decode(char *start, char **end);
 time_t parseHeaderDate(const char *date) ;
 bool parseRefresh(char *ref, int *delay_p) ;
 bool shortRefreshDelay(const char *r, int d);
 bool httpConnect(struct i_get *g);
+void *httpConnectBack1(void *ptr);
+void *httpConnectBack2(void *ptr);
+void *httpConnectBack3(void *ptr);
 void ebcurl_setError(CURLcode curlret, const char *url, int action, const char *curl_error);
 void setHTTPLanguage(const char *lang);
 int prompt_and_read(int prompt, char *buffer, int buffer_length, int error_message, bool hide_echo);
 int ebcurl_debug_handler(CURL * handle, curl_infotype info_desc, char *data, size_t size, struct i_get *g);
 int bg_jobs(bool iponly);
-void addNovsHost(char *host);
-void deleteNovsHosts(void);
 CURLcode setCurlURL(CURL * h, const char *url);
 bool frameExpand(bool expand, int ln1, int ln2);
 int frameExpandLine(int ln, jsobjtype fo);
@@ -209,16 +218,19 @@ void unframe(jsobjtype fobj, jsobjtype newdoc);
 void unframe2(jsobjtype fobj);
 
 /* sourcefile=main.c */
-const char *mailRedirect(const char *to, const char *from, const char *reply, const char *subj);
-bool javaOK(const char *url);
 void eb_curl_global_init(void);
 void ebClose(int n);
-bool isSQL(const char *s);
 void setDataSource(char *v);
+bool javaOK(const char *url);
+bool mustVerifyHost(const char *url);
+const char *findProxyForURL(const char *url);
+const char *findAgentForURL(const char *url);
+const char *mailRedirect(const char *to, const char *from, const char *reply, const char *subj);
 bool runEbFunction(const char *line);
 struct DBTABLE *findTableDescriptor(const char *sn);
 struct DBTABLE *newTableDescriptor(const char *name);
 void readConfigFile(void);
+const char *fetchReplace(const char *u);
 
 /* sourcefile=plugin.c */
 const struct MIMETYPE *findMimeByURL(const char *url, uchar *sxfirst);
@@ -237,21 +249,22 @@ bool validAccount(int n);
 bool sendMailCurrent(int sm_account, bool dosig);
 
 /* sourcefile=messages.c */
-void eeCheck(void) ;
+void eeCheck(void);
 void eb_puts(const char *s);
-void selectLanguage(void) ; //@ called once at startup
+void selectLanguage(void); //@ called once at startup
 const char *i_getString(int msg); //@ the i_ functions are ok in utf8 mode
-void i_puts(int msg) ;
-void i_printf(int msg, ...) ;
-void i_printfExit(int msg, ...) ;
-void i_stringAndMessage(char **s, int *l, int messageNum) ;
-void setError(int msg, ...) ; //? use these only in the foreground!
-void showError(void) ;
-void showErrorConditional(char cmd) ;
-void showErrorAbort(void) ;
+void i_puts(int msg);
+void i_printf(int msg, ...);
+void i_printfExit(int msg, ...);
+void i_stringAndMessage(char **s, int *l, int messageNum);
+void setError(int msg, ...); //? use these only in the foreground!
+void showError(void);
+void showErrorConditional(char cmd);
+void showErrorAbort(void);
 #if 0
-void i_caseShift(unsigned char *s, char action) ;
+void i_caseShift(unsigned char *s, char action);
 #endif
+bool helpUtility(void);
 
 // sourcefile=stringfile.c
 // Everything in this file is threadsafe except those marked with @,
@@ -266,7 +279,6 @@ char *allocZeroString(size_t n) ;
 char *reallocString(void *p, size_t n) ;
 void nzFree(void *s) ;
 void cnzFree(const void *s) ;
-uchar fromHex(char d, char e) ;
 char *appendString(char *s, const char *p) ;
 char *prependString(char *s, const char *p) ;
 void skipWhite(const char **s) ;
@@ -318,11 +330,13 @@ void copyPstring(pst s, const pst t) ;
 bool fdIntoMemory(int fd, char **data, int *len) ;
 bool fileIntoMemory(const char *filename, char **data, int *len) ;
 bool memoryOutToFile(const char *filename, const char *data, int len, int msgcreate, int msgwrite) ;
+void truncate0(const char *filename, int fh);
 void caseShift(char *s, char action) ;
+void camelCase(char *s);
 char fileTypeByName(const char *name, bool showlink) ;
 char fileTypeByHandle(int fd) ;
-int fileSizeByName(const char *name) ;
-int fileSizeByHandle(int fd) ;
+off_t fileSizeByName(const char *name) ;
+off_t fileSizeByHandle(int fd) ;
 time_t fileTimeByName(const char *name) ;
 char *conciseSize(size_t n); //?
 char *conciseTime(time_t t); //?
@@ -357,6 +371,7 @@ char *htmlEscape0(const char *s, bool do_and);
 #define htmlEscape(s) htmlEscape0((s), true)
 #define htmlEscapeTextarea(s) htmlEscape0((s), false)
 bool isURL(const char *url) ;
+bool isSQL(const char *url);
 bool isBrowseableURL(const char *url) ;
 bool isDataURI(const char *u);
 const char *getProtURL(const char *url) ; //?
@@ -377,6 +392,7 @@ char *altText(const char *base) ; //?
 char *encodePostData(const char *s, const char *keep_chars) ;
 char *decodePostData(const char *data, const char *name, int seqno) ;
 void decodeMailURL(const char *url, char **addr_p, char **subj_p, char **body_p) ;
+bool patternMatchURL(const char *url, const char *pattern);
 
 /* sourcefile=jseng-duk.c */
 void connectTagObject(struct htmlTag *t, jsobjtype p);
@@ -408,14 +424,19 @@ int set_property_function_nat(jsobjtype parent, const char *name, const char *bo
 int get_arraylength_nat(jsobjtype a);
 bool run_function_bool_nat(jsobjtype obj, const char *name);
 int run_function_onearg_nat(jsobjtype obj, const char *name, jsobjtype o);
+void run_function_onestring_nat(jsobjtype parent, const char *name, const char *s);
 char *run_script_nat(const char *s);
+void run_data_nat(jsobjtype o);
+void put_data_nat(jsobjtype cx, jsobjtype o, const char *s);
 
 /* sourcefile=css.c */
+void writeShortCache(void);
+bool matchMedia(char *t);
 void cssDocLoad(jsobjtype thisobj, char *s, bool pageload);
 void cssFree(struct ebFrame *f);
 jsobjtype querySelectorAll(const char *selstring, jsobjtype topobj);
 jsobjtype querySelector(const char *selstring, jsobjtype topobj);
-void cssAttributeCrunch(char *s);
+bool querySelector0(const char *selstring, jsobjtype topobj);
 void cssApply(jsobjtype thisobj, jsobjtype node, jsobjtype destination);
 void cssText(jsobjtype node, const char *rulestring);
 
