@@ -422,9 +422,10 @@ The prefix bl means breakline.
 static char *bl_start, *bl_cursor, *bl_end;
 static bool bl_overflow;
 /* This is a virtual column number, extra spaces for tab,
- * and skipping over invisible anchors. */
+ * one space for emoji, and skipping over invisible anchors. */
 static int colno;
-static const int optimalLine = 80;	/* optimal line length */
+int formatLineLength = 80;	// for html formatting or the bl command
+bool formatOverflow;
 static const int cutLineAfter = 36;	/* cut sentence after this column */
 static const int paraLine = 120;	/* paragraph in a line */
 static int longcut, pre_cr;
@@ -567,12 +568,14 @@ static void appendSpaceChunk(const char *chunk, int len, bool premode)
 		if (strchr(")\"|", d))
 			lright = colno, idxright = l;
 		lany = colno, idxany = l;
+		if (formatOverflow) {
 /* tack a short fragment onto the previous line. */
-		if (longcut && colno <= 15 && (nlc || lperiod == colno)) {
-			bl_start[longcut] = ' ';
-			if (!nlc)
-				len = spc = 0, nlc = 1;
-		}		/* pasting small fragment onto previous line */
+			if (longcut && colno <= 15 && (nlc || lperiod == colno)) {
+				bl_start[longcut] = ' ';
+				if (!nlc)
+					len = spc = 0, nlc = 1;
+			}	/* pasting small fragment onto previous line */
+		}
 	}			/* allowing line breaks */
 	if (lspace == 3)
 		nlc = 0;
@@ -638,7 +641,11 @@ static void appendPrintableChunk(const char *chunk, int len, bool premode)
 			continue;
 		}
 		if (visible) {
-			++colno;
+// each foreign char or emoji counts as one.
+// Ignore all but the first byte of a utf8.
+			if ((char)c >= 0	// ascii
+			    || (c & 0x40) == 0x40)
+				++colno;
 			continue;
 		}
 		if (isdigitByte(c))
@@ -652,7 +659,7 @@ static void appendPrintableChunk(const char *chunk, int len, bool premode)
 	lspace = 0;
 	if (premode)
 		return;
-	if (colno <= optimalLine)
+	if (colno <= formatLineLength)
 		return;
 /* Oops, line is getting long.  Let's see where we can cut it. */
 	i = j = 0;
@@ -836,8 +843,13 @@ char *htmlReformat(char *buf)
 		/* It's a tag */
 		tagno = strtol(h + 1, (char **)&nh, 10);
 		c = *nh++;
-		if (!c || !strchr("{}<>*", c))
-			i_printfExit(MSG_BadTagCode, tagno, c);
+		if (!c || !strchr("{}<>*", c)) {
+// this should never happen!
+			i_printf(MSG_BadTagCode, tagno, c);
+			appendOneChar('@');
+			nh = h + 1;
+			continue;
+		}
 		appendPrintableChunk(h, nh - h, premode);
 		preFormatCheck(tagno, &pretag, &slash);
 		if (pretag) {
@@ -1018,6 +1030,35 @@ void cutDuplicateEmails(char *tolist, char *cclist, const char *reply)
 	}
 }				/* cutDuplicateEmails */
 
+bool isEmailAddress(const char *s)
+{
+	bool atfound = false, dotfound = false;
+	if (!s || !*s)
+		return false;
+	for (; *s; ++s) {
+		char c = *s;
+		if (c < 0)	// nonascii
+			return false;
+		if (atfound) {
+			if (!isalnum(c) && c != '.' && c != '-')
+				return false;
+			if (c == '.') {
+				if (s[1] == '.' || s[1] == 0 || s[-1] == '.'
+				    || s[-1] == '@')
+					return false;
+				dotfound = true;
+			}
+			continue;
+		}
+// I think anything is ok before the @, except space.
+		if (c <= ' ')
+			return false;
+		if (c == '@')
+			atfound = true;
+	}
+	return atfound & dotfound;
+}
+
 /* return 1 for utf16, 2 for utf32, ored with 4 for big endian */
 int byteOrderMark(const uchar * buf, int buflen)
 {
@@ -1132,10 +1173,25 @@ Don't forget to free it when you're done.
 
 /* only 8859-1 and 8859-2 so far */
 static const int iso_unicodes[2][128] = {
-	{0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b,
+	{
+/*********************************************************************
+The first 32 nonascii chars in iso8859-1 are control characters,
+and almost never used.
+Much more common are the cp1252 characters, introduced by Microsoft.
+I'm gonna go with those, and hope I'm right more often than wrong.
+*********************************************************************/
+#define CP1252 1
+#if CP1252
+	 0x20AC, 0x81, 0x201A, 0x192, 0x201E, 0x2026, 0x2020, 0x2021,
+	 0x2C6, 0x2030, 0x160, 0x2039, 0x152, 0x8d, 0x17D, 0x8f,
+	 0x90, 0x2018, 0x2019, 0x201C, 0x201D, 0x2022, 0x2013, 0x2014,
+	 0x2DC, 0x2122, 0x161, 0x203A, 0x153, 0x9d, 0x17E, 0x178,
+#else
+	 0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b,
 	 0x8c, 0x8d, 0x8e, 0x8f,
 	 0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a,
 	 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
+#endif
 	 0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa,
 	 0xab, 0xac, 0xad, 0xae, 0xaf,
 	 0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba,
@@ -1175,6 +1231,7 @@ void iso2utf(const uchar * inbuf, int inbuflen, uchar ** outbuf_p,
 	uchar *outbuf;
 	const int *isoarray = iso_unicodes[type8859 - 1];
 	int ucode;
+	char *s;
 
 	if (!inbuflen) {
 		*outbuf_p = (uchar *) emptyString;
@@ -1185,11 +1242,15 @@ void iso2utf(const uchar * inbuf, int inbuflen, uchar ** outbuf_p,
 /* count chars, so we can allocate */
 	for (i = 0; i < inbuflen; ++i) {
 		c = inbuf[i];
-		if (c >= 0x80)
-			++nacount;
+		if (c >= 0x80) {
+			ucode = isoarray[c & 0x7f];
+			s = uni2utf8(ucode);
+			nacount += strlen(s) - 1;
+		}
 	}
 
 	outbuf = allocMem(inbuflen + nacount + 1);
+
 	for (i = j = 0; i < inbuflen; ++i) {
 		c = inbuf[i];
 		if (c < 0x80) {
@@ -1197,10 +1258,11 @@ void iso2utf(const uchar * inbuf, int inbuflen, uchar ** outbuf_p,
 			continue;
 		}
 		ucode = isoarray[c & 0x7f];
-		outbuf[j++] = (ucode >> 6) | 0xc0;
-		outbuf[j++] = (ucode & 0x3f) | 0x80;
+		s = uni2utf8(ucode);
+		strcpy((char *)outbuf + j, s);
+		j += strlen(s);
 	}
-	outbuf[j] = 0;		/* just for fun */
+	outbuf[j] = 0;
 
 	*outbuf_p = outbuf;
 	*outbuflen_p = j;
@@ -1247,6 +1309,25 @@ void utf2iso(const uchar * inbuf, int inbuflen, uchar ** outbuf_p,
 			}
 		}
 
+/* Convertable into 16 bit */
+		if ((c & 0xf0) == 0xe0 &&
+		    (inbuf[i + 1] & 0xc0) == 0x80 &&
+		    (inbuf[i + 2] & 0xc0) == 0x80) {
+			ucode = c & 0xf;
+			ucode <<= 6;
+			ucode |= (inbuf[i + 1] & 0x3f);
+			ucode <<= 6;
+			ucode |= (inbuf[i + 2] & 0x3f);
+			for (k = 0; k < 128; ++k)
+				if (isoarray[k] == ucode)
+					break;
+			if (k < 128) {
+				outbuf[j++] = k | 0x80;
+				i += 2;
+				continue;
+			}
+		}
+
 /* unicodes not found in our iso class are converted into stars */
 		c <<= 1;
 		++i;
@@ -1257,7 +1338,7 @@ void utf2iso(const uchar * inbuf, int inbuflen, uchar ** outbuf_p,
 		outbuf[j++] = '*';
 		--i;
 	}
-	outbuf[j] = 0;		/* just for fun */
+	outbuf[j] = 0;
 
 	*outbuf_p = outbuf;
 	*outbuflen_p = j;
@@ -1549,7 +1630,7 @@ char *force_utf8(char *buf, int buflen)
 	return NULL;
 }
 
-static char base64_chars[] =
+static const char base64_chars[] =
     "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
 /*
@@ -1758,3 +1839,241 @@ bool parseDataURI(const char *uri, char **mediatype, char **data, int *data_l)
 
 	return true;
 }				/* parseDataURI */
+
+uchar fromHex(char d, char e)
+{
+	d |= 0x20, e |= 0x20;
+	if (d >= 'a')
+		d -= ('a' - '9' - 1);
+	if (e >= 'a')
+		e -= ('a' - '9' - 1);
+	d -= '0', e -= '0';
+	return ((((uchar) d) << 4) | (uchar) e);
+}				/* fromHex */
+
+// find the color closest to the rgb value.
+// Input string is allocated; return is either the einput string
+// or another allocated string.
+char *closeColor(const char *s)
+{
+// indent formats an array of structures really weird; not like I would.
+	const struct reserved {
+		const char *name;
+		uchar r, g, b;
+	} colorlist[] = {
+		{
+		"aliceblue", 0xf0, 0xf8, 0xff}, {
+		"antiquewhite", 0xfa, 0xeb, 0xd7}, {
+		"aqua", 0x00, 0xff, 0xff}, {
+		"aquamarine", 0x7f, 0xff, 0xd4}, {
+		"azure", 0xf0, 0xff, 0xff}, {
+		"beige", 0xf5, 0xf5, 0xdc}, {
+		"bisque", 0xff, 0xe4, 0xc4}, {
+		"black", 0x00, 0x00, 0x00}, {
+		"blanchedalmond", 0xff, 0xeb, 0xcd}, {
+		"blue", 0x00, 0x00, 0xff}, {
+		"blueviolet", 0x8a, 0x2b, 0xe2}, {
+		"brown", 0xa5, 0x2a, 0x2a}, {
+		"burlywood", 0xde, 0xb8, 0x87}, {
+		"cadetblue", 0x5f, 0x9e, 0xa0}, {
+		"chartreuse", 0x7f, 0xff, 0x00}, {
+		"chocolate", 0xd2, 0x69, 0x1e}, {
+		"coral", 0xff, 0x7f, 0x50}, {
+		"cornflowerblue", 0x64, 0x95, 0xed}, {
+		"cornsilk", 0xff, 0xf8, 0xdc}, {
+		"crimson", 0xdc, 0x14, 0x3c}, {
+		"cyan", 0x00, 0xff, 0xff}, {
+		"darkblue", 0x00, 0x00, 0x8b}, {
+		"darkcyan", 0x00, 0x8b, 0x8b}, {
+		"darkgoldenrod", 0xb8, 0x86, 0x0b}, {
+		"darkgray", 0xa9, 0xa9, 0xa9}, {
+		"darkgreen", 0x00, 0x64, 0x00}, {
+		"darkkhaki", 0xbd, 0xb7, 0x6b}, {
+		"darkmagenta", 0x8b, 0x00, 0x8b}, {
+		"darkolivegreen", 0x55, 0x6b, 0x2f}, {
+		"darkorange", 0xff, 0x8c, 0x00}, {
+		"darkorchid", 0x99, 0x32, 0xcc}, {
+		"darkred", 0x8b, 0x00, 0x00}, {
+		"darksalmon", 0xe9, 0x96, 0x7a}, {
+		"darkseagreen", 0x8f, 0xbc, 0x8f}, {
+		"darkslateblue", 0x48, 0x3d, 0x8b}, {
+		"darkslategray", 0x2f, 0x4f, 0x4f}, {
+		"darkturquoise", 0x00, 0xce, 0xd1}, {
+		"darkviolet", 0x94, 0x00, 0xd3}, {
+		"deeppink", 0xff, 0x14, 0x93}, {
+		"deepskyblue", 0x00, 0xbf, 0xff}, {
+		"dimgray", 0x69, 0x69, 0x69}, {
+		"dodgerblue", 0x1e, 0x90, 0xff}, {
+		"feldspar", 0xd1, 0x92, 0x75}, {
+		"firebrick", 0xb2, 0x22, 0x22}, {
+		"floralwhite", 0xff, 0xfa, 0xf0}, {
+		"forestgreen", 0x22, 0x8b, 0x22}, {
+		"fuchsia", 0xff, 0x00, 0xff}, {
+		"gainsboro", 0xdc, 0xdc, 0xdc}, {
+		"ghostwhite", 0xf8, 0xf8, 0xff}, {
+		"gold", 0xff, 0xd7, 0x00}, {
+		"goldenrod", 0xda, 0xa5, 0x20}, {
+		"gray", 0x80, 0x80, 0x80}, {
+		"green", 0x00, 0x80, 0x00}, {
+		"greenyellow", 0xad, 0xff, 0x2f}, {
+		"honeydew", 0xf0, 0xff, 0xf0}, {
+		"hotpink", 0xff, 0x69, 0xb4}, {
+		"indianred", 0xcd, 0x5c, 0x5c}, {
+		"indigo", 0x4b, 0x00, 0x82}, {
+		"ivory", 0xff, 0xff, 0xf0}, {
+		"khaki", 0xf0, 0xe6, 0x8c}, {
+		"lavender", 0xe6, 0xe6, 0xfa}, {
+		"lavenderblush", 0xff, 0xf0, 0xf5}, {
+		"lawngreen", 0x7c, 0xfc, 0x00}, {
+		"lemonchiffon", 0xff, 0xfa, 0xcd}, {
+		"lightblue", 0xad, 0xd8, 0xe6}, {
+		"lightcoral", 0xf0, 0x80, 0x80}, {
+		"lightcyan", 0xe0, 0xff, 0xff}, {
+		"lightgoldenrodyellow", 0xfa, 0xfa, 0xd2}, {
+		"lightgrey", 0xd3, 0xd3, 0xd3}, {
+		"lightgreen", 0x90, 0xee, 0x90}, {
+		"lightpink", 0xff, 0xb6, 0xc1}, {
+		"lightsalmon", 0xff, 0xa0, 0x7a}, {
+		"lightseagreen", 0x20, 0xb2, 0xaa}, {
+		"lightskyblue", 0x87, 0xce, 0xfa}, {
+		"lightslateblue", 0x84, 0x70, 0xff}, {
+		"lightslategray", 0x77, 0x88, 0x99}, {
+		"lightsteelblue", 0xb0, 0xc4, 0xde}, {
+		"lightyellow", 0xff, 0xff, 0xe0}, {
+		"lime", 0x00, 0xff, 0x00}, {
+		"limegreen", 0x32, 0xcd, 0x32}, {
+		"linen", 0xfa, 0xf0, 0xe6}, {
+		"magenta", 0xff, 0x00, 0xff}, {
+		"maroon", 0x80, 0x00, 0x00}, {
+		"mediumaquamarine", 0x66, 0xcd, 0xaa}, {
+		"mediumblue", 0x00, 0x00, 0xcd}, {
+		"mediumorchid", 0xba, 0x55, 0xd3}, {
+		"mediumpurple", 0x93, 0x70, 0xd8}, {
+		"mediumseagreen", 0x3c, 0xb3, 0x71}, {
+		"mediumslateblue", 0x7b, 0x68, 0xee}, {
+		"mediumspringgreen", 0x00, 0xfa, 0x9a}, {
+		"mediumturquoise", 0x48, 0xd1, 0xcc}, {
+		"mediumvioletred", 0xc7, 0x15, 0x85}, {
+		"midnightblue", 0x19, 0x19, 0x70}, {
+		"mintcream", 0xf5, 0xff, 0xfa}, {
+		"mistyrose", 0xff, 0xe4, 0xe1}, {
+		"moccasin", 0xff, 0xe4, 0xb5}, {
+		"navajowhite", 0xff, 0xde, 0xad}, {
+		"navy", 0x00, 0x00, 0x80}, {
+		"oldlace", 0xfd, 0xf5, 0xe6}, {
+		"olive", 0x80, 0x80, 0x00}, {
+		"olivedrab", 0x6b, 0x8e, 0x23}, {
+		"orange", 0xff, 0xa5, 0x00}, {
+		"orangered", 0xff, 0x45, 0x00}, {
+		"orchid", 0xda, 0x70, 0xd6}, {
+		"palegoldenrod", 0xee, 0xe8, 0xaa}, {
+		"palegreen", 0x98, 0xfb, 0x98}, {
+		"paleturquoise", 0xaf, 0xee, 0xee}, {
+		"palevioletred", 0xd8, 0x70, 0x93}, {
+		"papayawhip", 0xff, 0xef, 0xd5}, {
+		"peachpuff", 0xff, 0xda, 0xb9}, {
+		"peru", 0xcd, 0x85, 0x3f}, {
+		"pink", 0xff, 0xc0, 0xcb}, {
+		"plum", 0xdd, 0xa0, 0xdd}, {
+		"powderblue", 0xb0, 0xe0, 0xe6}, {
+		"purple", 0x80, 0x00, 0x80}, {
+		"red", 0xff, 0x00, 0x00}, {
+		"rosybrown", 0xbc, 0x8f, 0x8f}, {
+		"royalblue", 0x41, 0x69, 0xe1}, {
+		"saddlebrown", 0x8b, 0x45, 0x13}, {
+		"salmon", 0xfa, 0x80, 0x72}, {
+		"sandybrown", 0xf4, 0xa4, 0x60}, {
+		"seagreen", 0x2e, 0x8b, 0x57}, {
+		"seashell", 0xff, 0xf5, 0xee}, {
+		"sienna", 0xa0, 0x52, 0x2d}, {
+		"silver", 0xc0, 0xc0, 0xc0}, {
+		"skyblue", 0x87, 0xce, 0xeb}, {
+		"slateblue", 0x6a, 0x5a, 0xcd}, {
+		"slategray", 0x70, 0x80, 0x90}, {
+		"snow", 0xff, 0xfa, 0xfa}, {
+		"springgreen", 0x00, 0xff, 0x7f}, {
+		"steelblue", 0x46, 0x82, 0xb4}, {
+		"tan", 0xd2, 0xb4, 0x8c}, {
+		"teal", 0x00, 0x80, 0x80}, {
+		"thistle", 0xd8, 0xbf, 0xd8}, {
+		"tomato", 0xff, 0x63, 0x47}, {
+		"turquoise", 0x40, 0xe0, 0xd0}, {
+		"violet", 0xee, 0x82, 0xee}, {
+		"violetred", 0xd0, 0x20, 0x90}, {
+		"wheat", 0xf5, 0xde, 0xb3}, {
+		"white", 0xff, 0xff, 0xff}, {
+		"whitesmoke", 0xf5, 0xf5, 0xf5}, {
+		"yellow", 0xff, 0xff, 0x00}, {
+		"yellowgreen", 0x9a, 0xcd, 0x32}, {
+		0}
+	};
+	const struct reserved *c, *best_c;
+	int best_val;
+	int r1, g1, b1;
+	const char *t;
+
+	if (!strncmp(s, "rgb(", 4)) {
+		t = s + 4;
+		if (!isdigit(*t))
+			goto fail;
+		r1 = strtol(t, (char **)&t, 10);
+		if (*t == ',')
+			++t;
+		while (*t == ' ')
+			++t;
+		if (!isdigit(*t))
+			goto fail;
+		g1 = strtol(t, (char **)&t, 10);
+		if (*t == ',')
+			++t;
+		while (*t == ' ')
+			++t;
+		if (!isdigit(*t))
+			goto fail;
+		b1 = strtol(t, (char **)&t, 10);
+		if (*t == ',')
+			++t;
+		while (*t == ' ')
+			++t;
+		if (*t != ')')
+			goto fail;
+	} else if (*s == '#' && isxdigit(s[1])) {
+		if (!isxdigit(s[2]) || !isxdigit(s[3]))
+			goto fail;
+		if (isxdigit(s[4]) && isxdigit(s[5]) && isxdigit(s[6])) {
+			r1 = fromHex(s[1], s[2]);
+			g1 = fromHex(s[3], s[4]);
+			b1 = fromHex(s[5], s[6]);
+		} else {
+// #xyz is short for #xxyyzz
+			r1 = fromHex(s[1], s[1]);
+			g1 = fromHex(s[2], s[2]);
+			b1 = fromHex(s[3], s[3]);
+		}
+	} else {
+// not an rgb format we recognize; should be just a word.
+		for (t = s; *t; ++t)
+			if (!isalpha(*t))
+				goto fail;
+		return (char *)s;
+	}
+
+	if (r1 < 0 || g1 < 0 || b1 < 0)
+		goto fail;
+	if (r1 > 255 || g1 > 255 || b1 > 255)
+		goto fail;
+
+// closest by rms; just check them all; kind of inefficient.
+	best_val = 255 * 255 * 3 + 1;
+	for (c = colorlist; c->name; ++c) {
+		int rms = (r1 - (int)c->r) * (r1 - (int)c->r) +
+		    (g1 - (int)c->g) * (g1 - (int)c->g) +
+		    (b1 - (int)c->b) * (b1 - (int)c->b);
+		if (rms < best_val)
+			best_val = rms, best_c = c;
+	}
+	return cloneString(best_c->name);
+
+fail:
+	return 0;
+}
