@@ -126,6 +126,7 @@ void html2nodes(const char *htmltext, bool startpage)
 	tidyOptSetBool(tdoc, TidyDropEmptyElems, no);
 	tidyOptSetBool(tdoc, TidyDropEmptyParas, no);
 	tidyOptSetBool(tdoc, TidyLiteralAttribs, yes);
+	tidyOptSetBool(tdoc, TidyStyleTags, no);
 
 	tidySetCharEncoding(tdoc, (cons_utf8 ? "utf8" : "latin1"));
 
@@ -298,6 +299,11 @@ static void convertNode(TidyNode node, int level, bool opentag)
 		tidyBufClear(&tnv);
 		tidyNodeGetValue(tdoc, node, &tnv);
 		if (tnv.size) {
+// check here for &#0; or &#2; both of those are bad!
+			for (i = 0; i < tnv.size; ++i)
+				if (tnv.bp[i] == 0
+				    || tnv.bp[i] == InternalCodeChar)
+					tnv.bp[i] = InternalCodeCharAlternate;
 			t->textval = cloneString((char *)tnv.bp);
 			tidyBufFree(&tnv);
 		}
