@@ -8,7 +8,6 @@
 #ifdef _MSC_VER
 #include <fcntl.h>
 #else
-#include <sys/wait.h>
 #include <signal.h>
 #endif
 #include <time.h>
@@ -18,13 +17,14 @@ int serverDataLen;
 CURL *global_http_handle;
 CURLSH *global_share_handle;
 bool pluginsOn = true;
-bool down_bg;			/* download in background */
+bool down_bg;			// download in background
+bool down_jsbg = true;		// download js in background
 char showProgress = 'd';	// dots
 static char *httpLanguage;	/* outgoing */
 
 struct BG_JOB {
 	struct BG_JOB *next, *prev;
-	int pid, state;
+	int state;
 	size_t fsize;		// file size
 	int file2;		// offset into filename
 	char file[4];
@@ -33,7 +33,6 @@ static struct listHead down_jobs = {
 	&down_jobs, &down_jobs
 };
 
-static void background_download(struct i_get *g);
 static void setup_download(struct i_get *g);
 static CURL *http_curl_init(struct i_get *g);
 static size_t curl_header_callback(char *header_line, size_t size, size_t nmemb,
@@ -42,7 +41,6 @@ static bool ftpConnect(struct i_get *g, char *creds_buf);
 static bool gopherConnect(struct i_get *g);
 static bool read_credentials(char *buffer);
 static const char *message_for_response_code(int code);
-static const char *findProxyForURL(const char *url);
 
 /* string is allocated. Quotes are removed. No other processing is done.
  * You may need to decode %xx bytes or such. */
@@ -131,10 +129,10 @@ static void scan_http_headers(struct i_get *g, bool fromCallback)
 	}
 
 	if (!g->hcl && (v = find_http_header(g, "content-length"))) {
-		g->hcl = atoi(v);
+		sscanf(v, "%lld", &g->hcl);
 		nzFree(v);
 		if (g->hcl)
-			debugPrint(4, "content length %d", g->hcl);
+			debugPrint(4, "content length %lld", g->hcl);
 	}
 
 	if (!g->etag && (v = find_http_header(g, "etag"))) {
@@ -229,6 +227,10 @@ static void i_get_free(struct i_get *g, bool nodata)
 // should not be necessary, but just to be safe:
 	g->headers = g->urlcopy = g->cdfn = g->etag = g->newloc = 0;
 	g->down_file = 0;
+	if (g->down_fd > 0) {
+		close(g->down_fd);
+		g->down_fd = 0;
+	}
 }
 
 /* actually run the curl request, http or ftp or whatever */
@@ -248,10 +250,10 @@ static CURLcode fetch_internet(struct i_get *g)
  * -1 user aborted the download
  * 0 standard in-memory download
  * 1 download but stop and ask user if he wants to download to disk
-* 2 disk download foreground
-* 3 disk download background parent
-* 4 disk download background child
-* 5 disk download background prefork
+* 2 disk download in foreground
+* 3 disk download parent thread
+* 4 disk download child thread
+* 5 disk download before the thread is spawned
  * 6 mime type says this should be a stream */
 size_t
 eb_curl_callback(char *incoming, size_t size, size_t nitems, struct i_get * g)
@@ -306,13 +308,12 @@ eb_curl_callback(char *incoming, size_t size, size_t nitems, struct i_get * g)
 // has to be the foreground http thread, so ok to call setErro,
 // which is not threadsafe.
 			setError(MSG_NoWrite2, g->down_file);
-			return -1;
+		} else {
+			i_printf(MSG_NoWrite2, g->down_file);
+			printf(", ");
+			i_puts(MSG_DownAbort);
 		}
-		i_printf(MSG_NoWrite2, g->down_file);
-		printf(", ");
-		i_puts(MSG_DownAbort);
-// return -1;
-		exit(1);
+		return -1;
 	}
 
 showdots:
@@ -322,7 +323,9 @@ showdots:
 	else
 		g->length += num_bytes;
 	dots2 = g->length / CHUNKSIZE;
-	if (showProgress != 'q' && dots1 < dots2) {
+// showing dots in parallel background download threads
+// gets jumbled and doesn't mean anything.
+	if (showProgress != 'q' && dots1 < dots2 && !g->down_force) {
 		if (showProgress == 'd') {
 			for (; dots1 < dots2; ++dots1)
 				putchar('.');
@@ -330,7 +333,7 @@ showdots:
 		}
 		if (showProgress == 'c' && g->hcl)
 			printf("%d/%d\n", dots2,
-			       (g->hcl + CHUNKSIZE - 1) / CHUNKSIZE);
+			       (int)((g->hcl + CHUNKSIZE - 1) / CHUNKSIZE));
 	}
 	return num_bytes;
 }
@@ -350,14 +353,15 @@ showdots:
  * All of the progress arguments to the function are unused. */
 
 static int
-curl_progress(void *unused, double dl_total, double dl_now,
+curl_progress(void *data_p, double dl_total, double dl_now,
 	      double ul_total, double ul_now)
 {
+	struct i_get *g = data_p;
 	int ret = 0;
-// ^c will interrupt an http or ftp download.
-// Perhaps that is hanging, and blocking edbrowse.
-	if (intFlag) {
-		i_puts(MSG_Interrupted);
+// ^c will interrupt an http or ftp download but not a background download
+	if (intFlag && g->down_force != 1) {
+		if (g->down_force == 0)
+			i_puts(MSG_Interrupted);
 		ret = 1;
 	}
 	return ret;
@@ -632,6 +636,8 @@ bool parseRefresh(char *ref, int *delay_p)
 	if (memEqualCI(u, "url=", 4)) {
 		char qc;
 		u += 4;
+		while (isspace(*u))
+			++u;
 		qc = *u;
 		if (qc == '"' || qc == '\'')
 			++u;
@@ -681,6 +687,7 @@ static void urlSanitize(struct i_get *g, const char *post)
 		g->urlcopy = percentURL(url, post);
 		g->urlcopy_l = strlen(g->urlcopy);
 	} else {
+		char *frag;
 		if (post)
 			g->urlcopy_l = post - url;
 		else
@@ -688,6 +695,10 @@ static void urlSanitize(struct i_get *g, const char *post)
 		g->urlcopy = allocMem(g->urlcopy_l + 2);
 		strncpy(g->urlcopy, url, g->urlcopy_l);
 		g->urlcopy[g->urlcopy_l] = 0;
+// percentURL strips off the hash, so we need to here.
+		frag = findHash(g->urlcopy);
+		if (frag)
+			*frag = 0;
 	}
 
 // get rid of : in http://this.that.com:/path, curl can't handle it.
@@ -798,8 +809,10 @@ mimestream:
 	}
 
 	h = http_curl_init(g);
-	if (!h)
+	if (!h) {		// should never happen
+		i_get_free(g, false);
 		return false;
+	}
 
 /* "Expect:" header causes some servers to lose.  Disable it. */
 	tmp_headers = curl_slist_append(custom_headers, "Expect:");
@@ -897,6 +910,31 @@ mimestream:
 	curlret = curl_easy_setopt(h, CURLOPT_REFERER, referrer);
 	if (curlret != CURLE_OK)
 		goto curl_fail;
+
+// look for custom headers from the calling function
+	if (g->custom_h) {
+		const char *u, *v;
+		u = g->custom_h;
+		while (*u) {
+			int d;
+			char *w;
+			v = strchr(u, '\n');
+			if (!v)
+				break;
+			d = v - u;
+			w = allocMem(d + 1);
+			memcpy(w, u, d);
+			w[d] = 0;
+			tmp_headers = curl_slist_append(custom_headers, w);
+			if (tmp_headers == NULL)
+				i_printfExit(MSG_NoMem);
+			custom_headers = tmp_headers;
+			debugPrint(4, "custom %s", w);
+			nzFree(w);
+			u = v + 1;
+		}
+	}
+
 	curlret = curl_easy_setopt(h, CURLOPT_HTTPHEADER, custom_headers);
 	if (curlret != CURLE_OK)
 		goto curl_fail;
@@ -946,6 +984,14 @@ mimestream:
 			goto mimestream;
 		}
 
+		if (head_request && g->down_force == 1) {
+			curl_easy_setopt(h, CURLOPT_NOBODY, 0l);
+			head_request = false;
+		}
+
+		if (g->down_force == 1)
+			truncate0(g->down_file, g->down_fd);
+
 perform:
 		g->is_http = g->cacheable = true;
 		curlret = fetch_internet(g);
@@ -1017,9 +1063,24 @@ they go where they go, so this doesn't come up very often.
 
 		if (g->down_state == 5) {
 /* user has directed a download of this file in the background. */
-			background_download(g);
-			if (g->down_state == 4)
-				goto perform;
+/* We spawn a thread to do this, then return, but g could go away */
+/* before the child thread has a chance to read its contents. */
+			struct i_get g0;
+			pthread_t tid;
+			nzFree(g->buffer);
+			g->buffer = NULL;
+			g->length = 0;
+			g0 = *g;	// structure copy
+			if (custom_headers)
+				curl_slist_free_all(custom_headers);
+			curl_easy_cleanup(h);
+			nzFree(postb);
+			nzFree(referrer);
+			pthread_create(&tid, NULL, httpConnectBack1,
+				       (void *)&g0);
+// I will assume the thread was created.
+// Don't call i_get_free(g); the child thread is using those strings.
+			return true;
 		}
 
 		if (g->down_state == 3 || g->down_state == -1) {
@@ -1030,22 +1091,32 @@ they go where they go, so this doesn't come up very often.
 		}
 
 		if (g->down_state == 4) {
+			bool r = true;
 			if (curlret != CURLE_OK) {
+				r = false;
 				ebcurl_setError(curlret, g->urlcopy, 1,
 						g->error);
 			} else {
-				i_printf(MSG_DownSuccess);
-				printf(": %s\n", g->down_file2);
+				curl_easy_getinfo(h, CURLINFO_RESPONSE_CODE,
+						  &(g->code));
+				if (g->code != 200) {
+					r = false;
+				} else {
+					i_printf(MSG_DownSuccess);
+					printf(": %s\n", g->down_file2);
+				}
 			}
-// We're going to exit, so don't need to do this stuff,
-// but some day this might be the end of a thread, not the end of a process.
-			i_get_free(g, true);
+			if (custom_headers)
+				curl_slist_free_all(custom_headers);
 			curl_easy_cleanup(h);
+			nzFree(postb);
 			nzFree(referrer);
-			exit(0);
+			i_get_free(g, true);
+			return r;
 		}
 
-		if (g->length >= CHUNKSIZE && showProgress == 'd')
+		if (g->length >= CHUNKSIZE && showProgress == 'd'
+		    && !g->down_force)
 			nl();	/* We printed dots, so terminate them with newline */
 
 		if (g->down_state == 2) {
@@ -1057,13 +1128,25 @@ they go where they go, so this doesn't come up very often.
 			return false;
 		}
 
-		if (curlret != CURLE_OK)
-			goto curl_fail;
+		if (curlret != CURLE_OK) {
+			if (!head_request)
+				goto curl_fail;
+			ebcurl_setError(curlret, g->urlcopy, 1, g->error);
+			debugPrint(3, "switch from head to get");
+			curl_easy_setopt(h, CURLOPT_NOBODY, 0l);
+			head_request = false;
+			goto perform;
+		}
+// get http code
 		curl_easy_getinfo(h, CURLINFO_RESPONSE_CODE, &g->code);
 		if (curlret != CURLE_OK)
 			goto curl_fail;
 
-		debugPrint(3, "http code %ld", g->code);
+		if (g->tsn)
+			debugPrint(3, "thread %d http code %ld", g->tsn,
+				   g->code);
+		else
+			debugPrint(3, "http code %ld", g->code);
 
 /* refresh header is an alternate form of redirection */
 		if (g->newloc && g->newloc_d >= 0) {
@@ -1255,6 +1338,114 @@ curl_fail:
 	return transfer_status;
 }				/* httpConnect */
 
+static int tsn;			// thread sequence number
+
+void *httpConnectBack1(void *ptr)
+{
+	struct i_get *g0 = ptr;
+	struct i_get g = *g0;	// structure copy
+	struct BG_JOB *job;
+	bool rc;
+	g.down_force = 1;
+	g.down_state = 4;
+// urlcopy will be recomputed on the next http call
+	nzFree(g.urlcopy);
+	g.urlcopy = 0;
+// Other things we should clean up?
+	g.tsn = ++tsn;
+	debugPrint(3, "bg thread %d", tsn);
+	i_puts(MSG_DownProgress);
+/* push job onto the list for tracking and display */
+	job = allocMem(sizeof(struct BG_JOB) + strlen(g.down_file));
+	job->state = 4;
+	strcpy(job->file, g.down_file);
+	job->file2 = g.down_file2 - g.down_file;
+// round file size up to the nearest chunk.
+// This will come out 0 only if the true size is 0.
+	job->fsize = ((g.hcl + (CHUNKSIZE - 1)) / CHUNKSIZE);
+	addToListBack(&down_jobs, job);
+	rc = httpConnect(&g);
+	job->state = (rc ? 0 : -1);
+	return NULL;
+}
+
+void *httpConnectBack2(void *ptr)
+{
+	struct htmlTag *t = ptr;
+	bool rc;
+	struct i_get g;
+	memset(&g, 0, sizeof(g));
+	g.thisfile = cf->fileName;
+	g.uriEncoded = true;
+	g.url = t->href;
+	g.down_force = 2;
+	g.tsn = ++tsn;
+	debugPrint(3, "jsbg thread %d", tsn);
+	rc = httpConnect(&g);
+	t->loadsuccess = rc;
+	if (!rc)
+		t->hcode = g.code;
+	else {
+// Rarely, a js file is not in utf8; convert it here, inside the thread.
+		char *b = force_utf8(g.buffer, g.length);
+		if (!b)
+			b = g.buffer;
+		else
+			nzFree(g.buffer);
+// don't know why t->value would be anything
+		nzFree(t->value);
+		t->value = b;
+	}
+	return NULL;
+}
+
+void *httpConnectBack3(void *ptr)
+{
+	struct htmlTag *t = ptr;
+	bool rc;
+	struct i_get g;
+	char *outgoing_body = 0, *outgoing_headers = 0;
+	memset(&g, 0, sizeof(g));
+	g.thisfile = cf->fileName;
+	g.uriEncoded = true;
+	g.url = t->href;
+	g.custom_h = t->innerHTML;
+	g.headers_p = &outgoing_headers;
+	g.down_force = 2;
+	g.tsn = ++tsn;
+	debugPrint(3, "xhr thread %d", tsn);
+	rc = httpConnect(&g);
+	outgoing_body = g.buffer;
+	t->loadsuccess = rc;
+	if (!rc)
+		t->hcode = g.code;
+	else {
+		char *a;
+		int l;
+// don't know why t->value would be anything
+		nzFree(t->value);
+		a = initString(&l);
+		if (outgoing_headers == 0)
+			outgoing_headers = emptyString;
+		if (outgoing_body == 0)
+			outgoing_body = emptyString;
+		stringAndNum(&a, &l, rc);
+		stringAndString(&a, &l, "\r\n\r\n");
+		stringAndNum(&a, &l, g.code);
+		stringAndString(&a, &l, "\r\n\r\n");
+		stringAndString(&a, &l, outgoing_headers);
+		stringAndString(&a, &l, outgoing_body);
+		while (l && isspace(a[l - 1]))
+			a[--l] = 0;
+		t->value = a;
+	}
+	nzFree(outgoing_headers);
+	nzFree(outgoing_body);
+	nzFree(t->innerHTML);
+	t->innerHTML = 0;
+	return NULL;
+}
+
 // copy text over to the buffer but change < to &lt; etc,
 // since this data will be browsed as if it were html.
 static void prepHtmlString(struct i_get *g, const char *q)
@@ -1295,7 +1486,7 @@ static void ftp_ls_line(struct i_get *g, char *line)
 	stringAndString(&g->buffer, &g->length, "<br>");
 
 	for (j = 0; line[j]; ++j)
-		if (!strchr("-rwxdls", line[j]))
+		if (!strchr("-rwxdlsS", line[j]))
 			break;
 
 	if (j == 10 && line[j] == ' ') {	/* long list */
@@ -1441,20 +1632,27 @@ static void gopher_ls_line(struct i_get *g, char *line)
 		stringAndChar(&g->buffer, &g->length, '\n');
 		return;
 	}
-// everything else becomes hyperlink
+// everything else becomes hyperlink apart from item type 7 which becomes form
 	if (host) {
 		char qc = '"';
 // I just assume host and path can be quoted with either " or '
 		if (strchr(host, qc)	// should never happen
 		    || strchr(pathname, qc))
 			qc = '\'';
-		stringAndString(&g->buffer, &g->length, "<a href=x");
+		if (first != '7')
+			stringAndString(&g->buffer, &g->length, "<a href=x");
+		else
+			stringAndString(&g->buffer, &g->length,
+					"<form method='get' action=x");
 		g->buffer[g->length - 1] = qc;
 
-		pathname = encodePostData(pathname, "./-_$");
 		if (!strncmp(pathname, "URL:", 4)) {
+// Full URL in path so use it unencoded
 			stringAndString(&g->buffer, &g->length, pathname + 4);
+			pathname = 0;
 		} else {
+// Just a path
+			pathname = encodePostData(pathname, "./-_$");
 			stringAndString(&g->buffer, &g->length, "gopher://");
 			stringAndString(&g->buffer, &g->length, host);
 			if (port && port != 70) {
@@ -1476,9 +1674,15 @@ static void gopher_ls_line(struct i_get *g, char *line)
 		s = 0;
 	if (s)
 		*s = 0;
+
 	prepHtmlString(g, text);
-	if (host)
-		stringAndString(&g->buffer, &g->length, "</a>");
+	if (host) {
+		if (first == '7')
+			stringAndString(&g->buffer, &g->length,
+					" <input type='text' /> <input type='submit' /></form>");
+		else
+			stringAndString(&g->buffer, &g->length, "</a>");
+	}
 	if (s) {
 		*s = '(';
 		prepHtmlString(g, s);
@@ -1555,7 +1759,9 @@ void ebcurl_setError(CURLcode curlret, const char *url, int action,
 		break;
 
 	case CURLE_PEER_FAILED_VERIFICATION:
+#if LIBCURL_VERSION_NUM < 0x073e00
 	case CURLE_SSL_CACERT:
+#endif
 		(*fn) (MSG_NoCertify, host);
 		break;
 
@@ -1638,9 +1844,6 @@ static bool ftpConnect(struct i_get *g, char *creds_buf)
 	if (stringEqual(creds_buf, ":") && memEqualCI(url, "ftp", 3))
 		strcpy(creds_buf, "anonymous:ftp@example.com");
 
-	g->down_state = 0;
-	g->down_file = g->down_file2 = NULL;
-	g->down_length = 0;
 	h = http_curl_init(g);
 	if (!h)
 		goto ftp_transfer_fail;
@@ -1653,28 +1856,39 @@ static bool ftpConnect(struct i_get *g, char *creds_buf)
 /* libcurl appends an implicit slash to URLs like "ftp://foo.com".
 * Be explicit, so that edbrowse knows that we have a directory. */
 	if (!strchr(g->urlcopy + protLength, '/'))
-		strcpy(g->urlcopy + g->urlcopy_l, "/");
+		strcpy(g->urlcopy + g->urlcopy_l++, "/");
 
 	curlret = setCurlURL(h, g->urlcopy);
 	if (curlret != CURLE_OK)
 		goto ftp_transfer_fail;
 
-	has_slash = g->urlcopy[g->urlcopy_l] == '/';
+	has_slash = g->urlcopy[g->urlcopy_l - 1] == '/';
 /* don't download a directory listing, we want to see that */
 /* Fetching a directory will fail in the special case of scp. */
-	g->down_state = (has_slash ? 0 : 1);
+	if (!g->down_force)
+		g->down_state = (has_slash ? 0 : 1);
+	g->down_length = 0;
 	g->down_msg = MSG_FTPDownload;
 	if (is_scp)
 		g->down_msg = MSG_SCPDownload;
 
-perform:
 	curlret = fetch_internet(g);
 
 	if (g->down_state == 5) {
 /* user has directed a download of this file in the background. */
-		background_download(g);
-		if (g->down_state == 4)
-			goto perform;
+/* We spawn a thread to do this, then return, but g could go away */
+/* before the child thread has a chance to read its contents. */
+		struct i_get g0;
+		pthread_t tid;
+		nzFree(g->buffer);
+		g->buffer = NULL;
+		g->length = 0;
+		g0 = *g;	// structure copy
+		curl_easy_cleanup(h);
+		pthread_create(&tid, NULL, httpConnectBack1, (void *)&g0);
+// I will assume the thread was created.
+// Don't call i_get_free(g); the child thread is using those strings.
+		return true;
 	}
 
 	if (g->down_state == 3 || g->down_state == -1) {
@@ -1684,11 +1898,17 @@ perform:
 	}
 
 	if (g->down_state == 4) {
-		if (curlret != CURLE_OK)
-			ebcurl_setError(curlret, g->urlcopy, 2, g->error);
-		i_printf(MSG_DownSuccess);
-		printf(": %s\n", g->down_file2);
-		exit(0);
+		bool r = true;
+		if (curlret != CURLE_OK) {
+			r = false;
+			ebcurl_setError(curlret, g->urlcopy, 1, g->error);
+		} else {
+			i_printf(MSG_DownSuccess);
+			printf(": %s\n", g->down_file2);
+		}
+		curl_easy_cleanup(h);
+		i_get_free(g, true);
+		return r;
 	}
 
 	if (g->length >= CHUNKSIZE && showProgress == 'd')
@@ -1709,7 +1929,7 @@ perform:
 		if (has_slash | is_scp)
 			transfer_success = false;
 		else {		/* try appending a slash. */
-			strcpy(g->urlcopy + g->urlcopy_l, "/");
+			strcpy(g->urlcopy + g->urlcopy_l++, "/");
 			g->down_state = 0;
 			cnzFree(g->down_file);
 			g->down_file = 0;
@@ -1780,6 +2000,7 @@ static bool gopherConnect(struct i_get *g)
 	has_slash = g->urlcopy[strlen(g->urlcopy) - 1] == '/';
 /* don't download a directory listing, we want to see that */
 	g->down_state = (has_slash ? 0 : 1);
+	g->down_length = 0;
 	g->down_msg = MSG_GopherDownload;
 // That's the default, let the leading character override
 	s = strchr(g->urlcopy + protLength, '/');
@@ -1798,14 +2019,26 @@ static bool gopherConnect(struct i_get *g)
 			has_slash = true;
 	}
 
-perform:
+	if (g->down_force)
+		g->down_state = 4;
+
 	curlret = fetch_internet(g);
 
 	if (g->down_state == 5) {
 /* user has directed a download of this file in the background. */
-		background_download(g);
-		if (g->down_state == 4)
-			goto perform;
+/* We spawn a thread to do this, then return, but g could go away */
+/* before the child thread has a chance to read its contents. */
+		struct i_get g0;
+		pthread_t tid;
+		nzFree(g->buffer);
+		g->buffer = NULL;
+		g->length = 0;
+		g0 = *g;	// structure copy
+		curl_easy_cleanup(h);
+		pthread_create(&tid, NULL, httpConnectBack1, (void *)&g0);
+// I will assume the thread was created.
+// Don't call i_get_free(g); the child thread is using those strings.
+		return true;
 	}
 
 	if (g->down_state == 3 || g->down_state == -1) {
@@ -1815,11 +2048,17 @@ perform:
 	}
 
 	if (g->down_state == 4) {
-		if (curlret != CURLE_OK)
-			ebcurl_setError(curlret, g->urlcopy, 2, g->error);
-		i_printf(MSG_DownSuccess);
-		printf(": %s\n", g->down_file2);
-		exit(0);
+		bool r = true;
+		if (curlret != CURLE_OK) {
+			r = false;
+			ebcurl_setError(curlret, g->urlcopy, 1, g->error);
+		} else {
+			i_printf(MSG_DownSuccess);
+			printf(": %s\n", g->down_file2);
+		}
+		curl_easy_cleanup(h);
+		i_get_free(g, true);
+		return r;
 	}
 
 	if (g->length >= CHUNKSIZE && showProgress == 'd')
@@ -1940,10 +2179,10 @@ static CURL *http_curl_init(struct i_get *g)
 	curl_easy_setopt(h, CURLOPT_DEBUGDATA, g);
 	curl_easy_setopt(h, CURLOPT_NOPROGRESS, 0);
 	curl_easy_setopt(h, CURLOPT_PROGRESSFUNCTION, curl_progress);
+	curl_easy_setopt(h, CURLOPT_PROGRESSDATA, g);
 	curl_easy_setopt(h, CURLOPT_CONNECTTIMEOUT, webTimeout);
 	curl_easy_setopt(h, CURLOPT_USERAGENT, currentAgent);
 	curl_easy_setopt(h, CURLOPT_SSLVERSION, CURL_SSLVERSION_DEFAULT);
-	curl_easy_setopt(h, CURLOPT_USERAGENT, currentAgent);
 /* We're doing this manually for now.
 	curl_easy_setopt(h, CURLOPT_FOLLOWLOCATION, allowRedirection);
 */
@@ -2265,6 +2504,7 @@ static void setup_download(struct i_get *g)
 {
 	const char *filepart;
 	const char *answer;
+	char *fp2, *s;
 
 /* if not run from a terminal then just return. */
 	if (!isInteractive) {
@@ -2276,17 +2516,25 @@ static void setup_download(struct i_get *g)
 		filepart = g->cdfn;
 	else
 		filepart = getFileURL(g->urlcopy, true);
+// transliterate to get rid of /
+	fp2 = cloneString(filepart);
+	for (s = fp2; *s; ++s)
+		if (*s == '/' || *s == '\\')
+			*s = '_';
+
 top:
-	answer = getFileName(g->down_msg, filepart, false, true);
+	answer = getFileName(g->down_msg, fp2, false, true);
 /* space for a filename means read into memory */
 	if (stringEqual(answer, " ")) {
 		g->down_state = 0;	/* in memory download */
+		nzFree(fp2);
 		return;
 	}
 
 	if (stringEqual(answer, "x") || stringEqual(answer, "X")) {
 		g->down_state = -1;
 		setError(MSG_DownAbort);
+		nzFree(fp2);
 		return;
 	}
 
@@ -2295,12 +2543,15 @@ top:
 		goto top;
 	}
 
-	g->down_fd = creat(answer, 0666);
+	g->down_fd = creat(answer, MODE_rw);
 	if (g->down_fd < 0) {
 		i_printf(MSG_NoCreate2, answer);
 		nl();
 		goto top;
 	}
+
+	nzFree(fp2);
+
 // we will free down_file, but not down_file2
 	g->down_file = g->down_file2 = cloneString(answer);
 	if (downDir) {
@@ -2311,63 +2562,10 @@ top:
 				++g->down_file2;
 		}
 	}
+
 	g->down_state = (down_bg ? 5 : 2);
 }				/* setup_download */
 
-#ifdef _MSC_VER			// need fork()
-/* At this point, down_state = 5 */
-static void background_download(struct i_get *g)
-{
-	g->down_state = -1;
-/* perhaps a better error message here */
-	setError(MSG_DownAbort);
-	return;
-}
-
-int bg_jobs(bool iponly)
-{
-	return 0;
-}
-
-#else // !_MSC_VER
-
-/* At this point, down_state = 5 */
-static void background_download(struct i_get *g)
-{
-	int down_pid = fork();
-	if (down_pid < 0) {	/* should never happen */
-		g->down_state = -1;
-/* perhaps a better error message here */
-		setError(MSG_DownAbort);
-		return;
-	}
-
-	if (down_pid) {		/* parent */
-		struct BG_JOB *job;
-		close(g->down_fd);
-/* the error message here isn't really an error, but a progress message */
-		setError(MSG_DownProgress);
-		g->down_state = 3;
-
-/* push job onto the list for tracking and display */
-		job = allocMem(sizeof(struct BG_JOB) + strlen(g->down_file));
-		job->pid = down_pid;
-		job->state = 4;
-		strcpy(job->file, g->down_file);
-		job->file2 = g->down_file2 - g->down_file;
-// round file size up to the nearest chunk.
-// This will come out 0 only if the true size is 0.
-		job->fsize = ((g->hcl + (CHUNKSIZE - 1)) / CHUNKSIZE);
-		addToListBack(&down_jobs, job);
-
-		return;
-	}
-
-/* ignore interrupt, not sure about quit and hangup */
-	signal(SIGINT, SIG_IGN);
-	g->down_state = 4;
-}				/* background_download */
-
 /* show background jobs and return the number of jobs pending */
 /* if iponly is true then just show in progress */
 int bg_jobs(bool iponly)
@@ -2375,19 +2573,6 @@ int bg_jobs(bool iponly)
 	bool present = false, part;
 	int numback = 0;
 	struct BG_JOB *j;
-	int pid, status;
-
-/* gather exit status of background jobs */
-	foreach(j, down_jobs) {
-		if (j->state != 4)
-			continue;
-		pid = waitpid(j->pid, &status, WNOHANG);
-		if (!pid)
-			continue;
-		j->state = -1;
-		if (WIFEXITED(status) && WEXITSTATUS(status) == 0)
-			j->state = 0;
-	}
 
 /* three passes */
 /* in progress */
@@ -2404,7 +2589,8 @@ int bg_jobs(bool iponly)
 		printf("%s", j->file + j->file2);
 		if (j->fsize)
 			printf(" %d/%zu",
-			       (fileSizeByName(j->file) / CHUNKSIZE), j->fsize);
+			       (int)(fileSizeByName(j->file) / CHUNKSIZE),
+			       j->fsize);
 		nl();
 	}
 	if (part)
@@ -2447,70 +2633,22 @@ int bg_jobs(bool iponly)
 		i_puts(MSG_Empty);
 
 	return numback;
-}				/* bg_jobs */
-#endif // #ifndef _MSC_VER // need fork()
-
-static char **novs_hosts;
-size_t novs_hosts_avail;
-size_t novs_hosts_max;
-
-void addNovsHost(char *host)
-{
-	if (novs_hosts_max == 0) {
-		novs_hosts_max = 32;
-		novs_hosts = allocZeroMem(novs_hosts_max * sizeof(char *));
-	} else if (novs_hosts_avail >= novs_hosts_max) {
-		novs_hosts_max *= 2;
-		novs_hosts =
-		    reallocMem(novs_hosts, novs_hosts_max * sizeof(char *));
-	}
-	novs_hosts[novs_hosts_avail++] = host;
-}				/* addNovsHost */
-
-/* Return true if the cert for this host should be verified. */
-static bool mustVerifyHost(const char *host)
-{
-	size_t this_host_len = strlen(host);
-	size_t i;
-
-	if (!verifyCertificates)
-		return false;
-
-	for (i = 0; i < novs_hosts_avail; i++) {
-		size_t l1 = strlen(novs_hosts[i]);
-		size_t l2 = this_host_len;
-		if (l1 > l2)
-			continue;
-		l2 -= l1;
-		if (!stringEqualCI(novs_hosts[i], host + l2))
-			continue;
-		if (l2 && host[l2 - 1] != '.')
-			continue;
-		return false;
-	}
-	return true;
-}				/* mustVerifyHost */
-
-void deleteNovsHosts(void)
-{
-	nzFree(novs_hosts);
-	novs_hosts = NULL;
-	novs_hosts_avail = novs_hosts_max = 0;
-}				/* deleteNovsHosts */
+}
 
 CURLcode setCurlURL(CURL * h, const char *url)
 {
-	char host[MAXHOSTLEN];
-	unsigned long verify;
+	unsigned long verify = mustVerifyHost(url);
 	const char *proxy = findProxyForURL(url);
+	const char *agent = findAgentForURL(url);
 	if (!proxy)
 		proxy = "";
 	else
-		debugPrint(3, "proxy %s", proxy);
-	if (!getProtHostURL(url, NULL, host))
-		return CURLE_URL_MALFORMAT;
-	verify = mustVerifyHost(host);
+		debugPrint(4, "proxy %s", proxy);
 	curl_easy_setopt(h, CURLOPT_PROXY, proxy);
+	if (agent) {
+		debugPrint(4, "agent %s", agent);
+		curl_easy_setopt(h, CURLOPT_USERAGENT, agent);
+	}
 	curl_easy_setopt(h, CURLOPT_SSL_VERIFYPEER, verify);
 	curl_easy_setopt(h, CURLOPT_SSL_VERIFYHOST, (verify ? 2 : 0));
 // certificate file is per handle, not global, so must be set here.
@@ -2520,73 +2658,6 @@ CURLcode setCurlURL(CURL * h, const char *url)
 	return curl_easy_setopt(h, CURLOPT_URL, url);
 }				/* setCurlURL */
 
-/*********************************************************************
-Given a protocol and a domain, find the proxy server
-to mediate your request.
-This is the C version, using entries in .ebrc.
-There is a javascript version of the same name, that we will support later.
-This is a beginning, and it can be used even when javascript is disabled.
-A return of null means DIRECT, and this is the default
-if we don't match any of the proxy entries.
-*********************************************************************/
-
-static const char *findProxyInternal(const char *prot, const char *domain)
-{
-	struct PXENT *px = proxyEntries;
-	int i;
-
-/* first match wins */
-	for (i = 0; i < maxproxy; ++i, ++px) {
-
-		if (px->prot) {
-			char *s = px->prot;
-			char *t;
-			int rc;
-			while (*s) {
-				t = strchr(s, '|');
-				if (t)
-					*t = 0;
-				rc = stringEqualCI(s, prot);
-				if (t)
-					*t = '|';
-				if (rc)
-					goto domain;
-				if (!t)
-					break;
-				s = t + 1;
-			}
-			continue;
-		}
-
-domain:
-		if (px->domain) {
-			int l1 = strlen(px->domain);
-			int l2 = strlen(domain);
-			if (l1 > l2)
-				continue;
-			l2 -= l1;
-			if (!stringEqualCI(px->domain, domain + l2))
-				continue;
-			if (l2 && domain[l2 - 1] != '.')
-				continue;
-		}
-
-		return px->proxy;
-	}
-
-	return 0;
-}				/* findProxyInternal */
-
-static const char *findProxyForURL(const char *url)
-{
-	char prot[MAXPROTLEN], host[MAXHOSTLEN];
-	if (!getProtHostURL(url, prot, host)) {
-/* this should never happen */
-		return 0;
-	}
-	return findProxyInternal(prot, host);
-}				/* findProxyForURL */
-
 /* expand a frame inline.
  * Pass a range of lines; you can expand all the frames in one go.
  * Return false if there is a problem fetching a web page,
@@ -2629,6 +2700,7 @@ int frameExpandLine(int ln, jsobjtype fo)
 	int tagno, start;
 	const char *s;
 	char *a;
+	char *jssrc = 0;
 	struct htmlTag *t;
 	struct ebFrame *save_cf, *new_cf, *last_f;
 	uchar save_local;
@@ -2665,15 +2737,23 @@ int frameExpandLine(int ln, jsobjtype fo)
 		t->href = a;
 	}
 	s = t->href;
+
+// javascript in the src, what is this for?
+	if (s && !strncmp(s, "javascript:", 11)) {
+		jssrc = (char *)s;
+		s = 0;
+	}
+
 	if (!s) {
 // No source. If this is your request then return an error.
 // But if we're dipping into the objects then it needs to expand
 // into a separate window, a separate js space, with an empty body.
-		if (!fo)
+		if (!fo && !jssrc)
 			return 2;
 // After expansion we need to be able to expand it,
 // because there's something there, well maybe.
 		t->href = cloneString("#");
+// jssrc is the old href and we are responsible for it
 	}
 
 	save_cf = cf = t->f0;
@@ -2682,7 +2762,9 @@ int frameExpandLine(int ln, jsobjtype fo)
 	last_f->next = cf = allocZeroMem(sizeof(struct ebFrame));
 	cf->owner = cw;
 	cf->frametag = t;
-	debugPrint(2, "fetch frame %s", (s ? s : "empty"));
+	debugPrint(2, "fetch frame %s",
+		   (s ? s : (jssrc ? "javascript" : "empty")));
+
 	if (s) {
 		bool rc = readFileArgv(s, (fo ? 2 : 1));
 		if (!rc) {
@@ -2696,7 +2778,7 @@ int frameExpandLine(int ln, jsobjtype fo)
 			return 3;
 		}
 
-/*********************************************************************
+       /*********************************************************************
 readFile could return success and yet serverData is null.
 This happens if httpConnect did something other than fetching data,
 like playing a stream. Does that happen, even in a frame?
@@ -2780,12 +2862,18 @@ we did that before and now it's being expanded. So bump step up to 2.
 		set_property_object(cf->winobj, "top", topobj);
 		set_property_object(cf->winobj, "frameElement", t->jv);
 		run_function_bool(cf->winobj, "eb$qs$start");
+		if (jssrc) {
+			jsRunScript(cf->winobj, jssrc, "frame.src", 1);
+		}
 		runScriptsPending();
 		runOnload();
 		runScriptsPending();
-		rebuildSelectors();
 		set_property_string(cf->docobj, "readyState", "complete");
+		run_event_bool(cf->docobj, "document", "onreadystatechange", 0);
+		runScriptsPending();
+		rebuildSelectors();
 	}
+	nzFree(jssrc);
 
 	if (cf->fileName) {
 		int j = strlen(cf->fileName);
@@ -2928,7 +3016,7 @@ bool reexpandFrame(void)
 		changeFileName = 0;
 	}
 
-/* don't print the size of what we just fetched */
+	/* don't print the size of what we just fetched */
 	fileSize = -1;
 
 	cf->hbase = cloneString(cf->fileName);
@@ -2958,8 +3046,10 @@ bool reexpandFrame(void)
 		runScriptsPending();
 		runOnload();
 		runScriptsPending();
-		rebuildSelectors();
 		set_property_string(cf->docobj, "readyState", "complete");
+		run_event_bool(cf->docobj, "document", "onreadystatechange", 0);
+		runScriptsPending();
+		rebuildSelectors();
 	}
 
 	j = strlen(cf->fileName);
