@@ -119,6 +119,7 @@ void connectTagObject(struct htmlTag *t, jsobjtype p)
 		debugPrint(1, "multiple tags connect to js pointer %p", p);
 	w->u.t = t;
 	t->jv = p;
+	set_property_number_nat(p, "eb$seqno", t->seqno);
 }
 
 void disconnectTagObject(struct htmlTag *t)
@@ -149,7 +150,7 @@ int js_main(void)
 		return 4;
 	}
 	duk_push_global_object(context0);
-	duk_push_boolean(context0, false);
+	duk_push_false(context0);
 	duk_put_prop_string(context0, -2, "compiled");
 	context0_obj = duk_get_heapptr(context0, -1);
 	duk_pop(context0);
@@ -215,6 +216,12 @@ static duk_ret_t native_mydoc(duk_context * cx)
 	return 1;
 }
 
+static duk_ret_t native_hasfocus(duk_context * cx)
+{
+	duk_push_boolean(cx, foregroundWindow);
+	return 1;
+}
+
 static duk_ret_t native_puts(duk_context * cx)
 {
 	const char *s = duk_safe_to_string(cx, -1);
@@ -224,6 +231,52 @@ static duk_ret_t native_puts(duk_context * cx)
 	return 0;
 }
 
+// write local file
+static duk_ret_t native_wlf(duk_context * cx)
+{
+	const char *s = duk_safe_to_string(cx, 0);
+	int len = strlen(s);
+	const char *filename = duk_safe_to_string(cx, 1);
+	int fh;
+	bool safe = false;
+	if (stringEqual(filename, "from") || stringEqual(filename, "jslocal"))
+		safe = true;
+	if (filename[0] == 'f') {
+		int i;
+		for (i = 1; isdigit(filename[i]); ++i) ;
+		if (i > 1 && (stringEqual(filename + i, ".js") ||
+			      stringEqual(filename + i, ".css")))
+			safe = true;
+	}
+	if (!safe)
+		return 0;
+	fh = open(filename, O_CREAT | O_TRUNC | O_WRONLY | O_TEXT, MODE_rw);
+	if (fh < 0) {
+		fprintf(stderr, "cannot create file %s\n", filename);
+		return 0;
+	}
+	if (write(fh, s, len) < len)
+		fprintf(stderr, "cannot write file %s\n", filename);
+	close(fh);
+	if (stringEqual(filename, "jslocal"))
+		writeShortCache();
+	return 0;
+}
+
+static duk_ret_t native_media(duk_context * cx)
+{
+	const char *s = duk_safe_to_string(cx, 0);
+	bool rc = false;
+	if (s && *s) {
+		char *t = cloneString(s);
+		rc = matchMedia(t);
+		nzFree(t);
+	}
+	duk_pop(cx);
+	duk_push_boolean(cx, rc);
+	return 1;
+}
+
 static duk_ret_t native_logputs(duk_context * cx)
 {
 	int minlev = duk_get_int(cx, 0);
@@ -337,7 +390,7 @@ static duk_ret_t getter_innerHTML(duk_context * cx)
 
 static duk_ret_t setter_innerHTML(duk_context * cx)
 {
-	jsobjtype thisobj;
+	jsobjtype thisobj, c1, c2;
 	char *run;
 	int run_l;
 	const char *h = duk_safe_to_string(cx, -1);
@@ -348,7 +401,7 @@ static duk_ret_t setter_innerHTML(duk_context * cx)
 	duk_push_this(cx);
 // remove the preexisting children.
 	if (duk_get_prop_string(cx, -1, "childNodes") && duk_is_array(cx, -1)) {
-		duk_set_length(cx, -1, 0);
+		c1 = duk_get_heapptr(cx, -1);
 	} else {
 // no child nodes array, don't do anything.
 // This should never happen.
@@ -356,13 +409,20 @@ static duk_ret_t setter_innerHTML(duk_context * cx)
 		debugPrint(5, "setter h 3");
 		return 0;
 	}
-	duk_pop(cx);
+// hold this away from garbage collection
+	duk_put_prop_string(cx, -2, "old$cn");
+// stack now holds html and this
+// make new childNodes array
+	duk_get_global_string(cx, "Array");
+	duk_pnew(cx, 0);
+	c2 = duk_get_heapptr(cx, -1);
+	duk_put_prop_string(cx, -2, "childNodes");
 // stack now holds html and this
 	duk_insert(cx, -2);
 	duk_put_prop_string(cx, -2, "inner$HTML");
+// stack now holds this
 
 	thisobj = duk_get_heapptr(cx, -1);
-	duk_pop(cx);
 
 // Put some tags around the html, so tidy can parse it.
 	run = initString(&run_l);
@@ -380,6 +440,21 @@ static duk_ret_t setter_innerHTML(duk_context * cx)
 	run_function_onearg_nat(context0_obj, "textarea$html$crossover",
 				thisobj);
 
+// mutation fix up from native code
+	duk_push_heapptr(cx, context0_obj);
+	duk_get_prop_string(cx, -1, "mutFixup");
+	if (duk_is_function(cx, -1)) {
+		duk_push_heapptr(cx, thisobj);
+		duk_push_false(cx);
+		duk_push_heapptr(cx, c2);
+		duk_push_heapptr(cx, c1);
+		duk_call(cx, 4);
+	}
+// stack is this mw0 retval
+	duk_pop_2(cx);
+	duk_del_prop_string(cx, -1, "old$cn");
+	duk_pop(cx);
+
 	return 0;
 }
 
@@ -787,7 +862,7 @@ static void append0(duk_context * cx, bool side)
 {
 	unsigned i, length;
 	jsobjtype child, thisobj;
-	char e[40];
+	char *e;
 	const char *thisname, *childname;
 
 /* we need one argument that is an object */
@@ -830,8 +905,9 @@ static void append0(duk_context * cx, bool side)
 /* pass this linkage information back to edbrowse, to update its dom tree */
 	thisname = embedNodeName(thisobj);
 	childname = embedNodeName(child);
-	sprintf(e, "l{a|%s,%s ", pointer2string(thisobj), thisname);
+	asprintf(&e, "l{a|%s,%s ", pointer2string(thisobj), thisname);
 	effectString(e);
+	free(e);
 	effectString(pointer2string(child));
 	effectChar(',');
 	effectString(childname);
@@ -859,7 +935,7 @@ static duk_ret_t native_insbf(duk_context * cx)
 	unsigned i, length;
 	int mark;
 	jsobjtype child, item, thisobj, h;
-	char e[40];
+	char *e;
 	const char *thisname, *childname, *itemname;
 
 /* we need two objects */
@@ -916,8 +992,9 @@ static duk_ret_t native_insbf(duk_context * cx)
 	thisname = embedNodeName(thisobj);
 	childname = embedNodeName(child);
 	itemname = embedNodeName(item);
-	sprintf(e, "l{b|%s,%s ", pointer2string(thisobj), thisname);
+	asprintf(&e, "l{b|%s,%s ", pointer2string(thisobj), thisname);
 	effectString(e);
+	free(e);
 	effectString(pointer2string(child));
 	effectChar(',');
 	effectString(childname);
@@ -938,20 +1015,20 @@ static duk_ret_t native_removeChild(duk_context * cx)
 	unsigned i, length;
 	int mark;
 	jsobjtype child, thisobj, h;
-	char e[40];
+	char *e;
 	const char *thisname, *childname;
 
 	debugPrint(5, "remove 1");
 // top of stack must be the object to remove.
 	if (!duk_is_object(cx, -1))
-		goto done;
+		goto fail;
 	child = duk_get_heapptr(cx, -1);
 	duk_push_this(cx);
 	thisobj = duk_get_heapptr(cx, -1);
 	duk_get_prop_string(cx, -1, "childNodes");
 	if (!duk_is_array(cx, -1)) {
 		duk_pop_2(cx);
-		goto done;
+		goto fail;
 	}
 	length = duk_get_length(cx, -1);
 	mark = -1;
@@ -967,7 +1044,7 @@ static duk_ret_t native_removeChild(duk_context * cx)
 
 	if (mark < 0) {
 		duk_pop_2(cx);
-		goto done;
+		goto fail;
 	}
 
 /* push the other elements down */
@@ -977,83 +1054,160 @@ static duk_ret_t native_removeChild(duk_context * cx)
 	}
 	duk_set_length(cx, -1, length - 1);
 	duk_pop_2(cx);
-	duk_del_prop_string(cx, -1, "parentNode");
+// missing parentnode must always be null
+	duk_push_null(cx);
+	duk_put_prop_string(cx, -2, "parentNode");
 
 /* pass this linkage information back to edbrowse, to update its dom tree */
 	thisname = embedNodeName(thisobj);
 	childname = embedNodeName(child);
-	sprintf(e, "l{r|%s,%s ", pointer2string(thisobj), thisname);
+	asprintf(&e, "l{r|%s,%s ", pointer2string(thisobj), thisname);
 	effectString(e);
+	free(e);
 	effectString(pointer2string(child));
 	effectChar(',');
 	effectString(childname);
 	effectString(" 0x0, ");
 	linkageNow('r', thisobj);
 
-done:
 	debugPrint(5, "remove 2");
+// mutation fix up from native code
+	duk_push_heapptr(cx, context0_obj);
+	duk_get_prop_string(cx, -1, "mutFixup");
+	if (duk_is_function(cx, -1)) {
+		duk_push_heapptr(cx, thisobj);
+		duk_push_false(cx);
+// exception here, push an integer where the node was.
+		duk_push_int(cx, mark);
+		duk_push_heapptr(cx, child);
+		duk_call(cx, 4);
+	}
+	duk_pop_2(cx);
+	return 1;
+
+fail:
+	debugPrint(5, "remove 2");
+	duk_pop(cx);
+	duk_push_null(cx);
 	return 1;
 }
 
 static duk_ret_t native_fetchHTTP(duk_context * cx)
 {
+	jsobjtype thisobj;
+	struct i_get g;
+	const char *incoming_url = duk_safe_to_string(cx, 0);
+	const char *incoming_method = duk_get_string(cx, 1);
+	const char *incoming_headers = duk_get_string(cx, 2);
+	const char *incoming_payload = duk_get_string(cx, 3);
+	char *outgoing_xhrheaders = NULL;
+	char *outgoing_xhrbody = NULL;
+	char *a = NULL, methchar = '?';
+	bool rc, async;
+
 	debugPrint(5, "xhr 1");
-	if (allowXHR) {
-		struct i_get g;
-		const char *incoming_url = duk_safe_to_string(cx, 0);
-		const char *incoming_method = duk_get_string(cx, 1);
-//              const char *incoming_headers = duk_get_string(cx, 2);
-		const char *incoming_payload = duk_get_string(cx, 3);
-		char *outgoing_xhrheaders = NULL;
-		char *outgoing_xhrbody = NULL;
-		char *a = NULL, methchar = '?';
+	duk_push_this(cx);
+	thisobj = duk_get_heapptr(cx, -1);
+	duk_get_prop_string(cx, -1, "async");
+	async = duk_get_boolean(cx, -1);
+	duk_pop_2(cx);
+	if (!down_jsbg)
+		async = false;
 
-		if (!incoming_url)
-			incoming_url = emptyString;
-		if (incoming_payload && *incoming_payload) {
-			if (incoming_method
-			    && stringEqualCI(incoming_method, "post"))
-				methchar = '\1';
-			if (asprintf(&a, "%s%c%s",
-				     incoming_url, methchar,
-				     incoming_payload) < 0)
-				i_printfExit(MSG_MemAllocError, 50);
-			incoming_url = a;
-		}
+// asynchronous xhr before browse and after browse go down different paths.
+// So far I can't get the before browse path to work,
+// at least on nasa.gov, which has lots of xhrs in its onload code.
+// It pushes things over to timers, which work, but the page is rendered
+// shortly after browse time instead of at browse time, which is annoying.
+	if (!cw->browseMode)
+		async = false;
 
-		debugPrint(3, "xhr %s", incoming_url);
-		memset(&g, 0, sizeof(g));
-		g.thisfile = cf->fileName;
-		g.uriEncoded = true;
-		g.url = incoming_url;
-		g.headers_p = &outgoing_xhrheaders;
-		httpConnect(&g);
-		outgoing_xhrbody = g.buffer;
-		nzFree(a);
-		if (intFlag) {
-			duk_get_global_string(cx, "eb$stopexec");
-// this next line should fail and stop the script!
-			duk_call(cx, 0);
-// It didn't stop the script, oh well.
-			duk_pop(cx);
-		}
-		if (outgoing_xhrheaders == NULL)
-			outgoing_xhrheaders = emptyString;
-		if (outgoing_xhrbody == NULL)
-			outgoing_xhrbody = emptyString;
+	if (!incoming_url)
+		incoming_url = emptyString;
+	if (incoming_payload && *incoming_payload) {
+		if (incoming_method && stringEqualCI(incoming_method, "post"))
+			methchar = '\1';
+		if (asprintf(&a, "%s%c%s",
+			     incoming_url, methchar, incoming_payload) < 0)
+			i_printfExit(MSG_MemAllocError, 50);
+		incoming_url = a;
+	}
+
+	debugPrint(3, "xhr send %s", incoming_url);
+
+// async and sync are completely different
+	if (async) {
+		const char *fpn = fakePropName();
+		struct htmlTag *t =
+		    newTag(cw->browseMode ? "object" : "script");
+		t->deleted = true;	// do not render this tag
+		t->step = 3;
+		t->async = true;
+		t->inxhr = true;
+		t->f0 = cf;
+		connectTagObject(t, thisobj);
 		duk_pop_n(cx, 4);
-		duk_push_string(cx, "");
-		duk_push_string(cx, outgoing_xhrheaders);
-		duk_push_string(cx, outgoing_xhrbody);
-		duk_join(cx, 2);
-		nzFree(outgoing_xhrheaders);
-		nzFree(outgoing_xhrbody);
+// This routine will return, and javascript might stop altogether; do we need
+// to protect this object from garbage collection?
+		duk_push_global_object(cx);
+		duk_push_this(cx);
+		duk_push_string(cx, fpn);
+		duk_def_prop(cx, 0,
+			     (DUK_DEFPROP_HAVE_VALUE |
+			      DUK_DEFPROP_SET_ENUMERABLE |
+			      DUK_DEFPROP_CLEAR_WRITABLE |
+			      DUK_DEFPROP_SET_CONFIGURABLE));
+		duk_pop(cx);	// don't need global any more
+		duk_push_this(cx);
+		duk_push_string(cx, fpn);
+		duk_put_prop_string(cx, 0, "backlink");
+		duk_pop(cx);
+// That takes care of garbage collection.
+// Now everything has to be allocated.
+		t->href = (a ? a : cloneString(incoming_url));
+// overloading the innerHTML field
+		t->innerHTML = cloneString(incoming_headers);
+		if (cw->browseMode)
+			scriptSetsTimeout(t);
+		pthread_create(&t->loadthread, NULL, httpConnectBack3,
+			       (void *)t);
+		duk_push_string(cx, "async");
+		return 1;
+	}
+
+	memset(&g, 0, sizeof(g));
+	g.thisfile = cf->fileName;
+	g.uriEncoded = true;
+	g.url = incoming_url;
+	g.custom_h = incoming_headers;
+	g.headers_p = &outgoing_xhrheaders;
+	rc = httpConnect(&g);
+	outgoing_xhrbody = g.buffer;
+	nzFree(a);
+	if (intFlag) {
+		duk_get_global_string(cx, "eb$stopexec");
+// this next line should fail and stop the script!
+		duk_call(cx, 0);
+// It didn't stop the script, oh well.
+		duk_pop(cx);
+	}
+	if (outgoing_xhrheaders == NULL)
+		outgoing_xhrheaders = emptyString;
+	if (outgoing_xhrbody == NULL)
+		outgoing_xhrbody = emptyString;
+	duk_pop_n(cx, 4);
+	duk_push_string(cx, "");
+	duk_push_string(cx, "\r\n\r\n");
+	duk_push_int(cx, rc);
+	duk_push_int(cx, g.code);
+	duk_push_string(cx, outgoing_xhrheaders);
+	duk_join(cx, 3);
+	duk_push_string(cx, outgoing_xhrbody);
+	duk_join(cx, 2);
+	nzFree(outgoing_xhrheaders);
+	nzFree(outgoing_xhrbody);
 // http fetch could bring new cookies into the current window.
 // Can I just call startCookie() again to refresh the cookie copy?
-	} else {
-		duk_pop_n(cx, 4);
-		duk_push_string(cx, emptyString);
-	}
 
 	debugPrint(5, "xhr 2");
 	return 1;
@@ -1274,6 +1428,22 @@ static duk_ret_t native_qs(duk_context * cx)
 	return 1;
 }
 
+// querySelector0
+static duk_ret_t native_qs0(duk_context * cx)
+{
+	jsobjtype root;
+	bool rc;
+	const char *selstring = duk_get_string(cx, 0);
+	duk_push_this(cx);
+	root = duk_get_heapptr(cx, -1);
+	duk_pop(cx);
+	jsInterruptCheck();
+	rc = querySelector0(selstring, root);
+	duk_pop(cx);
+	duk_push_boolean(cx, rc);
+	return 1;
+}
+
 static duk_ret_t native_cssApply(duk_context * cx)
 {
 	jsInterruptCheck();
@@ -1326,6 +1496,10 @@ void createJavaContext_nat(void)
 	duk_put_global_string(jcx, "my$doc");
 	duk_push_c_function(jcx, native_puts, 1);
 	duk_put_global_string(jcx, "eb$puts");
+	duk_push_c_function(jcx, native_wlf, 2);
+	duk_put_global_string(jcx, "eb$wlf");
+	duk_push_c_function(jcx, native_media, 1);
+	duk_put_global_string(jcx, "eb$media");
 	duk_push_c_function(jcx, native_btoa, 1);
 	duk_put_global_string(jcx, "btoa");
 	duk_push_c_function(jcx, native_atob, 1);
@@ -1374,12 +1548,17 @@ void createJavaContext_nat(void)
 	duk_put_global_string(jcx, "querySelectorAll");
 	duk_push_c_function(jcx, native_qs, DUK_VARARGS);
 	duk_put_global_string(jcx, "querySelector");
+	duk_push_c_function(jcx, native_qs0, 1);
+	duk_put_global_string(jcx, "querySelector0");
 	duk_push_c_function(jcx, native_cssApply, 3);
 	duk_put_global_string(jcx, "eb$cssApply");
 	duk_push_c_function(jcx, native_cssText, 1);
 	duk_put_global_string(jcx, "eb$cssText");
 
 	duk_push_heapptr(jcx, docobj);	// native document methods
+
+	duk_push_c_function(jcx, native_hasfocus, 0);
+	duk_put_prop_string(jcx, -2, "hasFocus");
 	duk_push_c_function(jcx, native_doc_write, DUK_VARARGS);
 	duk_put_prop_string(jcx, -2, "write");
 	duk_push_c_function(jcx, native_doc_writeln, DUK_VARARGS);
@@ -1462,12 +1641,33 @@ static enum ej_proptype top_proptype(void)
 	return EJ_PROP_NONE;	/* don't know */
 }				/* top_proptype */
 
+/*********************************************************************
+http://theconversation.com/how-to-build-a-moon-base-120259
+creates a script object with getter on src, which throws an error if src
+is empty or syntactically invalid.
+I check for src in prepareScript(), to see if I need to load the page.
+It's just a get call, but the getter throws an error,
+and the get call is unprotected, and edbrowse aborts. Ouch!
+Here is a simple get property call that is  called through
+duk_safe_call() and thus protected.
+I hope it doesn't introduce too much overhead, because it is almost never
+needed, but nor do I want edbrowse to abort!
+*********************************************************************/
+
+static duk_ret_t protected_get(duk_context * cx, void *udata)
+{
+	const char *name = udata;
+	duk_get_prop_string(cx, -1, name);
+	return 1;
+}
+
 enum ej_proptype typeof_property_nat(jsobjtype parent, const char *name)
 {
 	enum ej_proptype l;
+	int rc;
 	duk_push_heapptr(jcx, parent);
-	duk_get_prop_string(jcx, -1, name);
-	l = top_proptype();
+	rc = duk_safe_call(jcx, protected_get, (void *)name, 0, 1);
+	l = rc ? 0 : top_proptype();
 	duk_pop_2(jcx);
 	return l;
 }
@@ -1872,9 +2072,14 @@ static void uptrace(jsobjtype node)
 bool run_function_bool_nat(jsobjtype parent, const char *name)
 {
 	int dbl = 3;		// debug level
-	if (stringEqual(name, "ontimer") && !uvw)
-		dbl = 4;
+	int seqno = -1;
 	duk_push_heapptr(jcx, parent);
+	if (stringEqual(name, "ontimer")) {
+		dbl = 4;
+		if (duk_get_prop_string(jcx, -1, "tsn"))
+			seqno = duk_get_int(jcx, -1);
+		duk_pop(jcx);
+	}
 	if (!duk_get_prop_string(jcx, -1, name) || !duk_is_function(jcx, -1)) {
 #if 0
 		if (!errorMessage)
@@ -1884,10 +2089,13 @@ bool run_function_bool_nat(jsobjtype parent, const char *name)
 		return (debugLevel < 3);
 	}
 	duk_insert(jcx, -2);
-	debugPrint(dbl, "execute %s", name);
+	if (seqno > 0)
+		debugPrint(dbl, "exec %s %d", name, seqno);
+	else
+		debugPrint(dbl, "exec %s", name);
 	if (!duk_pcall_method(jcx, 0)) {
 		bool rc = true;
-		debugPrint(dbl, "execution complete");
+		debugPrint(dbl, "exec complete");
 		if (duk_is_boolean(jcx, -1))
 			rc = duk_get_boolean(jcx, -1);
 		if (duk_is_number(jcx, -1))
@@ -1906,7 +2114,7 @@ bool run_function_bool_nat(jsobjtype parent, const char *name)
 	processError();
 	debugPrint(3, "failure on %p.%s()", parent, name);
 	uptrace(parent);
-	debugPrint(3, "execution complete");
+	debugPrint(3, "exec complete");
 	return (debugLevel < 3);
 }				/* run_function_bool_nat */
 
@@ -1945,6 +2153,33 @@ int run_function_onearg_nat(jsobjtype parent, const char *name, jsobjtype child)
 	return rc;
 }				/* run_function_onearg_nat */
 
+// The single argument to the function has to be a string.
+void run_function_onestring_nat(jsobjtype parent, const char *name,
+				const char *s)
+{
+	duk_push_heapptr(jcx, parent);
+	if (!duk_get_prop_string(jcx, -1, name) || !duk_is_function(jcx, -1)) {
+#if 0
+		if (!errorMessage)
+			asprintf(&errorMessage, "no such function %s", name);
+#endif
+		duk_pop_2(jcx);
+		return;
+	}
+	duk_insert(jcx, -2);
+	duk_push_string(jcx, s);	// s is the only argument
+	if (!duk_pcall_method(jcx, 1)) {
+		duk_pop(jcx);
+		return;
+	}
+// error in execution
+	if (intFlag)
+		i_puts(MSG_Interrupted);
+	processError();
+	debugPrint(3, "failure on %p.%s[]", parent, name);
+	uptrace(parent);
+}				/* run_function_onestring_nat */
+
 jsobjtype instantiate_array_nat(jsobjtype parent, const char *name)
 {
 	jsobjtype a;
@@ -2049,7 +2284,41 @@ char *run_script_nat(const char *s)
 	char *result = 0;
 	bool rc;
 	const char *gc;
-	rc = duk_peval_string(jcx, s);
+	char *s2 = 0;
+
+// special debugging code to replace bp@ and trace@ with expanded macros.
+// Warning: breakpoints and tracing can change the flow of execution
+// prior to duktape commit 67c891d9e075cc49281304ff5955cae24faa1496
+	if (strstr(s, "bp@(") || strstr(s, "trace@(")) {
+		int l;
+		const char *u, *v1, *v2;
+		s2 = initString(&l);
+		u = s;
+		while (true) {
+			v1 = strstr(u, "bp@(");
+			v2 = strstr(u, "trace@(");
+			if (v1 && v2 && v2 < v1)
+				v1 = v2;
+			if (!v1)
+				v1 = v2;
+			if (!v1)
+				break;
+			stringAndBytes(&s2, &l, u, v1 - u);
+			stringAndString(&s2, &l, (*v1 == 'b' ?
+						  "(function(arg$,l$ne){if(l$ne) alert('break at line ' + l$ne); while(true){var res = prompt('bp'); if(!res) continue; if(res === '.') break; try { res = eval(res); alert(res); } catch(e) { alert(e.toString()); }}}).call(this,(typeof arguments=='object'?arguments:[]),\""
+						  :
+						  "(function(arg$,l$ne){ if(l$ne === step$go||typeof step$exp==='string'&&eval(step$exp)) step$l = 2; if(step$l == 0) return; if(step$l == 1) { alert3(l$ne); return; } if(l$ne) alert('break at line ' + l$ne); while(true){var res = prompt('bp'); if(!res) continue; if(res === '.') break; try { res = eval(res); alert(res); } catch(e) { alert(e.toString()); }}}).call(this,(typeof arguments=='object'?arguments:[]),\""));
+			v1 = strchr(v1, '(') + 1;
+			v2 = strchr(v1, ')');
+			stringAndBytes(&s2, &l, v1, v2 - v1);
+			stringAndString(&s2, &l, "\");");
+			u = ++v2;
+		}
+		stringAndString(&s2, &l, u);
+	}
+
+	rc = duk_peval_string(jcx, (s2 ? s2 : s));
+	nzFree(s2);
 	if (intFlag)
 		i_puts(MSG_Interrupted);
 	if (!rc) {
@@ -2067,3 +2336,47 @@ char *run_script_nat(const char *s)
 		duk_gc(jcx, 0);
 	return result;
 }
+
+// execute script.data code; more efficient than the above.
+void run_data_nat(jsobjtype o)
+{
+	bool rc;
+	const char *s, *gc;
+	duk_push_heapptr(jcx, o);
+	if (!duk_get_prop_string(jcx, -1, "data")) {
+// no data
+		duk_pop_2(jcx);
+		return;
+	}
+	s = duk_safe_to_string(jcx, -1);
+	if (!s || !*s)
+		return;
+// defer to the earlier routine if there are breakpoints
+	if (strstr(s, "bp@(") || strstr(s, "trace@(")) {
+		run_script_nat(s);
+		duk_pop_2(jcx);
+		return;
+	}
+	rc = duk_peval_string(jcx, s);
+	if (intFlag)
+		i_puts(MSG_Interrupted);
+	if (!rc) {
+		duk_pop_n(jcx, 3);
+	} else {
+		processError();
+		duk_pop_2(jcx);
+	}
+	gc = getenv("JSGC");
+	if (gc && *gc)
+		duk_gc(jcx, 0);
+}
+
+void put_data_nat(jsobjtype cx, jsobjtype o, const char *s)
+{
+	if (!cx || !o || !s || !*s)
+		return;
+	duk_push_heapptr(cx, o);
+	duk_push_string(cx, s);
+	duk_put_prop_string(cx, -2, "data");
+	duk_pop(cx);
+}
