@@ -14,7 +14,7 @@
 
 const char *progname;
 const char eol[] = "\r\n";
-const char *version = "3.7.4";
+const char *version = "3.7.6";
 char *changeFileName;
 char *configFile, *addressFile, *cookieFile;
 char *mailDir, *mailUnread, *mailStash, *mailReply;
@@ -25,7 +25,7 @@ char *ebTempDir, *ebUserDir;
 char *userAgents[MAXAGENT + 1];
 char *currentAgent;
 bool allowRedirection = true, allowJS = true, sendReferrer = true;
-bool allowXHR = true;
+bool blockJS;
 bool ftpActive;
 int webTimeout = 20, mailTimeout = 0;
 int displayLength = 500;
@@ -35,29 +35,10 @@ int localAccount, maxAccount;
 struct MACCOUNT accounts[MAXACCOUNT];
 int maxMime;
 struct MIMETYPE mimetypes[MAXMIME];
-/* filters to save emails in various files */
-#define MAXFILTER 500
-struct FILTERDESC {
-	const char *match;
-	const char *redirect;
-	char type;
-	long expire;
-};
-static struct FILTERDESC mailFilters[MAXFILTER];
-static int n_filters;
-int maxproxy;
-struct PXENT proxyEntries[MAXPROXY];
-/* for edbrowse functions defined in the config file */
-#define MAXEBSCRIPT 500		// this many scripts
-#define MAXNEST 20		// nested blocks
-static char *ebScript[MAXEBSCRIPT + 1];
-static char *ebScriptName[MAXEBSCRIPT + 1];
-#define MAXNOJS 500
-static const char *javaDis[MAXNOJS];
-static int javaDisCount;
 static struct DBTABLE dbtables[MAXDBT];
 static int numTables;
 volatile bool intFlag;
+time_t intStart;
 bool curlActive;
 bool ismc, isimap, passMail;
 char whichproc = 'e';		// edbrowse
@@ -75,10 +56,6 @@ struct ebSession sessionList[MAXSESSION], *cs;
 int maxSession;
 static pthread_mutex_t share_mutex = PTHREAD_MUTEX_INITIALIZER;
 
-#define MAXNOJS 500
-static const char *javaDis[MAXNOJS];
-static int javaDisCount;
-
 /*********************************************************************
 Redirect the incoming mail into a file, based on the subject or the sender.
 Along with the filters in .ebrc, this routine dips into your addressbook,
@@ -91,161 +68,93 @@ If you don't want an alias to act as a redirect filter,
 put a ! at the beginning of the alias name.
 *********************************************************************/
 
-const char *mailRedirect(const char *to, const char *from,
-			 const char *reply, const char *subj)
+static void *inputForever(void *ptr);
+static pthread_t foreground_thread;
+static void finishBrowse(void)
 {
-	int rlen = strlen(reply);
-	int slen = strlen(subj);
-	int tlen = strlen(to);
-	struct FILTERDESC *f;
-	const char *r;
-
-	for (f = mailFilters; f->match; ++f) {
-		const char *m = f->match;
-		int k, mlen = strlen(m);
-
-		r = f->redirect;
-
-		switch (f->type) {
-		case 2:
-			if (stringEqualCI(m, from))
-				return r;
-			if (stringEqualCI(m, reply))
-				return r;
-			if (*m == '@' && mlen < rlen &&
-			    stringEqualCI(m, reply + rlen - mlen))
-				return r;
-			break;
-
-		case 3:
-			if (stringEqualCI(m, to))
-				return r;
-			if (*m == '@' && mlen < tlen
-			    && stringEqualCI(m, to + tlen - mlen))
-				return r;
-			break;
-
-		case 4:
-			if (mlen > slen)
-				break;
-			if (mlen == slen) {
-				if (stringEqualCI(m, subj))
-					return r;
-				break;
-			}
-/* a prefix or suffix match is ok */
-/* have to be at least half the subject line */
-			if (slen > mlen + mlen)
-				break;
-			if (memEqualCI(m, subj, mlen))
-				return r;
-			k = slen - mlen;
-			if (memEqualCI(m, subj + k, mlen))
-				return r;
-			break;
-		}		/* switch */
-	}			/* loop */
-
-	r = reverseAlias(reply);
-	return r;
-}				/* mailRedirect */
-
-/*********************************************************************
-Are we ok to parse and execute javascript?
-*********************************************************************/
-
-bool javaOK(const char *url)
-{
-	int j, hl, dl;
-	const char *h, *d, *q, *path;
-	if (!allowJS)
-		return false;
-	if (!url)
-		return true;
-	if (isDataURI(url))
-		return true;
-	h = getHostURL(url);
-	if (!h)
-		return true;
-	hl = strlen(h);
-	path = getDataURL(url);
-	for (j = 0; j < javaDisCount; ++j) {
-		d = javaDis[j];
-		q = strchr(d, '/');
-		if (!q)
-			q = d + strlen(d);
-		dl = q - d;
-		if (dl > hl)
-			continue;
-		if (!memEqualCI(d, h + hl - dl, dl))
-			continue;
-		if (*q == '/') {
-			++q;
-			if (hl != dl)
-				continue;
-			if (!path)
-				continue;
-			if (strncmp(q, path, strlen(q)))
-				continue;
-			return false;
-		}		/* domain/path was specified */
-		if (hl == dl)
-			return false;
-		if (h[hl - dl - 1] == '.')
-			return false;
+	struct htmlTag *t;
+	int j;
+	char *a, *newbuf;
+	struct ebFrame *f;
+	whichproc = 'e';
+// tags should certainly be set
+	if (!tagList)
+		return;
+// set all scripts to complete
+	for (t = cw->scriptlist; t; t = t->same)
+		t->step = 6;
+// kill any timers
+	for (f = &cw->f0; f; f = f->next)
+		delTimers(f);
+	if (allowJS) {
+		allowJS = false;
+		blockJS = true;
+		i_puts(MSG_JavaOff);
 	}
-	return true;
-}				/* javaOK */
+	if (cw->browseMode)
+		return;
+// We were in the middle of the browse command; this is typical.
+	a = render(0);
+	newbuf = htmlReformat(a);
+	nzFree(a);
+	cw->rnlMode = cw->nlMode;
+	cw->nlMode = false;
+	cw->binMode = false;
+	cw->r_dot = cw->dot, cw->r_dol = cw->dol;
+	cw->dot = cw->dol = 0;
+	cw->r_map = cw->map;
+	cw->map = 0;
+	memcpy(cw->r_labels, cw->labels, sizeof(cw->labels));
+	memset(cw->labels, 0, sizeof(cw->labels));
+	j = strlen(newbuf);
+	addTextToBuffer((pst) newbuf, j, 0, false);
+	free(newbuf);
+	cw->undoable = false;
+	cw->changeMode = false;
+	if (cf->fileName) {
+		j = strlen(cf->fileName);
+		cf->fileName = reallocMem(cf->fileName, j + 8);
+		strcat(cf->fileName, ".browse");
+	}
+	cw->browseMode = true;
+}
 
 /* Catch interrupt and react appropriately. */
 static void catchSig(int n)
 {
+	time_t now;
+	pthread_t t1 = foreground_thread, t2;
+	int j;
 	intFlag = true;
-	if (inInput)
-		i_puts(MSG_EnterInterrupt);
 /* If we were reading from a file, or socket, this signal should
  * cause the read to fail.  Check for intFlag, so we know it was
  * interrupted, and not an io failure.
  * Then clean up appropriately. */
 	signal(SIGINT, catchSig);
-}				/* catchSig */
-
-bool isSQL(const char *s)
-{
-	char c = *s;
-	const char *c1 = 0;
-
-	if (!sqlPresent)
-		goto no;
-
-	if (isURL(s))
-		goto no;
-
-// look for word] or word:word]
-	if (!isalphaByte(c))
-		goto no;
-
-	for (++s; (c = *s); ++s) {
-		if (c == '_')
-			continue;
-		if (isalnumByte(c))
-			continue;
-		if (c == ':') {
-			if (c1)
-				goto no;
-			c1 = s;
-			continue;
-		}
-		if (c == ']')
-			goto yes;
+	if (inInput) {
+		i_puts(MSG_EnterInterrupt);
+		return;
 	}
-
-no:
-	return false;
-
-yes:
-	return true;
-}				/* isSQL */
+	if (!intStart) {	// first time hitting ^C
+		time(&intStart);
+		return;
+	}
+	time(&now);
+// should 45 seconds be configurable?
+	if (now < intStart + 45)
+		return;
+// Let's do something drastic here; start a new thread and exit the current one.
+	i_puts(MSG_IntForce);
+	if (whichproc == 'j')
+		finishBrowse();
+// Doing this stuff from a signal handler?  idk
+	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &j);
+// without this line the thread never dies, and js runs, and the CPU churns.
+	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &j);
+	if (pthread_create(&t2, NULL, inputForever, NULL))
+		return;		// didn't work
+	pthread_cancel(t1);
+}				/* catchSig */
 
 void setDataSource(char *v)
 {
@@ -359,6 +268,220 @@ void ebClose(int n)
 	exit(n);
 }				/* ebClose */
 
+static struct ebhost {
+// j = nojs, v = novs, p = proxy, f = function,
+// s = subject, t = to, r = reply, a = agentsite
+	char type, filler;
+	short n;
+// watch out, these fields are highly overloaded, depending on type
+	char *host;
+// for proxy entry we also have
+	char *prot, *domain;
+} *ebhosts;
+static size_t ebhosts_avail, ebhosts_max;
+
+static void add_ebhost(char *host, char type)
+{
+	if (ebhosts_max == 0) {
+		ebhosts_max = 32;
+		ebhosts = allocZeroMem(ebhosts_max * sizeof(struct ebhost));
+	} else if (ebhosts_avail >= ebhosts_max) {
+		ebhosts_max *= 2;
+		ebhosts =
+		    reallocMem(ebhosts, ebhosts_max * sizeof(struct ebhost));
+	}
+	ebhosts[ebhosts_avail].host = host;
+	ebhosts[ebhosts_avail++].type = type;
+}				/* add_ebhost */
+
+static void delete_ebhosts(void)
+{
+	nzFree(ebhosts);
+	ebhosts = NULL;
+	ebhosts_avail = ebhosts_max = 0;
+}				/* delete_ebhosts */
+
+static void add_proxy(char *v)
+{
+	char *q;
+	char *prot = 0, *domain = 0, *proxy = 0;
+	spaceCrunch(v, true, true);
+	q = strchr(v, ' ');
+	if (q) {
+		*q = 0;
+		if (!stringEqual(v, "*"))
+			prot = v;
+		v = q + 1;
+		q = strchr(v, ' ');
+		if (q) {
+			*q = 0;
+			if (!stringEqual(v, "*"))
+				domain = v;
+			v = q + 1;
+		}
+	}
+	if (!stringEqualCI(v, "direct"))
+		proxy = v;
+	add_ebhost(proxy, 'p');
+	ebhosts[ebhosts_avail - 1].prot = prot;
+	ebhosts[ebhosts_avail - 1].domain = domain;
+}
+
+// Are we ok to parse and execute javascript?
+bool javaOK(const char *url)
+{
+	int j;
+	if (!allowJS)
+		return false;
+	if (isDataURI(url))
+		return true;
+	for (j = 0; j < ebhosts_avail; ++j)
+		if (ebhosts[j].type == 'j' &&
+		    patternMatchURL(url, ebhosts[j].host))
+			return false;
+	return true;
+}				/* javaOK */
+
+/* Return true if the cert for this host should be verified. */
+bool mustVerifyHost(const char *url)
+{
+	int i;
+	if (!verifyCertificates)
+		return false;
+	for (i = 0; i < ebhosts_avail; i++)
+		if (ebhosts[i].type == 'v' &&
+		    patternMatchURL(url, ebhosts[i].host))
+			return false;
+	return true;
+}				/* mustVerifyHost */
+
+/*********************************************************************
+Given a protocol and a domain, find the proxy server
+to mediate your request.
+This is the C version, using entries in .ebrc.
+There is a javascript version of the same name, that we will support later.
+This is a beginning, and it can be used even when javascript is disabled.
+A return of null means DIRECT, and this is the default
+if we don't match any of the proxy entries.
+*********************************************************************/
+
+const char *findProxyForURL(const char *url)
+{
+	struct ebhost *px = ebhosts;
+	int i;
+	char prot[MAXPROTLEN], host[MAXHOSTLEN];
+
+	if (!getProtHostURL(url, prot, host)) {
+/* this should never happen */
+		return 0;
+	}
+
+/* first match wins */
+	for (i = 0; i < ebhosts_avail; ++i, ++px) {
+		if (px->type != 'p')
+			continue;
+
+		if (px->prot) {
+			char *s = px->prot;
+			char *t;
+			int rc;
+			while (*s) {
+				t = strchr(s, '|');
+				if (t)
+					*t = 0;
+				rc = stringEqualCI(s, prot);
+				if (t)
+					*t = '|';
+				if (rc)
+					goto domain;
+				if (!t)
+					break;
+				s = t + 1;
+			}
+			continue;
+		}
+
+domain:
+		if (!px->domain || patternMatchURL(url, px->domain))
+			return px->host;
+	}
+
+	return 0;
+}
+
+const char *findAgentForURL(const char *url)
+{
+	struct ebhost *px = ebhosts;
+	int i;
+	for (i = 0; i < ebhosts_avail; ++i, ++px)
+		if (px->type == 'a' && patternMatchURL(url, px->host))
+			return userAgents[px->n];
+	return 0;
+}
+
+const char *mailRedirect(const char *to, const char *from,
+			 const char *reply, const char *subj)
+{
+	int rlen = strlen(reply);
+	int slen = strlen(subj);
+	int tlen = strlen(to);
+	int mlen;		// length of match string
+	int i, k;
+	struct ebhost *f = ebhosts;
+	const char *m, *r;	// match and redirect
+
+	for (i = 0; i < ebhosts_avail; ++i, ++f) {
+		char type = f->type;
+		if (strchr("rts", type)) {
+			m = f->prot;
+			mlen = strlen(m);
+			r = f->host;
+		}
+
+		switch (type) {
+		case 'r':
+			if (stringEqualCI(m, from))
+				return r;
+			if (stringEqualCI(m, reply))
+				return r;
+			if (*m == '@' && mlen < rlen &&
+			    stringEqualCI(m, reply + rlen - mlen))
+				return r;
+			break;
+
+		case 't':
+			if (stringEqualCI(m, to))
+				return r;
+			if (*m == '@' && mlen < tlen
+			    && stringEqualCI(m, to + tlen - mlen))
+				return r;
+			break;
+
+		case 's':
+			if (mlen > slen)
+				break;
+			if (mlen == slen) {
+				if (stringEqualCI(m, subj))
+					return r;
+				break;
+			}
+/* a prefix or suffix match is ok */
+/* have to be at least half the subject line */
+			if (slen > mlen + mlen)
+				break;
+			if (memEqualCI(m, subj, mlen))
+				return r;
+			k = slen - mlen;
+			if (memEqualCI(m, subj + k, mlen))
+				return r;
+			break;
+		}		/* switch */
+	}			/* loop */
+
+	r = reverseAlias(reply);
+	return r;
+}				/* mailRedirect */
+
 static void setupEdbrowseTempDirectory(void)
 {
 	int userid;
@@ -388,13 +511,15 @@ static void setupEdbrowseTempDirectory(void)
 	if (fileTypeByName(ebTempDir, false) != 'd') {
 /* no such directory, try to make it */
 /* this temp edbrowse directory is used by everyone system wide */
-		if (mkdir(ebTempDir, 0777)) {
+		if (mkdir(ebTempDir, MODE_rwx)) {
 			i_printf(MSG_TempDir, ebTempDir);
 			ebTempDir = 0;
 			return;
 		}
+#ifndef DOSLIKE
 // yes, we called mkdir with 777 above, but that was cut by umask.
-		chmod(ebTempDir, 0777);
+		chmod(ebTempDir, MODE_rwx);
+#endif
 	}
 // make room for user ID on the end
 	ebUserDir = allocMem(strlen(ebTempDir) + 20);
@@ -421,6 +546,8 @@ static void setupEdbrowseTempDirectory(void)
  * So now you can edit *.c, on any operating system,
  * and it will do the right thing, with no work on my part. */
 
+static void loadReplacements(void);
+
 int main(int argc, char **argv)
 {
 	int cx, account;
@@ -480,7 +607,7 @@ int main(int argc, char **argv)
 	sprintf(configFile, "%s/.ebrc", home);
 /* if not present then create it, as was done above */
 	if (fileTypeByName(configFile, false) == 0) {
-		int fh = creat(configFile, 0600);
+		int fh = creat(configFile, MODE_private);
 		if (fh >= 0) {
 			write(fh, ebrc_string, strlen(ebrc_string));
 			close(fh);
@@ -530,6 +657,8 @@ int main(int argc, char **argv)
 	pcre_malloc = allocMem;
 	pcre_free = nzFree;
 
+	loadReplacements();
+
 	if (argc && stringEqual(argv[0], "-c")) {
 		if (argc == 1) {
 			*argv = configFile;
@@ -759,6 +888,13 @@ int main(int argc, char **argv)
 	if (cx > 1)
 		cxSwitch(1, false);
 
+	inputForever(NULL);
+	return 0;
+}
+
+static void *inputForever(void *ptr)
+{
+	foreground_thread = pthread_self();
 	while (true) {
 		pst p = inputLine();
 		pst save_p = clonePstring(p);
@@ -771,7 +907,8 @@ int main(int argc, char **argv)
 			linePending = save_p;
 		}
 	}			/* infinite loop */
-}				/* main */
+	return NULL;
+}
 
 /* Find the balancing brace in an edbrowse function */
 static const char *balance(const char *ip, int direction)
@@ -803,6 +940,7 @@ static const char *balance(const char *ip, int direction)
 	return ip;
 }				/* balance */
 
+#define MAXNEST 20		// nested blocks
 /* Run an edbrowse function, as defined in the config file. */
 /* This function must be reentrant. */
 bool runEbFunction(const char *line)
@@ -838,17 +976,20 @@ bool runEbFunction(const char *line)
 			setError(MSG_BadFunctionName);
 			goto fail;
 		}
-	for (j = 0; ebScript[j]; ++j)
-		if (stringEqualCI(linecopy, ebScriptName[j] + 1))
+	for (j = 0; j < ebhosts_avail; ++j)
+		if (ebhosts[j].type == 'f' &&
+		    stringEqualCI(linecopy, ebhosts[j].prot + 1))
 			break;
-	if (!ebScript[j]) {
+	if (j == ebhosts_avail) {
 		setError(MSG_NoSuchFunction, linecopy);
 		goto fail;
 	}
-
+// This or a downstream function could invoke config.
+// Don't know why anybody would do that!
+	fncopy = cloneString(ebhosts[j].host);
 /* skip past the leading \n */
-	ip = ebScript[j] + 1;
-	nofail = (ebScriptName[j][0] == '+');
+	ip = fncopy + 1;
+	nofail = (ebhosts[j].prot[0] == '+');
 	nest = 0;
 	ok = true;
 
@@ -875,11 +1016,6 @@ bool runEbFunction(const char *line)
 		argl[j] = strlen(s);
 	}
 
-// This or a downstream function could invoke config.
-// Don't know why anybody would do that!
-	fncopy = cloneString(ip);
-	ip = fncopy;
-
 	while ((code = *ip)) {
 		if (intFlag) {
 			setError(MSG_Interrupted);
@@ -977,6 +1113,7 @@ ahead:
 
 /* Here we go! */
 		debugPrint(3, "< %s", new);
+		jClearSync();
 		ok = edbrowseCommand(new, true);
 		free(new);
 
@@ -1032,20 +1169,13 @@ void unreadConfigFile(void)
 	nzFree(configMemory);
 	configMemory = 0;
 
-	memset(mailFilters, 0, sizeof(mailFilters));
-	n_filters = 0;
 	memset(accounts, 0, sizeof(accounts));
 	maxAccount = localAccount = 0;
 	memset(mimetypes, 0, sizeof(mimetypes));
 	maxMime = 0;
 	memset(dbtables, 0, sizeof(dbtables));
 	numTables = 0;
-	memset(proxyEntries, 0, sizeof(proxyEntries));
-	maxproxy = 0;
-	memset(ebScript, 0, sizeof(ebScript));
-	memset(ebScriptName, 0, sizeof(ebScriptName));
 	memset(userAgents + 1, 0, sizeof(userAgents) - sizeof(userAgents[0]));
-	javaDisCount = 0;
 
 	addressFile = NULL;
 	cookieFile = NULL;
@@ -1064,7 +1194,7 @@ void unreadConfigFile(void)
 
 	setDataSource(NULL);
 	setHTTPLanguage(eb_language);
-	deleteNovsHosts();
+	delete_ebhosts();
 }				/* unreadConfigFile */
 
 /* Order is important here: mail{}, mime{}, table{}, then global keywords */
@@ -1082,7 +1212,7 @@ static const char *const keywords[] = {
 	"downdir", "maildir", "agent",
 	"jar", "nojs", "cachedir",
 	"webtimer", "mailtimer", "certfile", "datasource", "proxy",
-	"linelength", "localizeweb", "notused33", "novs", "cachesize",
+	"agentsite", "localizeweb", "notused33", "novs", "cachesize",
 	"adbook", 0
 };
 
@@ -1109,7 +1239,6 @@ void readConfigFile(void)
 	char last[24];
 	int lidx = 0;
 	struct MACCOUNT *act;
-	struct PXENT *px;
 	struct MIMETYPE *mt;
 	struct DBTABLE *td;
 
@@ -1306,12 +1435,8 @@ putc:
 				++v;
 			if (!*v)
 				cfgLine1(MSG_EBRC_MatchNowh, s);
-			if (n_filters == MAXFILTER - 1)
-				cfgLine0(MSG_EBRC_Filters);
-			mailFilters[n_filters].match = s;
-			mailFilters[n_filters].redirect = v;
-			mailFilters[n_filters].type = mailblock;
-			++n_filters;
+			add_ebhost(v, "xxrts"[mailblock]);
+			ebhosts[ebhosts_avail - 1].prot = s;
 			continue;
 		}
 
@@ -1514,21 +1639,20 @@ putc:
 			ftype = fileTypeByName(v, false);
 			if (ftype && ftype != 'f')
 				cfgAbort1(MSG_EBRC_JarNotFile, v);
-			j = open(v, O_WRONLY | O_APPEND | O_CREAT, 0600);
+			j = open(v, O_WRONLY | O_APPEND | O_CREAT,
+				 MODE_private);
 			if (j < 0)
 				cfgAbort1(MSG_EBRC_JarNoWrite, v);
 			close(j);
 			continue;
 
 		case 24:	/* nojs */
-			if (javaDisCount == MAXNOJS)
-				cfgAbort1(MSG_EBRC_NoJS, MAXNOJS);
 			if (*v == '.')
 				++v;
 			q = strchr(v, '.');
 			if (!q || q[1] == 0)
 				cfgLine1(MSG_EBRC_DomainDot, v);
-			javaDis[javaDisCount++] = v;
+			add_ebhost(v, 'j');
 			continue;
 
 		case 25:	/* cachedir */
@@ -1560,33 +1684,22 @@ putc:
 			continue;
 
 		case 30:	/* proxy */
-			if (maxproxy == MAXPROXY)
-				cfgAbort1(MSG_EBRC_NoPROXY, MAXPROXY);
-			px = proxyEntries + maxproxy;
-			maxproxy++;
-			spaceCrunch(v, true, true);
-			q = strchr(v, ' ');
-			if (q) {
-				*q = 0;
-				if (!stringEqual(v, "*"))
-					px->prot = v;
-				v = q + 1;
-				q = strchr(v, ' ');
-				if (q) {
-					*q = 0;
-					if (!stringEqual(v, "*"))
-						px->domain = v;
-					v = q + 1;
-				}
-			}
-			if (!stringEqualCI(v, "direct"))
-				px->proxy = v;
+			add_proxy(v);
 			continue;
 
-		case 31:	/* linelength */
-			displayLength = atoi(v);
-			if (displayLength < 80)
-				displayLength = 80;
+		case 31:	// agentsite
+			spaceCrunch(v, true, true);
+			q = strchr(v, ' ');
+			if (!q || strchr(q + 1, ' ') ||
+			    (j = stringIsNum(q + 1)) < 0)
+				break;
+			if (j >= MAXAGENT || !userAgents[j]) {
+				cfgLine1(MSG_EBRC_NoAgent, j);
+				continue;
+			}
+			*q = 0;
+			add_ebhost(v, 'a');
+			ebhosts[ebhosts_avail - 1].n = j;
 			continue;
 
 		case 32:	/* localizeweb */
@@ -1601,7 +1714,7 @@ putc:
 			q = strchr(v, '.');
 			if (!q || q[1] == 0)
 				cfgLine1(MSG_EBRC_DomainDot, v);
-			addNovsHost(v);
+			add_ebhost(v, 'v');
 			continue;
 
 		case 35:	/* cachesize */
@@ -1727,7 +1840,6 @@ nokeyword:
 				goto putback;
 /* This ends the function */
 			*s = 0;	/* null terminate the script */
-			++sn;
 			continue;
 		}
 
@@ -1802,11 +1914,10 @@ nokeyword:
 
 		if (c == 'f') {
 			stack[++nest] = c;
-			if (sn == MAXEBSCRIPT)
-				cfgAbort1(MSG_EBRC_ManyFn, sn);
-			ebScriptName[sn] = s + 2;
+			sn = ebhosts_avail;
 			t[-1] = 0;
-			ebScript[sn] = t;
+			add_ebhost(t, 'f');
+			ebhosts[sn].prot = s + 2;
 			goto putback;
 		}
 
@@ -1819,7 +1930,7 @@ putback:
 	}			/* loop over lines */
 
 	if (nest)
-		cfgAbort1(MSG_EBRC_FnNotClosed, ebScriptName[sn]);
+		cfgAbort1(MSG_EBRC_FnNotClosed, ebhosts[sn].prot + 1);
 
 	if (mailblock | mimeblock)
 		cfgAbort0(MSG_EBRC_MNotClosed);
@@ -1827,3 +1938,73 @@ putback:
 	if (maxAccount && !localAccount)
 		localAccount = 1;
 }				/* readConfigFile */
+
+// local replacements for javascript and css
+struct JSR {
+	struct JSR *next;
+	char *url, *locf;
+};
+static struct JSR *jsr_top;
+
+const char *fetchReplace(const char *u)
+{
+	struct JSR *j = jsr_top;
+	const char *s;
+	int l;
+	if (!j)
+		return 0;	// high runner case
+// This feature only works if you are browsing a local website.
+// Save the home page to a file called base, add a <base> tag, and browse that.
+	if (!browseLocal)
+		return 0;
+	s = strchr(u, '?');
+	l = (s ? s - u : strlen(u));
+	while (j) {
+		if (!strncmp(j->url, u, l) && j->url[l] == 0)
+			return j->locf;
+		j = j->next;
+	}
+	return 0;
+}
+
+static void loadReplacements(void)
+{
+	FILE *f = fopen("jslocal", "r");
+	struct JSR *j;
+	char *s;
+	int n = 0;
+	char line[400];
+	if (!f)
+		return;
+	while (fgets(line, sizeof(line), f)) {
+		s = strchr(line, '\n');
+		if (!s) {
+			fprintf(stderr, "jslocal line too long\n");
+			fclose(f);
+			return;
+		}
+		*s = 0;
+		if (s > line && s[-1] == '\r')
+			*--s = 0;
+		if (line[0] == '#' || line[0] == 0)
+			continue;
+		s = strchr(line, ':');
+		if (!s) {
+			fprintf(stderr, "jslocal line has no :\n");
+			continue;
+		}
+		*s++ = 0;
+		if (strchr(s, '?')) {
+			fprintf(stderr, "jslocal line has ?\n");
+			continue;
+		}
+		j = allocMem(sizeof(struct JSR));
+		j->locf = cloneString(line);
+		j->url = cloneString(s);
+		j->next = jsr_top;
+		jsr_top = j;
+		++n;
+	}
+	fclose(f);
+	debugPrint(3, "%d js or css file replacements", n);
+}
