@@ -14,28 +14,28 @@ blank spaces for messages no longer used.
 Each file lang/msg-* must have exactly this many lines.
 *********************************************************************/
 
-#define EdbrowseMessageCount 659
+#define EdbrowseMessageCount 660
 
 // English
-extern const char *msg_en[], ebrc_en[];
+extern const char *msg_en[], ebrc_en[], qrg_en[];
 
 // French, Erwin Bliesenick, erwinb@no-log.org
-extern const char *msg_fr[], ebrc_fr[];
+extern const char *msg_fr[], ebrc_fr[], qrg_fr[];
 
 // Brazilian Portuguese, Cleverson Casarin, clcaul@gmail.com
-extern const char *msg_pt_br[], ebrc_pt_br[];
+extern const char *msg_pt_br[], ebrc_pt_br[], qrg_pt_br[];
 
 // German, Sebastian Humenda, shumenda@gmx.de
-extern const char *msg_de[], ebrc_de[];
+extern const char *msg_de[], ebrc_de[], qrg_de[];
 
 // Polish, Wojciech Gac, wojciech.s.gac@gmail.com
-extern const char *msg_pl[], ebrc_pl[];
+extern const char *msg_pl[], ebrc_pl[], qrg_pl[];
 
 // Russian, Wojciech Gac, wojciech.s.gac@gmail.com
-extern const char *msg_ru[], ebrc_ru[];
+extern const char *msg_ru[], ebrc_ru[], qrg_ru[];
 
 // Italian, Enrico Mioso, mrkiko.rs@gmail.com
-extern const char *msg_it[], ebrc_it[];
+extern const char *msg_it[], ebrc_it[], qrg_it[];
 
 /*********************************************************************
 This file and lang/msg-en line up.
@@ -563,8 +563,8 @@ enum {
 	MSG_GlobalCommand2,
 	MSG_Doubleclick,
 	MSG_OptionSync,
-	MSG_XhrOff,
-	MSG_XhrOn,
+	MSG_notused466,
+	MSG_notused467,
 	MSG_GetLocalJS,
 	MSG_GetJS,
 	MSG_GetJS2,
@@ -592,7 +592,7 @@ enum {
 	MSG_EBRC_NoCondFile,
 	MSG_EBRC_NoMatchStr,
 	MSG_EBRC_MatchNowh,
-	MSG_EBRC_Filters,
+	MSG_EBRC_NoAgent,
 	MSG_EBRC_BadKeyword,
 	MSG_EBRC_MailAttrOut,
 	MSG_EBRC_MimeAttrOut,
@@ -610,7 +610,7 @@ enum {
 	MSG_EBRC_ManyAgents,
 	MSG_EBRC_JarNotFile,
 	MSG_EBRC_JarNoWrite,
-	MSG_EBRC_NoJS,
+	MSG_EmailInput,
 	MSG_EBRC_DomainDot,
 	MSG_NoAccessSecure,
 	MSG_EBRC_SSLNoFile,
@@ -638,7 +638,7 @@ enum {
 	MSG_EBRC_ManyAcc,
 	MSG_EBRC_ManyTypes,
 	MSG_EBRC_ManyTables,
-	MSG_EBRC_ManyFn,
+	MSG_UrlInput,
 	MSG_EBRC_TooDeeply,
 	MSG_EBRC_FnNotClosed,
 	MSG_EBRC_MNotClosed,
@@ -652,7 +652,7 @@ enum {
 	MSG_ManyOpen,
 	MSG_InvalidSession,
 	MSG_InvalidLineNb,
-	MSG_EBRC_NoPROXY,
+	MSG_notused555,
 	MSG_SessionOutRange,
 	MSG_DoubleInit,
 	MSG_QuitNoActive,
@@ -665,23 +665,23 @@ enum {
 	MSG_FetchNotBackgnd,
 	MSG_NoMailDir,
 	MSG_NoDirChange,
-	MSG_notused568,
+	MSG_JavaBlock,
 	MSG_LogPass,
 	MSG_SortAlpha,
 	MSG_SortSize,
 	MSG_SortDate,
-	MSG_notused573,
-	MSG_notused574,
+	MSG_ConvUnix,
+	MSG_ConvDos,
 	MSG_BadTagCode,
-	MSG_notused576,
+	MSG_IntForce,
 	MSG_HtmlNotreentrant,
 	MSG_UnexSubmitForm,
 	MSG_NullListInform,
 	MSG_DebugEventOff,
 	MSG_DebugEventOn,
 	MSG_JavaContextError,
-	MSG_notused583,
-	MSG_notused584,
+	MSG_JSDownForeground,
+	MSG_JSDownBackground,
 	MSG_notused585,
 	MSG_TimersOff,
 	MSG_TimersOn,
@@ -753,8 +753,9 @@ enum {
 	MSG_SSLConnectError,
 	MSG_InputTTY,
 	MSG_InputReadLine,
-	MSG_notused656,
-	MSG_notused657,
+	MSG_DebugTraceOff,
+	MSG_DebugTraceOn,
 	MSG_SystemCmdFail,
-	MSG_notused659,
+	MSG_ColorOff,
+	MSG_ColorOn,
 };
