@@ -23,7 +23,7 @@ FILE *debugFile = NULL;
 char *debugFileName;
 bool demin = false;
 bool gotimers = true;
-char *uvw;
+bool uvw;
 char *downDir, *home;
 
 /*********************************************************************
@@ -60,8 +60,13 @@ void *reallocMem(void *p, size_t n)
 		i_printfExit(MSG_ReallocP);
 	if (!p)
 		i_printfExit(MSG_Realloc0, n);
-	if (p == emptyString)
-		return allocMem(n);
+	if (p == emptyString) {
+		p = allocMem(n);
+// keep the null byte that was present in emptyString.
+// fileIntoMemory() needs this to keep null on the end of an empty file
+// that was just read into memory.
+		*(char *)p = 0;
+	}
 	if (!(s = realloc(p, n)))
 		i_printfExit(MSG_ErrorRealloc, n);
 	return s;
@@ -95,17 +100,6 @@ void cnzFree(const void *v)
 	nzFree((void *)v);
 }				/* cnzFree */
 
-uchar fromHex(char d, char e)
-{
-	d |= 0x20, e |= 0x20;
-	if (d >= 'a')
-		d -= ('a' - '9' - 1);
-	if (e >= 'a')
-		e -= ('a' - '9' - 1);
-	d -= '0', e -= '0';
-	return ((((uchar) d) << 4) | (uchar) e);
-}				/* fromHex */
-
 char *appendString(char *s, const char *p)
 {
 	int slen = strlen(s);
@@ -774,7 +768,8 @@ bool memoryOutToFile(const char *filename, const char *data, int len,
 /* specify the error messages */
 		     int msgcreate, int msgwrite)
 {
-	int fh = open(filename, O_CREAT | O_TRUNC | O_WRONLY | O_BINARY, 0666);
+	int fh =
+	    open(filename, O_CREAT | O_TRUNC | O_WRONLY | O_BINARY, MODE_rw);
 	if (fh < 0) {
 		setError(msgcreate, filename, errno);
 		return false;
@@ -788,6 +783,25 @@ bool memoryOutToFile(const char *filename, const char *data, int len,
 	return true;
 }				/* memoryOutToFile */
 
+// portable function to truncate to 0
+void truncate0(const char *filename, int fh)
+{
+#ifndef DOSLIKE
+// unix is easy
+	if (fh < 0)
+		truncate(filename, 0l);
+	else
+		ftruncate(fh, 0l);
+#else
+	int fh2;
+	if (fh >= 0)
+		lseek(fh, 0L, 0);
+	fh2 = open(filename, O_WRONLY | O_TRUNC);
+	if (fh2 >= 0)		// it should be
+		close(fh2);
+#endif
+}
+
 /* shift string to upper, lower, or mixed case */
 /* action is u, l, or m. */
 void caseShift(char *s, char action)
@@ -830,6 +844,18 @@ void caseShift(char *s, char action)
 	}			/* loop */
 }				/* caseShift */
 
+// foo-bar has to become fooBar
+void camelCase(char *s)
+{
+	char *t, *w;
+	for (t = w = s; *t; ++t)
+		if (*t == '-' && isalpha(t[1]))
+			t[1] = toupper(t[1]);
+		else
+			*w++ = *t;
+	*w = 0;
+}
+
 /*********************************************************************
 Manage files, directories, and terminal IO.
 You'll see some conditional compilation when this program
@@ -925,7 +951,7 @@ char fileTypeByHandle(int fd)
 	return c;
 }				/* fileTypeByHandle */
 
-int fileSizeByName(const char *name)
+off_t fileSizeByName(const char *name)
 {
 	struct stat buf;
 	if (stat(name, &buf)) {
@@ -935,7 +961,7 @@ int fileSizeByName(const char *name)
 	return buf.st_size;
 }				/* fileSizeByName */
 
-int fileSizeByHandle(int fd)
+off_t fileSizeByHandle(int fd)
 {
 	struct stat buf;
 	if (fstat(fd, &buf)) {
