@@ -118,14 +118,14 @@ void unpercentString(char *s)
  * Some of the chars retain their reserved semantics and should not be changed.
  * This is a friggin guess!
  * All characters in the area between start and end, not including end,
- * are copied or transformed.
+ * are copied or transformed, except the hash, which is removed.
  * This function is used to sanitize user-supplied URLs.  */
 
 /* these punctuations are percentable, anywhere in a url.
  * The order is important.
  * Google has commas in encoded URLs, and wikipedia has parentheses,
  * so those are (sort of) ok. */
-static const char percentable[] = "+,()!*'[]$";
+static const char percentable[] = "+,()'\"\\<>!*[]$";
 static const char hexdigits[] = "0123456789abcdef";
 #define ESCAPED_CHAR_LENGTH 3
 
@@ -177,7 +177,7 @@ bool looksPercented(const char *start, const char *end)
 	if (!end)
 		end = start + strlen(start);
 	for (s = start; s < end; ++s)
-		if (*s < ' ' || strchr(percentable + 4, *s))
+		if (*s < ' ' || strchr(percentable + 5, *s))
 			return false;
 	return true;
 }				/* looksPercented */
@@ -211,6 +211,8 @@ char *htmlEscape0(const char *s, bool do_and)
 }				/* htmlEscape0 */
 
 /* Decide if it looks like a web url. */
+/* Don't do this in a href context  <a href=www.google.com> */
+static bool hrefContext;
 static bool httpDefault(const char *url)
 {
 	static const char *const domainSuffix[] = {
@@ -219,7 +221,10 @@ static bool httpDefault(const char *url)
 		"ca", "de", "jp", "nz", 0
 	};
 	int n, len;
-	const char *s, *lastdot, *end = url + strcspn(url, "/?#\1");
+	const char *s, *lastdot, *end;
+	if (hrefContext)
+		return false;
+	end = url + strcspn(url, "/?#\1");
 	if (end - url > 7 && stringEqual(end - 7, ".browse"))
 		end -= 7;
 	s = strrchr(url, ':');
@@ -325,25 +330,30 @@ static bool parseURL(const char *url, const char **proto, int *prlen, const char
 	}
 
 	if (p) {
-/* You have to have something after the colon */
 		q = p + 1;
 		if (*q == '/')
 			++q;
 		if (*q == '/')
 			++q;
 		skipWhite(&q);
-#if 0
-// javascript: is technically a url
-		if (!*q)
-			return false;
-#endif
+
+		if (!*q) {
+// You have to have something after the colon
+// but javascript: is technically a url, I guess
+			if (strncmp(url, "javascript:", 11))
+				return false;
+		}
+
 		if (proto)
 			*proto = url;
 		if (prlen)
 			*prlen = p - url;
 		a = protocolByName(url, p - url);
+#if 0
+// not sure why I had this code
 		if (a < 0 && q == p + 1)
 			return false;
+#endif
 		if (a >= 0 && !protocols[a].need_slashes)
 			++p;
 		else
@@ -445,6 +455,43 @@ bool isURL(const char *url)
 	return parseURL(url, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
 }				/* isURL */
 
+bool isSQL(const char *s)
+{
+	char c = *s;
+	const char *c1 = 0;
+
+	if (!sqlPresent)
+		goto no;
+
+	if (isURL(s))
+		goto no;
+
+// look for word] or word:word]
+	if (!isalphaByte(c))
+		goto no;
+
+	for (++s; (c = *s); ++s) {
+		if (c == '_')
+			continue;
+		if (isalnumByte(c))
+			continue;
+		if (c == ':') {
+			if (c1)
+				goto no;
+			c1 = s;
+			continue;
+		}
+		if (c == ']')
+			goto yes;
+	}
+
+no:
+	return false;
+
+yes:
+	return true;
+}				/* isSQL */
+
 // non-FTP URLs are always browsable.  FTP URLs are browsable if they end with
 //a slash. gopher urls are a bit more complicated, not yet implemented.
 bool isBrowseableURL(const char *url)
@@ -458,7 +505,7 @@ bool isBrowseableURL(const char *url)
 
 bool isDataURI(const char *u)
 {
-	return memEqualCI(u, "data:", 5);
+	return u && memEqualCI(u, "data:", 5);
 }				/* isDataURI */
 
 /* Helper functions to return pieces of the URL.
@@ -662,8 +709,9 @@ char *getFileURL(const char *url, bool chophash)
 		if (h)
 			e = h;
 	}
-/* if slash at the end then back up to the prior slash */
-	if (e == s && s > url) {
+// if slash at the end then back up to the prior slash
+// /.browse is like / at the end
+	if (s > url && (e == s || (e - s == 7 && !strncmp(s, ".browse", 7)))) {
 		while (s > url && s[-1] == '/')
 			--s;
 		e = s;
@@ -804,19 +852,21 @@ char *resolveURL(const char *base, const char *rel)
 		return cloneString(rel);
 
 	debugPrint(5, "resolve(%s|%s)", base, rel);
+	hrefContext = true;
 	if (!base)
 		base = emptyString;
 	if (!rel)
 		rel = emptyString;
 	n = allocString(strlen(base) + strlen(rel) + 12);
 
-	if (rel[0] == '#' && !strchr(rel, '/')) {
+	if (rel[0] == '#') {
 /* This is an anchor for the current document, don't resolve. */
 /* I assume the base does not have a #fragment on the end; that is not part of the base. */
 /* Thus I won't get url#foo#bar */
 		strcpy(n, rel);
 out_n:
 		debugPrint(5, "= %s", n);
+		hrefContext = false;
 		return n;
 	}
 
@@ -843,13 +893,14 @@ out_n:
 		n[0] = 0;
 		if (s != rel) {
 /* It didn't have http in front of it before, put it on now. */
+/* This is old; it shouldn't happen any more. */
 			strncpy(n, s, l);
 			strcpy(n + l, "://");
 		}
 		strcat(n, rel);
 		goto squash;
 	}
-	/* relative is already a url */
+// at this point rel is not a url.
 	s = base;
 	if (rel[0] == '/') {
 		s = getDataURL(base);
@@ -878,7 +929,7 @@ out_n:
 	for (p = 0; *s; ++s) {
 		if (*s == '/')
 			p = s;
-		if (strchr("?\1", *s))
+		if (strchr("#?\1", *s))
 			break;
 	}
 	if (!p) {
@@ -1014,7 +1065,7 @@ char *encodePostData(const char *s, const char *keep_chars)
 	if (s == emptyString)
 		return emptyString;
 	if (!keep_chars)
-		keep_chars = "-._~*()!";
+		keep_chars = "-._~()";
 	post = initString(&l);
 	while ((c = *s++)) {
 		if (isalnumByte(c))
@@ -1144,3 +1195,40 @@ void decodeMailURL(const char *url, char **addr_p, char **subj_p, char **body_p)
 	if (body_p)
 		*body_p = decodePostData(url, "body", 0);
 }				/* decodeMailURL */
+
+// Does a url match a pattern, from an entry in .ebrc
+// edbrowse.org matches edbrowse.org and foo.edbrowse.org
+// edbrowse.org/foo matches edbrowse.org/foo/bar
+bool patternMatchURL(const char *url, const char *pattern)
+{
+	char prot[MAXPROTLEN], host[MAXHOSTLEN];
+	const char *path, *q;
+	int hl, dl, ql;
+	if (!url || !pattern)
+		return false;
+	if (!url[0] || !pattern[0])
+		return false;
+// This function has to be threadsafe, so I call getProtHostURL,
+// which is also threadsafe.
+	if (!getProtHostURL(url, prot, host))
+		return false;
+	hl = strlen(host);
+	path = getDataURL(url);
+	q = strchr(pattern, '/');
+	if (!q)
+		q = pattern + strlen(pattern);
+	dl = q - pattern;
+	if (dl > hl)
+		return false;
+	if (!memEqualCI(pattern, host + hl - dl, dl))
+		return false;
+	if (*q == '/') {
+		++q;
+		if (hl != dl || !path)
+			return false;
+		ql = strlen(q);
+		return !strncmp(q, path, ql) &&
+		    (path[ql] == 0 || path[ql] == '/');
+	}			/* domain/path was specified */
+	return hl == dl || host[hl - dl - 1] == '.';
+}
